/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.communication;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.util.Vector;
import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetAgentInterface;
import de.fhg.iais.outdoor.simba.neuralnet.NeuronWrapper;
import de.fhg.iais.outdoor.simba.value.NormalizedDoubleValue;
import de.fhg.iais.outdoor.simba.xml.NeuralNetParser;

public class UDPSchedulerConnection {
	
	//TODO TestCases!
	 
	public static final int FAILURE = -1;
	public static final int COMMAND_COMPLETE = 1;
	public static final int COMMAND_PREPARE_GENERATION = 2;
	public static final int COMMAND_PREPARE_INDIVIDUAL = 3;
	public static final int COMMAND_PREPARE_TRY = 4;
	public static final int COMMAND_EXECUTE_STEP = 5;
	public static final int COMMAND_TERMINATE = 6;
	public static final int COMMAND_RESET_SYSTEM = 7;
	public static final int COMMAND_GET_SYSTEM_INFO = 8;
	public static final int COMMAND_ACK = 9;
	public static final int COMMAND_EXECUTE_NETWORK = 10;
	
	private NeuralNetAgentInterface mNeuralNetController;
	private DatagramSocket mSocket;
	private boolean mShutDown;
	private DataInputStream  mInput;
	private DataOutputStream  mOutput;
	private ByteArrayOutputStream mOutputStream;
	private ByteArrayInputStream mInputStream;
	private byte[] mInputBytes;
	private int mNumberOfTries;
	private int mNumberOfStepsPerTry;
	
	public UDPSchedulerConnection(NeuralNetAgentInterface controller) {
		mNeuralNetController = controller;
		mShutDown = false;
	}
	
	
	
	/**
	 * Start a UDP server that waits for a client to connect.
	 * This method also handles the entire communication between the 
	 * client and the server and executes the simulation steps triggered 
	 * by the client.
	 * 
	 * @param port the port to listen at for client messages.
	 */
	public void startServer(int port) {
		
		while(!mShutDown) {
			while(mSocket == null && !mShutDown) {
				try {
					if(mSocket == null) {
						mSocket = new DatagramSocket(port);
					}
					System.out.println("Server Created at port: " + mSocket.getLocalPort());
					Simulator.getInstance().reportStatus("UDP Server started at port " + port);
					//mServerSocket.setSoTimeout(1000);
				}
				catch(IOException e) {
						mSocket = null;
				}
			}
			if(mSocket == null) {
				return;
			}
			

			mInputBytes = new byte[1000];
			
			mInputStream = new ByteArrayInputStream(mInputBytes);
			mInput = new DataInputStream(mInputStream);
			mOutputStream = new ByteArrayOutputStream();
			mOutput = new DataOutputStream(mOutputStream);

			
			boolean connected = true;
			DatagramPacket outputDatagram = null;
			while(!mShutDown && connected) {		
				try {
					DatagramPacket packet = new DatagramPacket(mInputBytes, 1000);
					mSocket.receive(packet);
					mInput.reset();
					int command = mInput.readInt();
					switch(command) {
						case COMMAND_EXECUTE_STEP:
							{
								int stepNumber = mInput.readInt();
								
								Vector outputNeurons = mNeuralNetController.getOutputNeurons();
								for(int i = 0; i < outputNeurons.size(); ++i) {
									double value = mInput.readDouble();
									((NormalizedDoubleValue) outputNeurons.elementAt(i)).setNormalized(value);
								}

								Simulator.getInstance().getScheduler().executeSimulationStep(stepNumber);

								Vector inputNeurons = mNeuralNetController.getInputNeurons();
								Vector infoNeurons = mNeuralNetController.getInfoNeurons();

								mOutputStream.reset();
								mOutput.writeBoolean(Simulator.getInstance().getScheduler().shouldCurrentTryBeTerminated());

								for(int i = 0; i < inputNeurons.size(); ++i) {
									mOutput.writeDouble(((NormalizedDoubleValue) inputNeurons.elementAt(i)).getNormalized());
								}
								for(int i = 0; i < infoNeurons.size(); ++i) {
									mOutput.writeDouble(((NormalizedDoubleValue) infoNeurons.elementAt(i)).getNormalized());
								}
								mOutput.writeInt(COMMAND_COMPLETE);
								mOutput.flush();
								outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
								mSocket.send(outputDatagram);
								break;
							}
						case COMMAND_PREPARE_GENERATION:
							{
								int generation = mInput.readInt();
								mNumberOfTries = mInput.readInt();
								mNumberOfStepsPerTry = mInput.readInt();
								Simulator.getInstance().getScheduler().prepareNextGeneration(generation, mNumberOfTries, mNumberOfStepsPerTry);
								mOutputStream.reset();
								mOutput.writeInt(COMMAND_COMPLETE);
								mOutput.flush();								
								outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
								mSocket.send(outputDatagram);
								break;
							}
						case COMMAND_PREPARE_INDIVIDUAL:
							{
								int individualNumber = mInput.readInt();
								Simulator.getInstance().getScheduler().prepareNextIndividual(individualNumber);
								mOutputStream.reset();
								mOutput.writeInt(COMMAND_COMPLETE);
								mOutput.flush();
								outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
								mSocket.send(outputDatagram);
								break;
							}
						case COMMAND_PREPARE_TRY:
							{
								int tryNumber = mInput.readInt();
								Simulator.getInstance().getScheduler().prepareNextTry(tryNumber);
								mOutputStream.reset();
								mOutput.writeInt(COMMAND_COMPLETE);
								mOutput.flush();
								outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
								mSocket.send(outputDatagram);
								break;
							}
						case COMMAND_RESET_SYSTEM:
							{
								Simulator.getInstance().clearInstance();
								mOutputStream.reset();
								mOutput.writeInt(COMMAND_COMPLETE);
								mOutput.flush();
								outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
								mSocket.send(outputDatagram);
								break;
							}
						case COMMAND_TERMINATE:
							{
								Simulator.getInstance().getScheduler().terminateSimulationRun();
								mOutputStream.reset();
								mOutput.writeInt(COMMAND_COMPLETE);
								mOutput.flush();
								outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
								mSocket.send(outputDatagram);
								break;
							}
						case COMMAND_GET_SYSTEM_INFO:
							{
								//This usually is the first call a client does.
								System.out.println("Connection requested by client.");
								//Clear the current neural net to be able to receive external control commands.
								Simulator.getInstance().getNeuralNetManager().setCurrentNet(null);
								
								int numberOfInputNeurons = 0;
								int numberOfOutputNeurons = 0;
								int numberOfInfoNeurons = 0;
								if(mNeuralNetController != null) {
									numberOfInputNeurons = mNeuralNetController.getInputNeurons().size();
									numberOfOutputNeurons = mNeuralNetController.getOutputNeurons().size();
									numberOfInfoNeurons = mNeuralNetController.getInfoNeurons().size();
								}
								mOutputStream.reset();
								mOutput.writeInt(numberOfInputNeurons);
								mOutput.writeInt(numberOfOutputNeurons);
								mOutput.writeInt(numberOfInfoNeurons);
								mOutput.writeInt(COMMAND_COMPLETE);
								mOutput.flush();
								outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
								mSocket.send(outputDatagram);
								
								Vector inputNeurons = mNeuralNetController.getInputNeurons();
								Vector outputNeurons = mNeuralNetController.getOutputNeurons();
								Vector infoNeurons = mNeuralNetController.getInfoNeurons();
								
								for(int i = 0; i < numberOfInputNeurons; ++i) {
									mInputStream.reset();
									mSocket.receive(packet);
									mInput.reset();
									int ack = mInput.readInt();
									if(ack != COMMAND_ACK) {
										Logger.getLogger("Simulator")
											.warning("Did not receive an acknowledge signal from client! " + i);
									}
									else {
										NeuronWrapper neuron = (NeuronWrapper) inputNeurons.elementAt(i);
										
										mOutputStream.reset();
										mOutput.writeUTF(neuron.getName());
										mOutput.flush();
										outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
										mSocket.send(outputDatagram);
									}
								}
								
								for(int i = 0; i < numberOfOutputNeurons; ++i) {
									mInputStream.reset();
									mSocket.receive(packet);
									mInput.reset();
									int ack = mInput.readInt();
									if(ack != COMMAND_ACK) {
										Logger.getLogger("Simulator")
											.warning("Did not receive an acknowledge signal from client! " + i);
									}
									else {
										NeuronWrapper neuron = (NeuronWrapper) outputNeurons.elementAt(i);
										
										mOutputStream.reset();
										mOutput.writeUTF(neuron.getName());
										mOutput.flush();
										outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
										mSocket.send(outputDatagram);
									}
								}
								
								for(int i = 0; i < numberOfInfoNeurons; ++i) {
									mInputStream.reset();
									mSocket.receive(packet);
									mInput.reset();
									int ack = mInput.readInt();
									if(ack != COMMAND_ACK) {
										Logger.getLogger("Simulator")
											.warning("Did not receive an acknowledge signal from client! " + i);
									}
									else {
										NeuronWrapper neuron = (NeuronWrapper) infoNeurons.elementAt(i);
										
										mOutputStream.reset();
										mOutput.writeUTF(neuron.getName());
										mOutput.flush();
										outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
										mSocket.send(outputDatagram);
									}
								}
								
								
								
								mOutputStream.reset();
								mOutput.writeInt(COMMAND_COMPLETE);
								mOutput.flush();
								outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
								mSocket.send(outputDatagram);
								
								break;
							}
						case COMMAND_EXECUTE_NETWORK: {
								//Hinton sends the current network as XML. 
								//The network is executed (all tries of this individual)
								//and the result is then sent back to hinton.
								StringBuffer network = new StringBuffer();
								
								int individualId = mInput.readInt();
								int numberOfNetworkFragments = mInput.readInt();
								
								mOutputStream.reset();
								mOutput.writeInt(COMMAND_ACK);
								mOutput.flush();
								outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
								mSocket.send(outputDatagram);
								
								//receive network
								for(int i = 0; i < numberOfNetworkFragments; ++i) {
									mInputStream.reset();
									mSocket.receive(packet);
									mInput.reset();
									network.append(mInput.readUTF());
									
									mOutputStream.reset();
									mOutput.writeInt(COMMAND_ACK);
									mOutput.flush();
									outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
									mSocket.send(outputDatagram);
								}
								
								mInputStream.reset();
								mSocket.receive(packet);
								mInput.reset();
								if(!(mInput.readInt() == COMMAND_ACK)) {
									Logger.getLogger("Simulator")
										.warning("Could not receive ACK after receiving all xml fragments");
								}
								
								//try to create a net from the XML received.
								//NeuralNetParser.create
								NeuralNet net = NeuralNetParser.createNetFromEvosunXML(network.toString(), 
									Simulator.getInstance().getNeuralNetManager().getNeuralNetInterface());
								
								if(net != null) {
									Simulator.getInstance().getNeuralNetManager().setCurrentNet(net);
									
									Simulator.getInstance().getScheduler().prepareNextIndividual(individualId);
									for(int tries = 0; tries < mNumberOfTries; ++tries) {
										Simulator.getInstance().getScheduler().prepareNextTry(tries);
										for(int steps = 0; steps < mNumberOfStepsPerTry; ++steps) {
											Simulator.getInstance().getScheduler().executeSimulationStep(steps);
											if(Simulator.getInstance().getScheduler().shouldCurrentTryBeTerminated()) {
												break;
											}
										}
									}
									
								}
								else {
									Logger.getLogger("Simulator")
										.warning("UDPSchedulerConnection: Could not parse the neural network sent by hinton!");
								}
								Vector infoNeurons = mNeuralNetController.getInfoNeurons();

								//Write info neurons (including the fitness value)
								mOutputStream.reset();
								
								for(int i = 0; i < infoNeurons.size(); ++i) {
									mOutput.writeDouble(((NormalizedDoubleValue) infoNeurons.elementAt(i)).getNormalized());
								}
								
								mOutput.writeInt(COMMAND_COMPLETE);
								mOutput.flush();
								outputDatagram = new DatagramPacket(mOutputStream.toByteArray(), mOutputStream.toByteArray().length, packet.getSocketAddress());
								mSocket.send(outputDatagram);
								
								break;
							}
						default:
							System.out.println("Failure: Unexpected command in UDPSchedulerConnection.");
					}
				}
				catch(IOException e) {
					System.out.println("Client disconnected.");
					Simulator.getInstance().getScheduler().terminateSimulationRun();
					connected = false;
					mSocket = null;
				}
		
			}
			
		}
	
	}
	
	public void stopServer() {
		mShutDown = true;
	}
	
	
	

}
