/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.core;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.geom.AffineTransform;
import java.awt.geom.Line2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.Vector;
import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.PointValue;
import de.fhg.iais.outdoor.simba.value.SimObjectValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;


//TODO Describe the useage of the valueReplaced method.
public class SimObject implements SystemObject, ValueChangedListener {
	
	private boolean mGetCoordinateCalled = false;
	
	Logger mLogger;

	protected PointValue mPosition;
	protected DoubleValue mAngle;
	private PointValue mInitPosition;
	private DoubleValue mInitAngle;
	private SimObjectValue mReferenceObject;
	private IntValue mLayer;
	private BooleanValue mVisible;
	private BooleanValue mActive;
	private Vector mDependencies = new Vector();
	protected String mName;
	protected Polygon mShape;
	protected Color mShapeColor;
	protected boolean mFillShape;
	protected DoubleValue mTransparency;
	private boolean mCalculatedGlobalAngleThisTurn;
	private boolean mCalculatedGlobalShapeThisTurn;
	private boolean mCalculatedGlobalPositionThisTurn;
	private boolean mCalculatedGlobalShapeAsLinesThisTurn;
	private boolean mCalculatedAffineTransformThisTurn;
	private double mGlobalAngle;
	private Polygon mGlobalShape;
	private Point2D.Double mGlobalPosition;
	private Vector mGlobalShapeAsLines;
	private AffineTransform mAffineTransform;
	protected BooleanValue mMarked;
	protected boolean mFixedInitPosition;
	
	
	/**
	 * Standard Constructor. The initial position is set to 0.0/0.0.
	 *
	 */
	public SimObject() {
		this("SimObject", 0.0, 0.0, 0.0, new Polygon());
	}
	
	/**
	 * Constructor.
	 * @param name the name of this SimObject. The name might be modified if there
	 * is already another SimObject with the given name.
	 */
	public SimObject(String name) {
		this(name, 0.0, 0.0, 0.0, new Polygon());
	}
	
	/**
	 * Construcotr.
	 * @param x the x position of the object relative to its reference object
	 * @param y the y position of the object relative to its reference object
	 * @param angle the angle of the object relative to its reference object
	 * @param shape a Polygon describing the shape of the object (lines only).
	 */
	public SimObject(double x, double y, double angle, Polygon shape) {
		this("SimObject", x, y, angle, shape);
	}
	
	public SimObject(String name, double x, double y, double angle, Polygon shape) {
		this(name, x, y, angle, shape, true);
	}
	
	/**
	 * Construcotr.
	 * @param name the name of this SimObject. The name might be modified if there
	 * is already another SimObject with the given name.
	 * @param x the x position of the object relative to its reference object
	 * @param y the y position of the object relative to its reference object
	 * @param angle the angle of the object relative to its reference object
	 * @param shape a Polygon describing the shape of the object (lines only).
	 */
	public SimObject(String name, double x, double y, double angle, Polygon shape, boolean registerAtValueManager) {
		mName = name;
		
		if(registerAtValueManager) {
			int counter = 1;
			while(!Simulator.getInstance().getValueManager().addValue("/SimObjects/" + mName, new SimObjectValue(this))) {
				mName = name + counter;
				counter++;
			}
			name = mName;
		}
		
		mLogger = Logger.getLogger("Simulator");
		
		mFillShape = true;
		mShapeColor = Color.black;
		mShape = new Polygon(shape.xpoints, shape.ypoints, shape.npoints);
		mGlobalShape = new Polygon();
		mGlobalPosition = new Point2D.Double();
		mGlobalShapeAsLines = new Vector();
		mCalculatedGlobalAngleThisTurn = false;
		mCalculatedGlobalShapeThisTurn = false;
		mCalculatedGlobalPositionThisTurn = false;
		mCalculatedGlobalShapeAsLinesThisTurn = false;
		mCalculatedAffineTransformThisTurn = false;
		mGetCoordinateCalled = false;
		
		//Create internal Value objects.
		mPosition = new PointValue(x, y);
		mPosition.addValueReplacedListener(this);
		mPosition.addValueChangedListener(this);
		mInitPosition = new PointValue(x, y);
		mInitPosition.addValueReplacedListener(this);
		mFixedInitPosition = true;
		mAngle = new DoubleValue(angle);
		mAngle.addValueReplacedListener(this);
		mAngle.addValueChangedListener(this);
		mInitAngle = new DoubleValue(angle);
		mInitAngle.addValueReplacedListener(this);
		mLayer = new IntValue(0);
		mLayer.addValueReplacedListener(this);
		mActive = new BooleanValue(true);
		mActive.addValueReplacedListener(this);
		mVisible = new BooleanValue(true);
		mVisible.addValueReplacedListener(this);
		mTransparency = new DoubleValue(1.0);
		mTransparency.addValueReplacedListener(this);
		mReferenceObject = new SimObjectValue(null);
		mReferenceObject.addValueReplacedListener(this);
		mMarked = new BooleanValue(false);
		
		if(registerAtValueManager) {
			ValueManager vm = Simulator.getInstance().getValueManager();
			vm.addValue("/SimObjects/" + mName + "/LocalPosition", mPosition);
			vm.addValue("/SimObjects/" + mName + "/LocalAngle", mAngle);
			vm.addValue("/SimObjects/" + mName + "/InitPosition", mInitPosition);
			vm.addValue("/SimObjects/" + mName + "/InitAngle", mInitAngle);
			vm.addValue("/SimObjects/" + mName + "/Active", mActive);
			vm.addValue("/SimObjects/" + mName + "/Visible", mVisible);
			vm.addValue("/SimObjects/" + mName + "/Transparency", mTransparency);
			vm.addValue("/SimObjects/" + mName + "/ReferenceObject", mReferenceObject);
			vm.addValue("/SimObjects/" + mName + "/Marked", mMarked);
		}
		
		Simulator.getInstance().addSystemObject(this);
		
		mAffineTransform = new AffineTransform();
	}
	
	public boolean init() {
		return true;
	}
	
	public boolean cleanUp() {
		return true;
	}
	
	public boolean update() {
		//TODO allow disabling of preprocessed global angle / shape.
		invalidateCalculationBuffers();
		return true;
	}
	
	//TODO
	public void reset() {
		if(mFixedInitPosition) {
			mPosition.set(mInitPosition.get().x, mInitPosition.get().y);
			mAngle.set(mInitAngle.get());
		}
		
		mCalculatedGlobalAngleThisTurn = false;
		mCalculatedGlobalShapeThisTurn = false;
		mCalculatedGlobalPositionThisTurn = false;
		mCalculatedGlobalShapeAsLinesThisTurn = false;
		mCalculatedAffineTransformThisTurn = false;
		clearDependencies();
	}
	
	//TODO paint this object. (transformations have to be done to the painter.
	public void paintSelf(Graphics2D g) {
		Color c = g.getColor();
		Composite composite = g.getComposite();
		g.setColor(mShapeColor);
		g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float) mTransparency.get()));
		if(mFillShape) {
			g.fill(mShape);
		}
		else {
			g.drawPolygon(mShape);
		}
		g.setComposite(composite);
		g.setColor(c);
	}
	
	/**
	 * This method is called whenever this SimObject was added to an EnvironmentModel.
	 * Subclasses of SimObject can used this method to register sub components at the
	 * EnvironmentModel.
	 * 
	 * @param model the EnvironmentModel where the child SimObjects should be registered.
	 */
	public void addChildren(EnvironmentModel model) {
		
	}
	
	/**
	 * Sets the visibility attribute of this SimObject.
	 * Invisible objects are simulated, but not painted in visualizations.
	 * Users can therefore select the SimObjects they want to see in the visualization.
	 * 
	 * TODO Check whether a visibility level would allow more flexibilty for drawing.
	 * 
	 * @param visible the new state of the visibility attribute.
	 */
	public void setVisible(boolean visible) {
		mVisible.set(visible);
	}
	
	/**
	 * Returns true if this SimObject is visible, thus is being painted in visualizations.
	 * 
	 * @return true if this SimObject is visible, otherwise false.
	 */
	public boolean isVisible() {
		return mVisible.get();
	}
	
	/**
	 * Sets the color of the Polgon shape. If the polygon is filled, then the entire
	 * shape is also filled with this color.
	 * 
	 * @param color the new color of the shape when drawn.
	 */
	public void setShapeColor(Color color) {
		if(color != null) {
			mShapeColor = color;
		}
	}
	
	/**
	 * Returns the color the shape of this SimObject is painted in visualizations.
	 * 
	 * @return the current color of the shape.
	 */
	public Color getShapeColor() {
		return mShapeColor;
	}
	
	/**
	 * Sets the alpha channel for the object shape. 
	 * The object is fully transparent at a value of 0.0, and fully visible
	 * at 1.0.
	 * @param transparency the transparency for the shape during drawing.
	 */
	public void setShapeTransparency(double transparency) {
		mTransparency.set(transparency);
	}
	
	/**
	 * Returns the transparency of this object.
	 * @return the transparency.
	 */
	public double getShapeTransparency() {
		return mTransparency.get();
	}
	
	/**
	 * Determins whether the shape is filled or only outlined during drawing.
	 * If it is set to be filled, then the shape color will also be used to
	 * fill the shape.
	 * 
	 * @param fill indicates whether to fill the shape when drawing.
	 */
	public void setFillShape(boolean fill) {
		mFillShape = fill;
	}
	
	/**
	 * Returns true if the shape is filled during drawing.
	 * @return true if the shape is filled during drawing, otherwise false.
	 */
	public boolean hasFilledShape() {
		return mFillShape;
	}
	
	/**
	 * Returns the name of this SimObject.
	 * 
	 * @return the name
	 */
	public String getName() {
		return mName;
	}
	
	/**
	 * Sets the current reference object. The coordinates of the SimObject
	 * always is relative to the reference objects coordinates.
	 * An independent, global object should have the environment object as
	 * reference object. To avoid numerical problems, never create cycles 
	 * of references.
	 * 
	 * @param object the reference object of this SimObject.
	 */
	public void setReferenceObject(SimObject object) {
		mReferenceObject.set(object);
	}
	
	/**
	 * Returns the current reference object.
	 * 
	 * @return the current reference object.
	 */
	public SimObject getReferenceObject() {
		return mReferenceObject.get();
	}
	
	public SimObjectValue getReferenceObjectValue() {
		return mReferenceObject;
	}
	
	/**
	 * Returns the current coordinate point of this object relative to
	 * its reference object.
	 * 
	 * @return the x coordinate.
	 */
	public Point2D.Double getPosition() {
		return mPosition.get();
	}
	
	public PointValue getPositionValue() {
		return mPosition;
	}
	
	/**
	 * Sets the x and y coordinate of this object relative to its 
	 * reference object.
	 * @param x the new x coordinate relative to its reference object.
	 * @param y the new y coordinate relative to its reference object.
	 */
	public void setPosition(double x, double y) {
		mPosition.set(x, y);
	}
	
	public void setPosition(Point2D.Double point) {
		mPosition.set(point.x, point.y);
	}
	
	
	/**
	 * Returns the position in global coordinates, taking into account
	 * the reference object of this SimObject.
	 * 
	 * @return the global position of the SimObject.
	 */
	public Point2D.Double getGlobalPosition() {
		if(!mCalculatedGlobalPositionThisTurn) {
			SimObject referenceObject = mReferenceObject.get();
			Point2D.Double position = mPosition.get();
			if(referenceObject != null) {
				AffineTransform transform = new AffineTransform();
				referenceObject.addAffineTransform(transform);
				transform.transform(position, mGlobalPosition);
			}
			else {
				mGlobalPosition.setLocation(position.x, position.y);
			}
			mCalculatedGlobalPositionThisTurn = true;
		}
		return mGlobalPosition;
	}
	
	/**
	 * Sets the angle of this SimObject relative to its reference object.
	 * @param angle the new angle.
	 */
	public void setAngle(double angle) {
		mAngle.set(angle);
	}
	
	/**
	 * Returns the angle relative to the reference object.
	 * 
	 * @return the angle.
	 */
	public double getAngle() {
		return mAngle.get();
	}
	
	public DoubleValue getAnglePosition() {
		return mAngle;
	}
	
	public DoubleValue getInitialAngleValue() {
		return mInitAngle;
	}
	
	
	/**
	 * Returns the global angle, taking into accound the own angle
	 * and the angles of all relative SimObects in the direct line.
	 * 
	 * @return the global ycoordinate.
	 */
	public synchronized double getGlobalAngle() {
		if(mGetCoordinateCalled) {
			mLogger.warning("There was a cycle detected in method getGlobalAngle().");
			mGetCoordinateCalled = false;
			return 0.0;
		}
		if(!mCalculatedGlobalAngleThisTurn) {
			mGetCoordinateCalled = true;
			mGlobalAngle = mAngle.get();
			if(mReferenceObject.get() != null) {
				mGlobalAngle += mReferenceObject.get().getGlobalAngle();
			}
			while(mGlobalAngle > 360) {
				mGlobalAngle -= 360;
			}
			while(mGlobalAngle < 0.0) {
				mGlobalAngle += 360;
			}
			mGetCoordinateCalled = false;
			mCalculatedGlobalAngleThisTurn = true;
		}
		return mGlobalAngle;
	}
	
	/**
	 * Returns a Polygon with points described in global coordinates.
	 * @return a Polygon in global coordinates.
	 */
	public Polygon getGlobalShape() {
		if(!mCalculatedGlobalShapeThisTurn) {
			mGlobalShape.reset();
			AffineTransform transformation = new AffineTransform();
			addAffineTransform(transformation);
			PathIterator i = mShape.getPathIterator(transformation);
			if(!i.isDone()) {
				double[] coords = new double[2];
				while(!i.isDone()) {
					i.currentSegment(coords);
					i.next();
					if(!i.isDone()) {
						mGlobalShape.addPoint((int) coords[0], (int)coords[1]);
					}
				}
			}
			mCalculatedGlobalShapeThisTurn = true;
		}
		return mGlobalShape;
	}
	
	
	
	public Vector getShapeAsGlobalLineSegments() {
		if(!mCalculatedGlobalShapeAsLinesThisTurn) {
			mGlobalShapeAsLines.clear();
			AffineTransform transformation = new AffineTransform();
			addAffineTransform(transformation);
//			Polygon shape = getGlobalShape();
			PathIterator i = mShape.getPathIterator(transformation);
			if(!i.isDone()) {
				double[] coords = new double[2];
				double[] first = new double[2];
				i.currentSegment(coords);
				i.currentSegment(first);
				double[] secondCoords = new double[2];
				i.next();
				while(!i.isDone()) {
					i.currentSegment(secondCoords);
					if(coords[0] != secondCoords[0] || coords[1] != secondCoords[1]) {
						//add a new line object only if the two points do NOT coincide.
						mGlobalShapeAsLines.add(new Line2D.Double(coords[0], coords[1], secondCoords[0], secondCoords[1]));
					}
					i.next();
					//System.out.println("Found: 1: " + coords[0] + " " + coords[1] + " / " + secondCoords[0] + " " + secondCoords[1]);
					coords[0] = secondCoords[0];
					coords[1] = secondCoords[1];
				}
				if(coords[0] != first[0] || coords[1] != first[1]) {
					//add a new line object only if the two points do NOT coincide.
					mGlobalShapeAsLines.add(new Line2D.Double(coords[0], coords[1], first[0], first[1]));
				}
				mCalculatedGlobalShapeAsLinesThisTurn = true;
			}
			
			//TODO Reuse of global shape would speed up performance on the cost of accuracy (int instead of double)
//			int[] xpoints = shape.xpoints;
//			int[] ypoints = shape.ypoints;
//			
//			for(int i = 1; i < shape.npoints; ++i) {
//				if(xpoints[i - 1] != xpoints[i] || ypoints[i - 1] != ypoints[i]) {
//					mGlobalShapeAsLines.add(new Line2D.Double(xpoints[i - 1], ypoints[i - 1], xpoints[i], ypoints[i]));
//				}
//			}
//			if(shape.npoints > 2) {
//				int i = shape.npoints - 1;
//				if(xpoints[i] != xpoints[0] || ypoints[i] != ypoints[0]) {
//					mGlobalShapeAsLines.add(new Line2D.Double(xpoints[i], ypoints[i], xpoints[0], ypoints[0]));
//				}
//			}
			
			mCalculatedGlobalShapeAsLinesThisTurn = true;
		}
		return new Vector(mGlobalShapeAsLines);
	}
	
	
	/**
	 * Adds the own coordinate tranformations to the AffineTransformation.
	 * 
	 * @param transform the transformation object where the own transformations have to be added.
	 */
	public synchronized void addAffineTransform(AffineTransform transform) {
		if(!mCalculatedAffineTransformThisTurn) {
			if(mGetCoordinateCalled) {
				mLogger.warning("There was a cycle detected in method addAffineTransform().");
				mGetCoordinateCalled = false;
				return;
			}
			mGetCoordinateCalled = true;
			mAffineTransform = new AffineTransform();
			SimObject referenceObject = mReferenceObject.get();
			if(referenceObject != null) {
				referenceObject.addAffineTransform(mAffineTransform);
			}
			mAffineTransform.translate(mPosition.getX(), mPosition.getY());
			mAffineTransform.rotate(Math.toRadians(mAngle.get())); //angle in rad.
			
			mGetCoordinateCalled = false;
			mCalculatedAffineTransformThisTurn = true;
		}
		transform.concatenate(mAffineTransform);
	}
	
	/**
	 * Returns an AffineTransform object that maps local points to the global coordinate system.
	 * @return an AffineTransform to map local points to the global coordinate system.
	 */
	public AffineTransform getAffineTransform() {
		AffineTransform transformation = new AffineTransform();
		addAffineTransform(transformation);
		return transformation;
	}
	

	
	/**
	 * Sets the Polygon that represents the shape of the SimObject.
	 * 
	 * @param shape the new shape of the SimObject.
	 */
	public void setShape(Polygon shape) {
		mShape = new Polygon(shape.xpoints, shape.ypoints, shape.npoints);
	}
	
	
	/**
	 * Returns the Polygon that represents the shape of this SimObject.
	 * 
	 * @return the shape of the SimObject.
	 */
	public Polygon getShape() {
		return mShape;
	}
	
	/**
	 * Sets the drawing layer of this SimObject. The greater the layer 
	 * number, the later the object is drawn. The layer parameter can 
	 * be used to control the order in which objects are drawn. This is 
	 * important because overlapping objects hide each other, so this way
	 * it can be controlled which object is in the front.
	 * 
	 * @param layer the current layer of the SimObject.
	 */
	public void setLayer(int layer) {
		mLayer.set(layer);
	}
	
	/**
	 * Returns the drawing layer of this SimObject.
	 * @see setLayer
	 * @return the current drawing layer.
	 */
	public int getLayer() {
		return mLayer.get();
	}
	
	public IntValue getLayerObject() {
		return mLayer;
	}
	
	/**
	 * Adds an object this SimObject is dependent on. Dependency here means
	 * that the an object has to be updated before this SimObject is updated.
	 * @param object an object this SimObject depends on.
	 * Avoid circular dependencies.
	 * @return true if the object could be added to the dependency list.
	 */
	public boolean addDependency(SimObject object) {
		if(object == null || mDependencies.contains(object)) {
			return false;
		}
		mDependencies.add(object);
		return true;
	}
	
	/**
	 * Removes an object this SimObject was dependent on. This can be used to 
	 * remove an automatically set dependency manually.
	 * 
	 * @param object the object to remove from the dependency list.
	 */
	public void removeDependency(SimObject object) {
		if(object != null && mDependencies.contains(object)) {
			mDependencies.remove(object);
		}
	}
	
	/**
	 * Adds the content of the SimObjectSet to the list of dependencies.
	 * Note: the objects are simply added with addDependency, so there are
	 * no duplicates with already contained dependencies.
	 * TODO check if it makes sense to separate dependency lists, like with obstacles in Scanner.
	 * 
	 * @param objects
	 * @return
	 */
	public boolean addDependencies(SimObjectSet objects) {
		if(objects == null) {
			return false;
		}
		for(Iterator i = objects.getIterator(); i.hasNext();) {
			addDependency((SimObject) i.next());
		}
		return true;
	}
	
	/**
	 * Removes all dependencies from the SimObject.
	 * If a reference object is set, then this object remains 
	 * in the dependency list.
	 */
	public void clearDependencies() {
		mDependencies.clear();
	}
	
	/**
	 * Returns a copy of the list of dependencies.
	 * @return the list of obejcts this SimObject depends on.
	 */
	public Vector getDependencies() {
		Vector dependencies = new Vector(mDependencies);
		if(mReferenceObject.get() != null) {
			dependencies.add(mReferenceObject.get());
		}
		return dependencies;
	}
	
	/**
	 * Sets this SimObject active (true) or inactive (false).
	 * Inactive objects are neither present in the simulation, nor 
	 * are painted in the visualizations.
	 * 
	 * This can be used to remove an Object from the Simulation temporarily.
	 * 
	 * @param active indicates whether this SimObject is currently simulated.
	 */
	public void setActive(boolean active) {
		mActive.set(active);
	}
	
	/**
	 * Returns true if this SimObject currently is active.
	 * @see setActive()
	 * @return true if this SimObject is currently active.
	 */
	public boolean isActive() {
		return mActive.get();
	}
	
	public BooleanValue getActiveValue() {
		return mActive;
	}
	
	/**
	 * Called whenever an observed Value object is replaced by another one in the 
	 * ValueManager. The object can then change its observatoin to the new Value object
	 * and cancel the observation of the old Value. When overriding this method
	 * in subclassed do not forget to call this method from the subclass.
	 */
	public void valueReplaced(Value oldValue, Value newValue) {
		mTransparency = (DoubleValue) synchronizeValue(mTransparency, oldValue, newValue);
		mPosition = (PointValue) synchronizeValue(mPosition, oldValue, newValue);
		mInitPosition = (PointValue) synchronizeValue(mInitPosition, oldValue, newValue);
		mAngle = (DoubleValue) synchronizeValue(mAngle, oldValue, newValue);
		mInitAngle = (DoubleValue) synchronizeValue(mInitAngle, oldValue, newValue);
		mLayer = (IntValue) synchronizeValue(mLayer, oldValue, newValue);
		mActive = (BooleanValue) synchronizeValue(mActive, oldValue, newValue);
		mVisible = (BooleanValue) synchronizeValue(mVisible, oldValue, newValue);
		mTransparency  = (DoubleValue) synchronizeValue(mTransparency, oldValue, newValue);
		mReferenceObject = (SimObjectValue) synchronizeValue(mReferenceObject, oldValue, newValue);
		mMarked = (BooleanValue ) synchronizeValue(mMarked, oldValue, newValue);
	}
	
	public void setMarked(boolean marked) {
		mMarked.set(marked);
	}
	
	public boolean isMarked() {
		return mMarked.get();
	}
	
	public BooleanValue getMarkedValue() {
		return mMarked;
	}
	
	
	
	public void valueChanged(Value value) {
//		if(value == null) {
//			return;
//		}
//		if(value == mPosition) {
//			invalidateCalculationBuffers();
//		}
//		else if(value == mAngle) {
//			invalidateCalculationBuffers();
//		}
	}
	
	public void invalidateCalculationBuffers() {
		mCalculatedGlobalAngleThisTurn = false;
		mCalculatedGlobalShapeThisTurn = false;
		mCalculatedGlobalPositionThisTurn = false;
		mCalculatedGlobalShapeAsLinesThisTurn = false;
		mCalculatedAffineTransformThisTurn = false;
	}
	

	protected Value synchronizeValue(Value value, Value oldValue, Value newValue) {
		return ValueManager.synchronizeValue(value, oldValue, newValue, this);
	}
	
	public void setFixedInitPosition(boolean fixed) {
		mFixedInitPosition = fixed;
	}
	
	public boolean hasFixedInitPosition() {
		return mFixedInitPosition;
	}
	
}
