/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.core;


import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;
import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.evolution.EvolutionManager;
import de.fhg.iais.outdoor.simba.fitness.FitnessManager;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetManager;
import de.fhg.iais.outdoor.simba.statistic.StatisticManager;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.StringValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.visualization.OptimizePerformanceListener;
import de.fhg.iais.outdoor.simba.visualization.VisualizationManager;

public class Simulator implements ValueChangedListener {
	
//	TODO TestCases!
	
	private static Simulator sInstance;
	private ValueManager mValueManager;
	private FitnessManager mFitnessManager;
	private Scheduler mScheduler;
	private VisualizationManager mVisualizationManager;
//	private ConfigurationManager mConfigManager;
	private PlugInManager mPlugInManager;
	private Vector mSystemObjects = new Vector();
	private IntValue mSimulationDelay;
	private BooleanValue mPauseSimulation;
	private StringValue mStatusMessages;
	private boolean mInitializedSuccessful;
	private BooleanValue mOptimizePerformance;
	private Vector mOptimizePerformanceListeners = new Vector();
	private NeuralNetManager mNeuralNetManager;
	private EvolutionManager mEvolutionManager;
	private StatisticManager mStatisticManager;
	private String mConfigDirectoryName;
	private static boolean mSimulatorCreated = false;
	private static boolean mSchedulerCreated = false;
	
	private Properties mSystemProperties;
	
	/**
	 * Constructor.
	 *
	 */
	public Simulator() {
		mConfigDirectoryName = "config/";
		setupSimulator();
	}
	
	public void setupSimulator() {
		File configDir = new File(mConfigDirectoryName);
		if(!configDir.exists() || !configDir.isDirectory()) {
			configDir.mkdir();
		}
		
		mSystemProperties = new Properties();
		try {
			mSystemProperties.load(
					new FileInputStream(mConfigDirectoryName + "properties.conf"));
		} catch (FileNotFoundException e) {
		} catch (IOException e) {
		}
		
		mValueManager = new ValueManager();
		mFitnessManager = new FitnessManager();
		mVisualizationManager = new VisualizationManager(mValueManager);
		mPlugInManager = new PlugInManager();
		mEvolutionManager = new EvolutionManager(mValueManager);
		mNeuralNetManager = new NeuralNetManager(mValueManager);
		mStatisticManager = new StatisticManager();
		mScheduler = new Scheduler(this);
		
		
//		mConfigManager = new ConfigurationManager();
		mSimulationDelay = new IntValue(0);
		getValueManager().addValue("/Simulation/Delay", mSimulationDelay);
		mPauseSimulation = new BooleanValue(false);
		getValueManager().addValue("/Simulation/Pause", mPauseSimulation);
		mInitializedSuccessful = false;
		mStatusMessages = new StringValue();
		getValueManager().addValue("/StatusMessages", mStatusMessages);
		mOptimizePerformance = new BooleanValue(false);
		mValueManager.addValue("/Visualization/OptimizePerformance", mOptimizePerformance);
		mOptimizePerformance.addValueChangedListener(this);
		
		addSystemObject(mNeuralNetManager);
		addSystemObject(mFitnessManager);
		addSystemObject(mEvolutionManager);
		addSystemObject(mStatisticManager);
	}
	
	/**
	 * Returns the singleton object of this class.
	 * This singleton object should be used by all simulation objects,
	 * so they do not require backreferences to the Simulator object used.
	 * 
	 * @return the singleton instance of this class.
	 */
	public static Simulator getInstance() {
		if(!mSimulatorCreated) {
			synchronized(Simulator.class) {
				sInstance = new Simulator();
				mSimulatorCreated = true;
			}
		}
		return sInstance;
	}
	
	/**
	 * Sets the currently used scheduler object.
	 * @param scheduler the curently used scheduler.
	 */
	public void setScheduler(Scheduler scheduler) {
		mScheduler = scheduler;
	}
	
	/**
	 * Returns the currently used scheduler object.
	 * @return the currently used scheudler.
	 */
	public Scheduler getScheduler() {
		if(!mSchedulerCreated || mScheduler == null) {
				synchronized(this) {
				mSchedulerCreated = false;
				mScheduler = new Scheduler(this);
				mSchedulerCreated = true;
			}
		}
		return mScheduler;
	}
	
	/**
	 * Returns the global ValueManager.
	 * @return
	 */
	public ValueManager getValueManager() {
		return mValueManager;
	}
	
	/**
	 * Returns the global VisualizationManager.
	 * @return the VisualizationManager.
	 */
	public VisualizationManager getVisualizationManager() {
		return mVisualizationManager;
	}
	
	/**
	 * Returns the global PlugInManager.
	 * @return the PlugInManager.
	 */
	public PlugInManager getPlugInManager() {
		return mPlugInManager;
	}
	
	/**
	 * Returns the global FitnessManager.
	 * @return the FitnessManager.
	 */
	public FitnessManager getFitnessManager() {
		return mFitnessManager;
	}
	
	/**
	 * Returns the global neuralNetManager.
	 * TODO check how to generatlize this, e.g. with a ControllerManager.
	 * 
	 * @return the NeuralNetManager.
	 */
	public NeuralNetManager getNeuralNetManager() {
		return mNeuralNetManager;
	}
	
	/**
	 * Returns the global EvolutionManager.
	 * @return the EvolutionManager.
	 */
	public EvolutionManager getEvolutionManager() {
		return mEvolutionManager;
	}
	
	public StatisticManager getStatisticManager() {
		return mStatisticManager;
	}
	
	/**
	 * Returns the global ConfigurationManager.
	 * @return the ConfigurationManager.
	 */
//	public ConfigurationManager getConfigurationManager() {
//		return mConfigManager;
//	}
	
	/**
	 * Initializes the Simulator and all related objects.
	 * 
	 * @return true if successful, otherwise false. 
	 */
	public boolean init() {
		boolean ok = true;
		if(mInitializedSuccessful) {
			Logger.getLogger("Simulator").warning("Initialization started more than once! This may lead to problems!");
		}
		for(Iterator i = mSystemObjects.iterator(); i.hasNext();) {
			SystemObject obj = (SystemObject) i.next();
			if(!obj.init()) {
				Logger.getLogger("Simulator")
					.warning("Simulator: Could not initialize SystemObject [" + obj + "]");
				ok = false;
			}
		}
		if(mScheduler == null) {
			Logger.getLogger("Simulator")
				.warning("Simulator: There was no Scheduler object set before the initialization!");
			return false;
		}
		if(! mScheduler.init()) {
			ok = false;
		}
		mInitializedSuccessful = ok;
		if(!mInitializedSuccessful) {
			Logger.getLogger("Simulator")
				.warning("Simulator: The initialization failed! The simulator can not be started!");
		}
		return ok;
	}
	
	/**
	 * Adds a SystemObject to the Simulator. Registered SystemObjects are
	 * initialized at startup and cleaned up at shutdown automatically.
	 * 
	 * @param obj the SystemObject to add.
	 * @return true if successful, false if already contained.
	 */
	public boolean addSystemObject(SystemObject obj) {
		if(obj == null || mSystemObjects.contains(obj)) {
			return false;
		}
		mSystemObjects.add(obj);
		return true;
	}
	
	/**
	 * Called before the system is destroyed, so any object gets the 
	 * chance to release ressources.
	 * 
	 * @return true if successful.
	 */
	public boolean shutDown() {
		boolean ok = true;
		for(Iterator i = mSystemObjects.iterator(); i.hasNext();) {
			SystemObject obj = (SystemObject) i.next();
			if(!obj.cleanUp()) {
				Logger.getLogger("Simulator")
					.warning("Simulator: Could not clean up SystemObject [" + obj + "]");
				ok = false;
			}
		}
		try {
			mSystemProperties.store(
					new FileOutputStream(mConfigDirectoryName + "properties.conf"), "Simba System Properties");
		} catch (FileNotFoundException e) {
		} catch (IOException e) {
		}
		return ok;
	}
	
	public void clearInstance() {
		shutDown();
		mSystemObjects.clear();
		mOptimizePerformanceListeners.clear();
		setupSimulator();
	}
	
	
	/**
	 * Emulates the execution of an evolution run.
	 * This is useful to start the simulator for tests during setup.
	 * 
	 * @param generations
	 * @param individuals
	 * @param tries
	 * @param steps
	 */
	public void emulateEvolutionRun(int generations, int individuals, int tries, int steps) {
		if(!mInitializedSuccessful) {
			Logger.getLogger("Simulator")
				.warning("Simulator: Can not execute the simulator without a successful initialization!");
			return;
		}
		mScheduler.emulateEvolutionRun(generations, individuals, tries, steps);
	}
	
	/**
	 * Sets the delay after each simulation step. This delay can be used to slow down
	 * the simulation to a convenient speed if the simulation is observed. 
	 * The delay is measured in milliseconds.
	 * 
	 * @param delay the current delay after each step.
	 */
	public void setSimulationDelay(int delay) {
		mSimulationDelay.set(delay);
	}
	
	/**
	 * Returns the millisecons that is waited after each step.
	 * @return the delay after each step.
	 */
	public int getSimulationDelay() {
		return mSimulationDelay.get();
	}
	
	/**
	 * Sets the delay after each simulation step. This delay can be used to slow down
	 * the simulation to a convenient speed if the simulation is observed. 
	 * The delay is measured in milliseconds.
	 * 
	 * @param delay the current delay after each step.
	 */
	public void pauseSimulation(boolean pause) {
		mPauseSimulation.set(pause);
	}
	
	/**
	 * Returns the millisecons that is waited after each step.
	 * @return the delay after each step.
	 */
	public boolean isPaused() {
		return mPauseSimulation.get();
	}
	
	/**
	 * Replaces the current message in the StringValue [/StatusMessages].
	 * To use this information users can register a ValueChanged listener
	 * at this StringValue to get notified if a new status message was set.
	 * 
	 * @param message the new status message.
	 */
	public void reportStatus(String message) {
		mStatusMessages.set(message);
	}
	
	/**
	 * Sets the internal performance optimization flag to true.
	 * This signal can be used by visualizations to optimize their 
	 * visualization behavior, e.g. by not displaying information.
	 * 
	 * @param optimize
	 */
	public synchronized void optimizePerformance(boolean optimize) {
		mOptimizePerformance.set(optimize);
	}
	
	public boolean isPerformanceOptimized() {
		return mOptimizePerformance.get();
	}
	
	public BooleanValue getPerformanceOptimizationValue() {
		return mOptimizePerformance;
	}
	
	public synchronized boolean addOptimizePerformanceListener(OptimizePerformanceListener listener) {
		if(listener == null || mOptimizePerformanceListeners.contains(listener)) {
			return false;
		}
		mOptimizePerformanceListeners.add(listener);
		return true;
	}
	
	public synchronized boolean removeOptimizePerformanceListener(OptimizePerformanceListener listener) {
		if(listener == null || !mOptimizePerformanceListeners.contains(listener)) {
			return false;
		}
		mOptimizePerformanceListeners.remove(listener);
		return true;
	}
	
	public Vector getOptimizePerformanceListeners() {
		return new Vector(mOptimizePerformanceListeners);
	}
	

	public Vector getSystemObjects() {
		return new Vector(mSystemObjects);
	}
	
	/**
	 * Returns the global system properties object. This object can be used by any component
	 * to add global properties. The system properties are saved automatically after 
	 * the clean up phase.
	 * @return
	 */
	public Properties getSystemProperties() {
		return mSystemProperties;
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		mSimulationDelay = (IntValue) ValueManager.synchronizeValue(mSimulationDelay, oldValue, newValue, this);
		mPauseSimulation = (BooleanValue) ValueManager.synchronizeValue(mPauseSimulation, oldValue, newValue, this);
		mStatusMessages = (StringValue) ValueManager.synchronizeValue(mStatusMessages, oldValue, newValue, this);
		mOptimizePerformance = (BooleanValue) ValueManager.synchronizeValue(mOptimizePerformance, oldValue, newValue, this);
	}

	public void valueChanged(Value value) {
		if(value == mPauseSimulation) {
			if(mPauseSimulation.get()) {
				reportStatus("Simulation was paused.");
			}
			else {
				reportStatus("Simulation was resumed.");
			}
		}
		if(value == mOptimizePerformance) {
			for(Iterator i = mOptimizePerformanceListeners.iterator(); i.hasNext();) {
				((OptimizePerformanceListener) i.next()).optimizePerformance(mOptimizePerformance.get());
			}
		}
	}
	
	public String getConfigDirectoryName() {
		return mConfigDirectoryName;
	}

}
