/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.devices;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;

import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.Sensor;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class AngleToPointScanner extends Sensor {
	
	protected DoubleValue mMinScanAngle;
	protected DoubleValue mMaxScanAngle;
	protected DoubleValue mAngle;
	private double mCurrentAngle;
	protected Color mObstacleDetectedColor;
	protected int mDrawingSize;
	
	protected Vector mObstacles = new Vector();
	
	public AngleToPointScanner(String name, double x, double y, double angle) 
	{
		this(name, x, y, angle, -180.0, 180.0);
	}
	
	
	public AngleToPointScanner(String name, double x, double y, double angle, 
			double minScanAngle, double maxScanAngle) 
	{
		super(name, x, y, angle);
		
		if(minScanAngle < -180.0) {
			minScanAngle = -180.0;
		}
		if(maxScanAngle > 180.0) {
			maxScanAngle = 180.0;
		}
		mMinScanAngle = new DoubleValue(minScanAngle);
		mMaxScanAngle = new DoubleValue(maxScanAngle);
		mAngle = new DoubleValue(0.0);
		mCurrentAngle = 0.0;
		
		mObstacleDetectedColor = Color.red;
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		vm.addValue("/SimObjects/" + getName() + "/MinScanAngle", mMinScanAngle);
		vm.addValue("/SimObjects/" + getName() + "/MaxScanAngle", mMaxScanAngle);
		vm.addValue("/SimObjects/" + getName() + "/Angle", mAngle);
		
		mMinScanAngle.addValueReplacedListener(this);
		mMaxScanAngle.addValueReplacedListener(this);
		mAngle.addValueReplacedListener(this);
		
		mDrawingSize = 10;
	}
	
	/**
	 * Calculates the distance to the closest obstacle.
	 * 
	 */
	public boolean update() {
		boolean ok = super.update();
	
		mCurrentAngle = 200.0; //greater than the max bounds -180/180
	
		Point2D.Double ownPosition = getGlobalPosition();
		double ownAngle = getGlobalAngle();
		
		for(int i = 0; i < mObstacles.size(); ++i) {
			SimObject obs = (SimObject) mObstacles.elementAt(i);
			if(!obs.isActive()) {
				continue;
			}
			
			Point2D.Double objPosition = obs.getGlobalPosition();
			
			double x = objPosition.x - ownPosition.x;
			double y = objPosition.y - ownPosition.y;
			
			double angle = Math.toDegrees(Math.atan2(y, x)) - ownAngle;
		
			while(angle > 180.0) {
				angle -= 360.0;
			}
			while(angle < -180.0) {
				angle += 360.0;
			}
		
			if(Math.abs(angle) < Math.abs(mCurrentAngle)) {
				mCurrentAngle = angle;
			}
		}

		if(mCurrentAngle > 180.0) {
			mCurrentAngle = 0.0;
		}
		double dist = Math.max(Math.min(addNoise(mCurrentAngle), mMaxScanAngle.get()), mMinScanAngle.get());
		mAngle.set(dist);

		return ok;
	}
	
	
	
	public void paintSelf(Graphics2D g) {
		Color c = g.getColor();
		
		g.setColor(mObstacleDetectedColor);
		if(mCurrentAngle != 0.0) {
			g.drawArc(0 - ((int) ((double) mDrawingSize / 2.0)), 0 - ((int) ((double) mDrawingSize/ 2.0)), mDrawingSize, mDrawingSize, 0, ((int) Math.round(mCurrentAngle * -1)));
			g.drawLine(0, 0, ((int) ((double) mDrawingSize / 2.0)), 0);
			g.drawLine(0, 0, (int) (((double) mDrawingSize) * Math.cos(Math.toRadians(mCurrentAngle))), (int) (((double) mDrawingSize) * Math.sin(Math.toRadians(mCurrentAngle))));
		}
		g.setColor(c);
	}
	
	
	/**
	 * Returns the last measured value of this sensor.
	 * @return
	 */
	public double getSensorReading() {
		return mAngle.get();
	}
	
	/**
	 * Sets the maximal scan distance of this sensor.
	 * @param distance the maximal scan distance.
	 */
	public void setMaxScanAngle(double distance) {
		mMaxScanAngle.set(distance);
	}
	
	/**
	 * Returns the maximal scan distance of this sensor.
	 * @return the maximal scan distance.
	 */
	public double getMaxScanAngle() {
		return mMaxScanAngle.get();
	}
	
	
	/**
	 * Sets the minimal scan distance of this sensor.
	 * @param distance the miniman scan distance.
	 */
	public void setMinScanAngle(double distance) {
		mMinScanAngle.set(distance);
	}
	
	/**
	 * Returns the minimal scan distance of this sensor. 
	 * @return the minimal scan distance.
	 */
	public double getMinScanAngle() {
		return mMinScanAngle.get();
	}
	
	
	/**
	 * Sets the color used to indicate that an obstacle was detected.
	 * If the color object is null, then the color of the scanner
	 * is the same whether an obstacle is detected or not.
	 * 
	 * @param color the color to use when on obstacle was deteceted.
	 */
	public void setObstacleDetectedColor(Color color) {
		mObstacleDetectedColor = color;
	}
	
	/**
	 * Returns the Color object used to draw the scanner when an obstacle
	 * was detected.
	 * 
	 * @return the color of the scanner when an obstacle was detected.
	 */
	public Color getObstacleDetectedColor() {
		return mObstacleDetectedColor;
	}
	
	
	/**
	 * Adds an obstacle this distance scanner considers during its scans.
	 * 
	 * @param obstacle an obstacle that can be detected by this scanner.
	 * @return true if the obstacle object could be added.
	 */
	public boolean addObstacle(SimObject obstacle) {
		if(obstacle == null || mObstacles.contains(obstacle)) {
			return false;
		}
		mObstacles.add(obstacle);
		addDependency(obstacle);
		return true;
	}
	
	
	/**
	 * Returns a copy of the obstacle list of this scanner.
	 * 
	 * @return a copy of the obstacle list.
	 */
	public Vector getObstacles() {
		return new Vector(mObstacles);
	}
	
	
	public void setDrawingSize(int size) {
		mDrawingSize = size;
	}
	
	public int getDrawingSize() {
		return mDrawingSize;
	}
	
	

	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		
		mMinScanAngle = (DoubleValue) synchronizeValue(mMinScanAngle, oldValue, newValue);
		mMaxScanAngle = (DoubleValue) synchronizeValue(mMaxScanAngle, oldValue, newValue);
		mAngle = (DoubleValue) synchronizeValue(mAngle, oldValue, newValue);
	}
	
	

}
