/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.devices;

import java.awt.Polygon;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;

import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class CircularPolygon extends SimObject {

	protected DoubleValue mRadius;
	protected IntValue mCorners;
	protected DoubleValue mMaxAngle;
	
	
	/**
	 * Constructs a ciruclarPolygon. This is a polygon with a fixed number of corners,
	 * regularly distributed over a circle in the range of 0.0 and maxAngle angle.
	 * All corners are exactly (maxAngle / numberOfCorners) degree apart from their 
	 * neighbouring corners measured from the origin (0.0/0.0). The distance of all points 
	 * to the origin is given by a radius.
	 * This constructor sets the maximal angle to 360.0 degree,
	 * so the CircularPolygon always approximates a regular circle with the given radius.   
	 *
	 * @param name the name of this CircularPolygon.
	 * @param x the x position.
	 * @param y the y position.
	 * @param angle the orientation.
	 * @param radius the radius, thus the distance of all points from the origin.
	 * @param corners the number of corners for the given segment.
	 */
	public CircularPolygon(String name, double x, double y, double angle, double radius, int corners) {
		this(name, x, y, angle, radius, corners, 360.0, true);
	}
	
	/**
	 * Constructs a ciruclarPolygon. This is a polygon with a fixed number of corners,
	 * regularly distributed over a circle in the range of 0.0 and maxAngle angle.
	 * All corners are exactly (maxAngle / numberOfCorners) degree apart from their 
	 * neighbouring corners measured from the origin (0.0/0.0). The distance of all points 
	 * to the origin is given by a radius.
	 * This constructor sets the maximal angle to 360.0 degree,
	 * so the CircularPolygon always approximates a regular circle with the given radius.   
	 *
	 * @param name the name of this CircularPolygon.
	 * @param x the x position.
	 * @param y the y position.
	 * @param angle the orientation.
	 * @param radius the radius, thus the distance of all points from the origin.
	 * @param corners the number of corners for the given segment.
	 */
	public CircularPolygon(String name, double x, double y, double angle, double radius, int corners, double maxAngle) {
		this(name, x, y, angle, radius, corners, maxAngle, true);
	}
	
	
	/**
	 * Constructs a ciruclarPolygon. This is a polygon with a fixed number of corners,
	 * regularly distributed over a circle in the range of 0.0 and maxAngle angle.
	 * All corners are exactly (maxAngle / numberOfCorners) degree apart from their 
	 * neighbouring corners measured from the origin (0.0/0.0). The distance of all points 
	 * to the origin is given by a radius.
	 * If the angle is 360 degree, then the CircularPolygon approximates a regular circle
	 * with a given radius.   
	 * If the angle is less then 360 degree, then an additional point will be inserted 
	 * at the origin (0.0/0.0). The results are slices of a circle, like in chart diagrams.
	 * 
	 * @param name the name of this CircularPolygon.
	 * @param x the x position.
	 * @param y the y position.
	 * @param angle the orientation.
	 * @param radius the radius, thus the distance of all points from the origin.
	 * @param corners the number of corners for the given segment.
	 * @param maxAngle the maximal angle to which corners are fitted. 360 leads to a circle, less than 360 to a circle slice.
	 */
	public CircularPolygon(String name, double x, double y, double angle, double radius, int corners, double maxAngle, boolean registerAtValueManager) {
		super(name, x, y, angle, new Polygon(), registerAtValueManager);
		mRadius = new DoubleValue(radius);
		mCorners = new IntValue(corners);
		mMaxAngle = new DoubleValue(maxAngle);
		
		mRadius.addValueChangedListener(this);
		mCorners.addValueChangedListener(this);
		mMaxAngle.addValueChangedListener(this);
		
		if(registerAtValueManager) {
			ValueManager vm = Simulator.getInstance().getValueManager();
			vm.addValue("/SimObjects/" + mName + "/Radius", mRadius);
			vm.addValue("/SimObjects/" + mName + "/Corners", mCorners);
			vm.addValue("/SimObjects/" + mName + "/MaxAngle", mMaxAngle);
		}
		
		updateShape();
	}
	
	
	/**
	 * Calculates the shape object for the given number of corners, 
	 * the maximal angle and the current radius.
	 */
	protected void updateShape() {
		int corners = mCorners.get();
		double maxAngle = mMaxAngle.get();
		if(maxAngle < 360.0) {
			//one corener is required for the center point if the radius is less than a circle.
			//another corner is required since the first and the last point are not the same.
			corners -= 2;
		}
		double angle = Math.toRadians(maxAngle / (double) corners);
		Point2D.Double startpoint = new Point2D.Double(mRadius.get(), 0.0);
		AffineTransform transform = new AffineTransform();
		mShape = new Polygon();
		
		Point2D.Double point = new Point2D.Double(startpoint.x, startpoint.y);
		
		for(int i = 0; i <= corners; ++i) {
			mShape.addPoint((int) Math.round(point.x), (int) Math.round(point.y));
			transform.rotate(angle);
			transform.transform(startpoint, point);
		}
		if(maxAngle < 360.0) {
			mShape.addPoint(0, 0);
		}
	}
	
	/**
	 * Sets the number of corners of this CircularPolygon.
	 * 
	 * @param segments the new number of corners.
	 */
	public void setNumberOfCorners(int corners) {
		mCorners.set(corners);
		updateShape();
	}
	
	/**
	 * Returns the number of corners of this CircularPolygon.
	 * @return the number of corners.
	 */
	public int getNumberOfCorners() {
		return mCorners.get();
	}
	
	/**
	 * Sets the radius of this CircularPolygon.
	 * 
	 * @param radius the new radius.
	 */
	public void setRadius(double radius) {
		mRadius.set(radius);
		updateShape();
	}
	
	/**
	 * Returns the current radius.
	 * @return the radius.
	 */
	public double getRadius() {
		return mRadius.get();
	}
	
	/**
	 * Sets the maximal angle this circular polygon will cover. 
	 * Examples: 90 degree will lead to a quarter of a circle.
	 * 180 degree would be a half circle.
	 * 360 is a full circle.
	 * 
	 * @param angle the maximal angle.
	 */
	public void setMaximalAngle(double angle) {
		mMaxAngle.set(angle);
		updateShape();
	}
	
	/**
	 * Returns the maximal angle.
	 * @return the maximal angle.
	 */
	public double getMaximalAngle() {
		return mMaxAngle.get();
	}

	public void valueChanged(Value value) {
		super.valueChanged(value);
		if(value == null) {
			return;
		}
		if(value == mCorners || value == mMaxAngle || value == mRadius) {
			updateShape();
		}
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mRadius = (DoubleValue) ValueManager.synchronizeValue(mRadius, oldValue, newValue, this);
		mCorners = (IntValue) ValueManager.synchronizeValue(mCorners, oldValue, newValue, this);
		mMaxAngle = (DoubleValue) ValueManager.synchronizeValue(mMaxAngle, oldValue, newValue, this);
	}
	
	
	
}
