/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.devices;

import java.awt.Color;
import java.awt.Polygon;
import java.awt.geom.Point2D;
import java.util.Hashtable;

import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.value.Value;

public class ColorBeacon extends Box implements LightSource {

	private Color mCurrentColor;
	private CircularPolygon mLightArea;
	private Hashtable mColorTable = new Hashtable();
	

	public ColorBeacon(String name, double x, double y, double angle, 
			double lightRadius, int corners, double lightAngle) 
	{
		super(name, x, y, angle, 10.0, 10.0);
		
		setShapeColor(Color.red);
		
		mCurrentColor = Color.YELLOW;
		
		mLightArea = new CircularPolygon(getName() + "/LightArea", 0.0, 0.0, 0.0, lightRadius, corners, lightAngle);
		mLightArea.setShapeColor(mCurrentColor);
		mLightArea.setShapeTransparency(0.5);
		
		mLightArea.setReferenceObject(this);

	}
	
	

	public void setActive(boolean active) {
		super.setActive(active);
	}



	/**
	* Calculates the current brightness of the Beacon.
	*/
	public boolean update() {
		boolean ok = super.update();
		return ok;
	}

	public void reset() {
		super.reset();
	}

	/**
	* This method is called whenever this SimObject was added to an EnvironmentModel.
	* Subclasses of SimObject can used this method to register sub components at the
	* EnvironmentModel.
	* 
	* @param model the EnvironmentModel where the child SimObjects should be registered.
	*/
	public void addChildren(EnvironmentModel model) {
		super.addChildren(model);
		model.addSimObject(mLightArea);
	}


	/**
	* Returns the child component that indicates the bulb of the light source.
	* @return the bulb child component.
	*/
	public CircularPolygon getLightArea() {
		return mLightArea;
	}
	
	
	/**
	 * Sets the Hashtable as color table. The hashtable is NOT copied.
	 * The color table must not be null!
	 * @param colorTable
	 */
	public void setColorTable(Hashtable colorTable) {
		if(mColorTable != null) {
			mColorTable = colorTable;
		}
	}
	
	
	/**
	 * Returns the current color of the beacon.
	 * @return
	 */
	public Color getColor() {
		return mCurrentColor;
	}
	
	
	/**
	 * Sets the color of this beacon. If color is null, then the old color
	 * persists.
	 * @param color
	 */
	public void setColor(Color color) {
		if(color == null) {
			return;
		}
		mCurrentColor = color;
		mLightArea.setShapeColor(color);
	}
	

	/**
	* Returns the brightness of the LightSource at the given point.
	* @param globalPosition the position to check for brightness in global coordinates.
	* @return the brightness at the given global position.
	*/
	public double getBrightness(Point2D.Double globalPosition) {
		if(!mLightArea.isActive() || !isActive()) {
			return 0.0;
		}
		Polygon globalShape = mLightArea.getGlobalShape();
		if(globalShape.contains(globalPosition)) {
			Object obj = mColorTable.get(mCurrentColor);
			if(obj != null && (obj instanceof Double)) {
				return ((Double) obj).doubleValue();
			}
			return 0.0;
		}
		return 0.0;
	}

	/**
	* Called whenever an observed Value object is replaced by another one in the 
	* ValueManager. The object can then change its observatoin to the new Value object
	* and cancel the observation of the old Value.
	*/
	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
	}
	
}
