/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.devices;

import java.awt.Color;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.Sensor;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.SimObjectSet;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.NormalizedDoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public abstract class DistanceSensor extends Sensor {
	
	protected DoubleValue mMinScanDistance;
	protected DoubleValue mMaxScanDistance;
	protected NormalizedDoubleValue mDistance;
	protected Color mObstacleDetectedColor;
	protected boolean mTreatObstaclesAsMassive;
	
	
	//TODO consider opening angle.
	
	protected SimObjectSet mObstacles = new SimObjectSet();
	protected Vector mObstacleLists = new Vector();
	protected Vector mIgnorableObjects = new Vector();
	
	
	public DistanceSensor(String name, double x, double y, double angle,
			double minScanDistance, double maxScanDistance)
	{
		this(name, x, y, angle, minScanDistance, maxScanDistance, true);
	}
	
	public DistanceSensor(String name, double x, double y, double angle,
			double minScanDistance, double maxScanDistance, boolean treatObstaclesAsMassive) 
	{
		super(name, x, y, angle);
		
		mMinScanDistance = new DoubleValue(minScanDistance);
		mMaxScanDistance = new DoubleValue(maxScanDistance);
		mDistance = new NormalizedDoubleValue(0.0, minScanDistance, maxScanDistance, 0.0, 1.0);
		mTreatObstaclesAsMassive = treatObstaclesAsMassive;
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		vm.addValue("/SimObjects/" + name + "/MinScanDistance", mMinScanDistance);
		vm.addValue("/SimObjects/" + name + "/MaxScanDistance", mMaxScanDistance);
		vm.addValue("/SimObjects/" + name + "/Distance", mDistance);
		
		mMinScanDistance.addValueReplacedListener(this);
		mMaxScanDistance.addValueReplacedListener(this);
		mDistance.addValueReplacedListener(this);
		mMinScanDistance.addValueChangedListener(this);
		mMaxScanDistance.addValueChangedListener(this);
		
		mObstacleDetectedColor = Color.RED;
	}
	
	
	/**
	 * Returns the last measured value of this sensor.
	 * @return
	 */
	public double getSensorReading() {
		return mDistance.get();
	}
	
	/**
	 * Returns the internal NormalizedDoubleValue object that
	 * contains the distance value. 
	 * 
	 * @return
	 */
	public NormalizedDoubleValue getDistanceValue() {
		return mDistance;
	}
	
	/**
	 * Sets the maximal scan distance of this sensor.
	 * @param distance the maximal scan distance.
	 */
	public void setMaxScanDistance(double distance) {
		mMaxScanDistance.set(distance);
	}
	
	/**
	 * Returns the maximal scan distance of this sensor.
	 * @return the maximal scan distance.
	 */
	public double getMaxScanDistance() {
		return mMaxScanDistance.get();
	}
	
	
	/**
	 * Sets the minimal scan distance of this sensor.
	 * @param distance the miniman scan distance.
	 */
	public void setMinScanDistance(double distance) {
		mMinScanDistance.set(distance);
	}
	
	/**
	 * Returns the minimal scan distance of this sensor. 
	 * @return the minimal scan distance.
	 */
	public double getMinScanDistance() {
		return mMinScanDistance.get();
	}
	
	
	/**
	 * Sets the color used to indicate that an obstacle was detected.
	 * If the color object is null, then the color of the scanner
	 * is the same whether an obstacle is detected or not.
	 * 
	 * @param color the color to use when on obstacle was deteceted.
	 */
	public void setObstacleDetectedColor(Color color) {
		mObstacleDetectedColor = color;
	}
	
	/**
	 * Returns the Color object used to draw the scanner when an obstacle
	 * was detected.
	 * 
	 * @return the color of the scanner when an obstacle was detected.
	 */
	public Color getObstacleDetectedColor() {
		return mObstacleDetectedColor;
	}
	
	
	/**
	 * Adds an obstacle this distance scanner considers during its scans.
	 * 
	 * @param obstacle an obstacle that can be detected by this scanner.
	 * @return true if the obstacle object could be added.
	 */
	public boolean addObstacle(SimObject obstacle) {
		if(mObstacles.add(obstacle)) {
			addDependency(obstacle);
			return true;
		}
		return false;
	}
	
	/**
	 * Adds an external set of SimObjects that are considered as obstacles.
	 * 
	 * @param obstacles the SimObjectSet to add.
	 * @return
	 */
	public boolean addObstacleSet(SimObjectSet obstacles) {
		if(obstacles == null || mObstacleLists.contains(obstacles)) {
			return false;
		}
		mObstacleLists.add(obstacles);
		return true;
	}
	
	/**
	 * Removes a previously added SimObjectSet.
	 * @param obstacles
	 * @return
	 */
	public boolean removeObstacleSet(SimObjectSet obstacles) {
		if(obstacles == null || !mObstacleLists.contains(obstacles)) {
			return false;
		}
		mObstacleLists.remove(obstacles);
		return true;
	}
	
	public void clearObstacleSets() {
		mObstacleLists.clear();
	}
	
	
	/**
	 * Returns a copy of the obstacle list of this scanner.
	 * 
	 * @return a copy of the obstacle list.
	 */
	public Vector getObstacles() {
		return mObstacles.getElements();
	}
	
	public Vector getObstacleSets() {
		return new Vector(mObstacleLists);
	}
	
	public boolean addIgnorableObject(SimObject object) {
		if(object == null || mIgnorableObjects.contains(object)) {
			return false;
		}
		mIgnorableObjects.add(object);
		return true;
	}
	
	public boolean removeIgnorableObject(SimObject object) {
		if(object == null || !mIgnorableObjects.contains(object)) {
			return false;
		}
		mIgnorableObjects.remove(object);
		return true;
	}
	
	public void removeAllIgnorableObjects() {
		mIgnorableObjects.clear();
	}
	
	public Vector getIgnorableObjects() {
		return new Vector(mIgnorableObjects);
	}
	
	
	public void setTreatObstaclesAsMassive(boolean treatAsMassive) {
		mTreatObstaclesAsMassive = treatAsMassive;
	}
	
	public boolean isTreatingObstaclesAsMassive() {
		return mTreatObstaclesAsMassive;
	}
	
	
	
	//TODO add get for all obstacles including lists.
	
	

	/**
	 * Called whenever an observed Value object is replaced by another one in the 
	 * ValueManager. The object can then change its observation to the new Value object
	 * and cancel the observation of the old Value. When overriding this method
	 * in subclassed do not forget to call this method from the subclass.
	 */
	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mMinScanDistance = (DoubleValue) synchronizeValue(mMinScanDistance, oldValue, newValue);
		mMaxScanDistance = (DoubleValue) synchronizeValue(mMaxScanDistance, oldValue, newValue);
		mDistance = (NormalizedDoubleValue) synchronizeValue(mDistance, oldValue, newValue);
		
		mDistance.setNormalizedValueRange(mMinScanDistance.get(), mMaxScanDistance.get());
	}

	public void valueChanged(Value value) {
		super.valueChanged(value);
		if(value != null) {
			if(value == mMinScanDistance || value == mMaxScanDistance) {
				mDistance.setValueRange(mMinScanDistance.get(), mMaxScanDistance.get());
			}
		}
	}
	

}
