/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.devices;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.Sensor;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class LightSensor extends Sensor {
	
	protected DoubleValue mMinBrightness;
	protected DoubleValue mMaxBrightness;
	protected DoubleValue mSensorValue;
	protected Vector mLightSources = new Vector();
	protected boolean mMeasureMax;
	
	public LightSensor(String name, double x, double y, double angle,
			double minBrightness, double maxBrightness)
	{
		this(name, x, y, angle, minBrightness, maxBrightness, true);
	}
	
	
	public LightSensor(String name, double x, double y, double angle, 
			double minBrightness, double maxBrightness, boolean measureMax) 
	{
		super(name, x, y, angle, new Polygon());
		
		mMeasureMax = measureMax;
		
		mMinBrightness = new DoubleValue(minBrightness);
		mMaxBrightness = new DoubleValue(maxBrightness);
		mSensorValue = new DoubleValue(0.0);
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		vm.addValue("/SimObjects/" + name + "/MinBrightness", mMinBrightness);
		vm.addValue("/SimObjects/" + name + "/MaxBrightness", mMaxBrightness);
		vm.addValue("/SimObjects/" + name + "/Brightness", mSensorValue);
		
		mMinBrightness.addValueReplacedListener(this);
		mMaxBrightness.addValueReplacedListener(this);
		mSensorValue.addValueReplacedListener(this);
	}

	/**
	 * Calculates the distance to the closest obstacle.
	 * 
	 */
	public boolean update() {
		boolean ok = super.update();
		if(!isActive()) {
			if(getSensorReading() != mMaxBrightness.get()) {
				mSensorValue.set(mMaxBrightness.get());
			}
			return ok;
		}
		
	
		Point2D.Double globalPosition = getGlobalPosition();
		
		if(mMeasureMax) {
			double maxValue = mMinBrightness.get();
			for(Iterator i = mLightSources.iterator(); i.hasNext();) {
				LightSource ls = (LightSource) i.next();
				double brightness = ls.getBrightness(globalPosition);
				if(brightness > maxValue) {
					maxValue = brightness;
				}
			}
			if(maxValue > mMaxBrightness.get()) {
				maxValue = mMaxBrightness.get();
			}
			mSensorValue.set(maxValue);
		}
		else {
			double minValue = mMaxBrightness.get();
			for(Iterator i = mLightSources.iterator(); i.hasNext();) {
				LightSource ls = (LightSource) i.next();
				double brightness = ls.getBrightness(globalPosition);
				if(brightness < minValue) {
					minValue = brightness;
				}
			}
			if(minValue < mMinBrightness.get()) {
				minValue = mMinBrightness.get();
			}
			mSensorValue.set(Math.max(0, minValue)); //do not allow a brigthness less than zero.
		}
		
		return ok;
	}
	
	public void paintSelf(Graphics2D g) {
		Color c = g.getColor();
		g.setColor(getShapeColor());
		if(mMeasureMax) {
			double brightness = mSensorValue.get();
			if(brightness > mMinBrightness.get()) {
				double diff = mMaxBrightness.get() - mMinBrightness.get();
				double abs = brightness - mMinBrightness.get();
				g.fillRect(0, -2, (int) (10.0 / diff * abs), 4);
			}
		}
		else {
			double brightness = mSensorValue.get();
			if(brightness < mMaxBrightness.get()) {
				double diff = mMaxBrightness.get() - mMinBrightness.get();
				double abs = brightness - mMinBrightness.get();
				g.fillRect(0, -2, 10 - ((int) (10.0 / diff * abs)), 4);
			}
		}
		g.setColor(c);
	}
	
	//TODO
	public void reset() {
		super.reset();
	}
	
	
	/**
	 * Returns the last measured value of this sensor.
	 * @return
	 */
	public double getSensorReading() {
		return mSensorValue.get();
	}
	
	public double getMinimalBrightness() {
		return mMinBrightness.get();
	}
	
	public double getMaximalBrightness() {
		return mMaxBrightness.get();
	}
	
	
	/**
	 * Adds a LightSource to the list of detectable light sources.
	 * 
	 * @param lightSource the light source to add.
	 * @return true if successful, false if already contained.
	 */
	public boolean addLightSource(LightSource lightSource) {
		if(lightSource == null || mLightSources.contains(lightSource)) {
			return false;
		}
		mLightSources.add(lightSource);
		
		if(lightSource instanceof SimObject) {
			SimObject obj = (SimObject) lightSource;
			addDependency(obj);
		}
		return true;
	}
	
	/**
	 * Removes a light source from the list of detectable light sources.
	 * 
	 * @param lightSource the light source to remove.
	 * @return true if successful, false if not contained.
	 */
	public boolean removeLightSource(LightSource lightSource) {
		if(lightSource == null || !mLightSources.contains(lightSource)) {
			return false;
		}
		mLightSources.remove(lightSource);
		return true;
	}
	
	public void removeAllLightSources() {
		mLightSources.clear();
	}
	
	/**
	 * Called whenever an observed Value object is replaced by another one in the 
	 * ValueManager. The object can then change its observation to the new Value object
	 * and cancel the observation of the old Value. When overriding this method
	 * in subclassed do not forget to call this method from the subclass.
	 */
	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mMinBrightness = (DoubleValue) synchronizeValue(mMinBrightness, oldValue, newValue);
		mMaxBrightness = (DoubleValue) synchronizeValue(mMaxBrightness, oldValue, newValue);
		mSensorValue = (DoubleValue) synchronizeValue(mSensorValue, oldValue, newValue);
	}
	
}
