/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.evolution;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DateFormat;
import java.util.Date;
import java.util.LinkedList;

public class EvolutionLogger {
	
	private BufferedWriter mFileWriter;
	private boolean mActive;
	private int mIndentation;
	private DateFormat mDateFormatter;
	private LinkedList mClosingTags = new LinkedList();
	private String mFileName;
	
	public EvolutionLogger() {
		mIndentation = 0;
		mDateFormatter = DateFormat.getInstance();
	}
	
	public synchronized boolean startEvolutionLogger(String filename) {
		if(mFileWriter != null) {
			closeEvolutionLogger();
		}
		mIndentation = 0;
		try {
			mFileWriter = new BufferedWriter(new FileWriter(filename));
			mFileName = filename;
			mActive = true;
			addHeader();
			
		} catch (IOException e) {
			return false;
		}
		return true;
	}
	
	public boolean logXML(String xml) {
		return logXML(xml, "");
	}
	
	public synchronized boolean logXML(String xml, String closingTag) {
		
		if(mFileWriter == null) {
			return false;
		}
		String line = "";
		for(int i = 0; i < mIndentation; ++i) {
			line += "\t";
		}
		String[] lines = xml.split("\n");
		try {
			for(int i = 0; i < lines.length; ++i) {
				mFileWriter.write(line + lines[i] + "\n");
			}
			mFileWriter.flush();
		} catch (IOException e) {
			return false;
		}
		
		if(!closingTag.equals("")) {
			mIndentation++;
			mClosingTags.addFirst(closingTag);
		}
		return true;
	}
	
	public synchronized boolean closeTag() {
		if(mClosingTags.isEmpty()) {
			return false;
		}
		mIndentation--;
		logXML((String) mClosingTags.removeFirst(), "");
		return true;
	}
	
	public synchronized boolean logComment(String comment) {
		if(mFileWriter == null) {
			return false;
		}
		String line = "";
		for(int i = 0; i < mIndentation; ++i) {
			line += "\t";
		}
		String[] lines = comment.split("\n");
		try {
			mFileWriter.write(line + "<!-- ");
			for(int i = 0; i < lines.length; ++i) {
				if(i == 0) {
					mFileWriter.write(lines[i]);
				}
				else {
					mFileWriter.write(line + "\t" + lines[i]);
				}
				if(i < lines.length - 1) {
					mFileWriter.write("\n");
				}
				else {
					mFileWriter.write(" -->\n");
				}
			}
			mFileWriter.flush();
		} catch (IOException e) {
			return false;
		}
		return true;
	}
	
	public synchronized void closeEvolutionLogger() {
		if(mFileWriter != null && mActive) {
			try {
				while(!mClosingTags.isEmpty()) {
					closeTag();
				}
				
				addFooter();
				mFileWriter.close();
				mActive = false;
			} catch (IOException e) {
			}
		}
		mFileWriter = null;
	}
	
	protected void addHeader() {
		if(mFileWriter == null) {
			return;
		}
		try {
			mFileWriter.write("<!-- Logger Time: " + mDateFormatter.format(new Date(System.currentTimeMillis())) + " -->\n");
			mFileWriter.write("<evolution>\n");
			mFileWriter.flush();
			mIndentation++;
			
		} catch (IOException e) {
			return;
		}
	}
	
	protected void addFooter() {
		if(mFileWriter == null) {
			return;
		}
		try {
			mFileWriter.write("</evolution>\n");
			mFileWriter.write("<!-- End Time: " + mDateFormatter.format(new Date(System.currentTimeMillis())) + " -->\n");
			mFileWriter.flush();
			mIndentation--;
			
		} catch (IOException e) {
			return;
		}
	}
	
	public boolean isActive() {
		return mActive;
	}
	
	public String getFileName() {
		return mFileName;
	}

}
