/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.evolution;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;
import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetAgentInterface;
import de.fhg.iais.outdoor.simba.neuralnet.Neuron;
import de.fhg.iais.outdoor.simba.neuralnet.Synapse;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.NormalizedDoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class EvolutionOperatorNNAddSynapse extends EvolutionOperator {
	
	private NormalizedDoubleValue mProbabilityForInsertionPerNeuron;
	private NormalizedDoubleValue mProbabilityOfLoop;
	private IntValue mMaxNumberOfSynapsesToInsert;
	private IntValue mMaxNumberOfLoopsToInsert;
	
	public EvolutionOperatorNNAddSynapse() {
		super("NNAddSynapse", 500);
		
		mProbabilityForInsertionPerNeuron = new NormalizedDoubleValue(0.0, 0.0, 1.0, 0.0, 1.0);
		addParameter(mProbabilityForInsertionPerNeuron, "InsertionProbabilityPerNeuron");
		
		mProbabilityOfLoop = new NormalizedDoubleValue(0.0, 0.0, 1.0, 0.0, 1.0);
		addParameter(mProbabilityOfLoop, "LoopProbabilityPerNeuron");
		
		mMaxNumberOfSynapsesToInsert = new IntValue(10);
		addParameter(mMaxNumberOfSynapsesToInsert, "MaxNumberOfNewSynapses");
		
		mMaxNumberOfLoopsToInsert = new IntValue(1);
		addParameter(mMaxNumberOfLoopsToInsert, "MaxNumberOfNewLoops");
	}

	public boolean executeOperation(EvolutionIndividual individual) {
	
		NeuralNet net = null;
		try {
			net = (NeuralNet) individual.getGenom();
		}
		catch(ClassCastException e) {
		}
		if(net == null) {
			Logger.getLogger("Simulator")
			.warning("The EvolutionIndividual did not have a NeuralNet as genom.");
			return true;
		}
		
		NeuralNetAgentInterface agentInterface = Simulator.getInstance().getNeuralNetManager().getNeuralNetInterface();
		Vector inputNeurons = new Vector();
		if(agentInterface != null) {
			inputNeurons = agentInterface.getInputNeurons();
		}
		
		Vector neuronVector = net.getNeurons();
		LinkedList neurons = new LinkedList();
		//Ignore all neurons that are locked (not allowed to be source or target of a synapse).
		for(Iterator i = neuronVector.iterator(); i.hasNext();) {
			Neuron n = (Neuron) i.next();
			if(n.getProtectionMode().equals(Neuron.LOCKED)) {
				continue;
			}
			neurons.add(n);
		}
		
		//add neuron to neuron synapses.
		int maxNumberOfSypapsesToInsert = mMaxNumberOfSynapsesToInsert.get();
		for(int i = 0; i < maxNumberOfSypapsesToInsert && !neurons.isEmpty(); ++i) {
			
			if(mRandom.nextDouble() < mProbabilityForInsertionPerNeuron.get()) {
				//add synapse, find a suitable source
				
				Neuron n = (Neuron) neurons.get(mRandom.nextInt(neurons.size()));
				neurons.remove(n);
				
				//make sure that the neuron does not get more synapses as allowed.
				int maxOutgoingSynapses = n.getMaxOutgoingSynapsesValue().get();
				if(maxOutgoingSynapses >= 0 && n.getSynapses().size() >= maxOutgoingSynapses) {
					continue;
				}
				
				Vector targetCandidates = new Vector();
				
				for(Iterator j = neurons.iterator(); j.hasNext();) {
					Neuron targetCandidate = (Neuron) j.next();
					boolean isCandidate = true;
					
					//Remove all input neurons.
					if(inputNeurons.contains(targetCandidate.getActivationValue())) {
						continue;
					}
					//Remove self loops (considered separetely later).
					if(targetCandidate == n) {
						continue;
					}
					//Remove all locked and protected neurons.
					if(targetCandidate.getProtectionMode().equals(Neuron.LOCKED))
					{
						continue;
					}
					//Remove all neurons that can not accept more incomming synapses.
					int maxIncommingSynapses = targetCandidate.getMaxIncomingSynapsesValue().get();
					if(maxIncommingSynapses >= 0 && targetCandidate.getSynapses().size() >= maxIncommingSynapses) {
						continue;
					}
					Vector synapses = targetCandidate.getSynapses();
					for(Iterator k = synapses.iterator(); k.hasNext();) {
						Neuron synapseSource = ((Synapse) k.next()).getSourceNeuron();
						if(synapseSource == n) {
							isCandidate = false;
							break;
						}
						
					}
					if(isCandidate) {
						targetCandidates.add(targetCandidate);
					}
				}
				if(targetCandidates.isEmpty()) {
					continue;
				}
				int index = mRandom.nextInt(targetCandidates.size());
				Synapse synapse = new Synapse(n);
				((Neuron) targetCandidates.elementAt(index)).addSynapse(synapse);
			}
		}
		//add neuron loop synapses.
		int maxNumberOfLoopsToInsert = mMaxNumberOfLoopsToInsert.get();
		for(int i = 0; i < maxNumberOfLoopsToInsert && !neurons.isEmpty(); ++i) {
			
			if(mRandom.nextDouble() < mProbabilityOfLoop.get()) {

				Neuron n = (Neuron) neurons.get(mRandom.nextInt(neurons.size()));
				neurons.remove(n);
				
				//Ignore neurons that are protected (may not be the target of a synapse).
				if(n.getProtectionMode().equals(Neuron.LOCKED)) {
					continue;
				}
				
				//make sure that the neuron does not get more synapses as allowed.
				int maxOutgoingSynapses = n.getMaxOutgoingSynapsesValue().get();
				if(maxOutgoingSynapses >= 0 && n.getSynapses().size() >= maxOutgoingSynapses) {
					continue;
				}
				//Remove all neurons that can not accept more incomming synapses.
				int maxIncommingSynapses = n.getMaxIncomingSynapsesValue().get();
				if(maxIncommingSynapses >= 0 && n.getSynapses().size() >= maxIncommingSynapses) {
					continue;
				}
				
				if(inputNeurons.contains(n.getActivationValue())) {
					continue;
				}
				
				Vector synapses = n.getSynapses();
				
				boolean hasLoop = false;
				for(Iterator k = synapses.iterator(); k.hasNext();) {
					if(((Synapse) k.next()).getSourceNeuron() == n) {
						hasLoop = true;
						break;
					}
				}
				if(!hasLoop) {
					Synapse synapse = new Synapse(n);
					n.addSynapse(synapse);
				}
				
			}
		}
		
		return true;
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mProbabilityForInsertionPerNeuron = (NormalizedDoubleValue) ValueManager.synchronizeValue(mProbabilityForInsertionPerNeuron, oldValue, newValue, this);
		mProbabilityOfLoop = (NormalizedDoubleValue) ValueManager.synchronizeValue(mProbabilityOfLoop, oldValue, newValue, this);
		mMaxNumberOfLoopsToInsert = (IntValue) ValueManager.synchronizeValue(mMaxNumberOfLoopsToInsert, oldValue, newValue, this);
		mMaxNumberOfSynapsesToInsert = (IntValue) ValueManager.synchronizeValue(mMaxNumberOfSynapsesToInsert, oldValue, newValue, this);
	}
	
	
	
	

}
