/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.evolution;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;
import java.util.logging.Logger;


import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.neuralnet.Synapse;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.NormalizedDoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class EvolutionOperatorNNChangeStrength extends EvolutionOperator {
	
	private NormalizedDoubleValue mChangeProbability;
	private IntValue mMaxNumberOfChangedSynapses;
	private DoubleValue mMaxSynapseChange;
	private BooleanValue mChangeAllZeroSynapses;
	private DoubleValue mMaxSynapseStrength;
	
	public EvolutionOperatorNNChangeStrength() {
		super("NNChangeStrength", 500);

		mChangeProbability = new NormalizedDoubleValue(0.0, 0.0, 1.0, 0.0, 1.0);
		addParameter(mChangeProbability, "ChangeProbability");
		
		mMaxSynapseChange = new DoubleValue(0.0);
		addParameter(mMaxSynapseChange, "MaxSynapseChange");
		
		mMaxSynapseStrength = new DoubleValue(8.0);
		addParameter(mMaxSynapseStrength, "MaxSynapseStrength");
		
		mMaxNumberOfChangedSynapses = new IntValue(10);
		addParameter(mMaxNumberOfChangedSynapses, "MaxNumberOfChangedSynapses");
		
		mChangeAllZeroSynapses = new BooleanValue(true);
		addParameter(mChangeAllZeroSynapses, "ChangeAllZeroSynapses");
		
	}

	public boolean executeOperation(EvolutionIndividual individual) {

		NeuralNet net = null;
		try {
			net = (NeuralNet) individual.getGenom();
		}
		catch(ClassCastException e) {
		}
		if(net == null) {
			Logger.getLogger("Simulator")
			.warning("The EvolutionIndividual did not have a NeuralNet as genom.");
			return true;
		}
		
		Vector synapseVector = net.getSynapses();
		
		//Remove all fixed synapses from the list of change candidates. 
		Vector synapses = new Vector();
		for(Iterator i = synapseVector.iterator(); i.hasNext();) {
			Synapse s = (Synapse) i.next();
			if(!s.getProtectionMode().equals(Synapse.FIXED)) {
				synapses.add(s);
			}
		}
		
		LinkedList list = new LinkedList(synapses);
		
		int maxNumberOfChanged = mMaxNumberOfChangedSynapses.get();
		
		
		if(mChangeAllZeroSynapses.get()) {
			for(Iterator i = synapses.iterator(); i.hasNext();) {
				Synapse s = (Synapse) i.next();
				if(s.getStrength() == 0.0) {
					//change strength
					changeStrength(s);
					maxNumberOfChanged--;
				}
			}
		}
		
		for(int i = 0; i < maxNumberOfChanged && !list.isEmpty(); ++i) {
			
			if(mRandom.nextDouble() < mChangeProbability.get()) {
				int index = mRandom.nextInt(list.size());
				Synapse s = (Synapse) list.get(index);
				list.remove(index);
				changeStrength(s);
			}
			
		}

		return true;
	}
	
	protected void changeStrength(Synapse synapse) {
		double maxStrength = mMaxSynapseStrength.get();
		
		double change = mMaxSynapseChange.get() * mRandom.nextDouble();
		if(mRandom.nextBoolean()) {
			change *= -1;
		}
		change += synapse.getStrength();
		if(change > maxStrength) {
			change = maxStrength;
		}
		else if(change < -maxStrength) {
			change = -maxStrength;
		}
		synapse.setStrength(change);
		
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mChangeProbability = (NormalizedDoubleValue) ValueManager.synchronizeValue(mChangeProbability, oldValue, newValue, this);
		mMaxNumberOfChangedSynapses = (IntValue) ValueManager.synchronizeValue(mMaxNumberOfChangedSynapses, oldValue, newValue, this);
		mChangeAllZeroSynapses = (BooleanValue) ValueManager.synchronizeValue(mChangeAllZeroSynapses, oldValue, newValue, this);
		mMaxSynapseChange = (DoubleValue) ValueManager.synchronizeValue(mMaxSynapseChange, oldValue, newValue, this);
		mMaxSynapseStrength = (DoubleValue) ValueManager.synchronizeValue(mMaxSynapseStrength, oldValue, newValue, this);
		
	}

}
