/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.evolution;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;
import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetAgentInterface;
import de.fhg.iais.outdoor.simba.neuralnet.Neuron;
import de.fhg.iais.outdoor.simba.neuralnet.Synapse;
import de.fhg.iais.outdoor.simba.neuralnet.TransferFunction;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.NormalizedDoubleValue;

public class EvolutionOperatorNNChangeTransferFunction extends
		EvolutionOperator 
{

	private NormalizedDoubleValue mProbabilityOfStandardTransferFunction;
	private NormalizedDoubleValue mProbabilityOfChange;
	private IntValue mMaxNumberOfChangedNeurons;
	private BooleanValue mExcludeOutputNeurons;
	
	public EvolutionOperatorNNChangeTransferFunction() {
		super("NNChangeTransferFunction", 2);
		
		mProbabilityOfChange = new NormalizedDoubleValue(0.0, 0.0, 1.0, 0.0, 1.0);
		mProbabilityOfStandardTransferFunction = new NormalizedDoubleValue(1.0, 0.0, 1.0, 0.0, 1.0);
		mMaxNumberOfChangedNeurons = new IntValue(1);
		mExcludeOutputNeurons = new BooleanValue(true);
		
		addParameter(mProbabilityOfChange, "ProbabilityPerNeuron");
		addParameter(mProbabilityOfStandardTransferFunction, "ProbabilityOfStandard");
		addParameter(mMaxNumberOfChangedNeurons, "MaxNumberOfChangedNeurons");
		addParameter(mExcludeOutputNeurons, "ExcludeOutputNeurons");
	}

	public boolean executeOperation(EvolutionIndividual individual) {
		
		NeuralNet net = null;
		try {
			net = (NeuralNet) individual.getGenom();
		}
		catch(ClassCastException e) {
		}
		if(net == null) {
			Logger.getLogger("Simulator")
			.warning("The EvolutionIndividual did not have a NeuralNet as genom.");
			return true;
		}

		Vector neuronVector = net.getNeurons();
		
		NeuralNetAgentInterface agentInterface = Simulator.getInstance().getNeuralNetManager().getNeuralNetInterface();
		Vector iNeurons = new Vector();
		Vector oNeurons = new Vector();
		if(agentInterface != null) {
			iNeurons = agentInterface.getInputNeurons();
			oNeurons = agentInterface.getOutputNeurons();
		}
		
		//Remove all fixed neurons and all input neurons from the list of change candidates. 
		Vector neurons = new Vector();
		boolean excludeOutputNeurons = mExcludeOutputNeurons.get();
		for(Iterator i = neuronVector.iterator(); i.hasNext();) {
			Neuron n = (Neuron) i.next();
			if(!n.getProtectionMode().equals(Synapse.FIXED) && !iNeurons.contains(n.getActivationValue())) {
				if(!excludeOutputNeurons || (!oNeurons.contains(n.getActivationValue()))) {
					neurons.add(n);
				}
			}
		}
		
		LinkedList list = new LinkedList(neurons);
		
		int maxNumberOfChanged = mMaxNumberOfChangedNeurons.get();
		Vector transferFunctions = Simulator.getInstance().getNeuralNetManager().getTransferFunctionPrototypes();
		TransferFunction mDefaultFunction = net.getDefaultTransferFunction();
		transferFunctions.remove(mDefaultFunction);
		
		for(int i = 0; i < maxNumberOfChanged && !list.isEmpty(); ++i) {
			
			if(mRandom.nextDouble() < mProbabilityOfChange.get()) {
				int index = mRandom.nextInt(list.size());
				Neuron n = (Neuron) list.get(index);
				list.remove(index);

				if(n.getTransferFunction() == mDefaultFunction) {
					TransferFunction tf = (TransferFunction) 
							transferFunctions.elementAt(mRandom.nextInt(transferFunctions.size()));
					n.setTransferFunction(tf);
				}
				else {
					if(mRandom.nextDouble() <= mProbabilityOfStandardTransferFunction.get()) {
						n.setTransferFunction(mDefaultFunction);
					}
					else {
						TransferFunction tf = (TransferFunction) 
						transferFunctions.elementAt(mRandom.nextInt(transferFunctions.size()));
						n.setTransferFunction(tf);
					}
				}
			}
			
		}
		
		
		return true;
	}
	
	
}
