/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.evolution;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.NormalizedDoubleValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class NiecheRouletteWheelSelection extends NeuralNetSelectionMethod {

	private NormalizedDoubleValue mPercentageToConsider;
	private IntValue mNumberOfNieches;

	public NiecheRouletteWheelSelection() {
		super("NiecheRouletteWheelSelection");
		
		mPercentageToConsider = new NormalizedDoubleValue(50.0, 0.0, 100.0, 0.0, 100.0);
		addParameter(mPercentageToConsider, "PercentageToConsider");

		mNumberOfNieches = new IntValue(4);
		addParameter(mNumberOfNieches, "NumberOfNieches");
	}
	
	
	/**
	 * Returns a Vector with the specified number of children stubs.
	 * Children stubs are EvolutionIndividuals that do not contain a genom, 
	 * but at least one parent individual that may serve as a blue print
	 * for the new genom.
	 */
	public Vector createChildrenStubs(Vector parents, int numberOfChildren) {
		
		int fitnessFunctionIndex = getFitnessFunctionIndexValue().get();
		
		Vector children = new Vector();

		double fitnessSum = 0.0;
		Vector fitnessValues = new Vector();
		
		LinkedList fitnessSortedIndividuals = new LinkedList();
		
		//fill the Vector with the best individuals and calculate the fitness sum.
		for(Iterator i = parents.iterator(); i.hasNext();) {
			EvolutionIndividual ind = (EvolutionIndividual) i.next();
			if(ind.getFitnessValues().size() > fitnessFunctionIndex) {
				double fitness = Math.min(0.0, ((Double) ind.getFitnessValues().elementAt(fitnessFunctionIndex)).doubleValue());
				fitnessSum += fitness;
				fitnessValues.add(new Double(fitness));
				
				//add to sorted vector
				boolean added = false;
				for(int j = 0; j < fitnessSortedIndividuals.size(); ++j) {
					if(fitness > ((Double) ((EvolutionIndividual) fitnessSortedIndividuals.get(j))
							.getFitnessValues().elementAt(fitnessFunctionIndex))
							.doubleValue()) 
					{
						fitnessSortedIndividuals.add(j, ind);
						added = true;
						break;
					}
				}
				if(!added) {
					fitnessSortedIndividuals.add(ind);
				}
				
			}
			else {
				fitnessValues.add(new Double(0.0));
			}
		}
		

		//Calculate the percentage of fitness for each individual.
		Vector percentage = new Vector();
		
		for(Iterator i = fitnessValues.iterator(); i.hasNext();) {
			Double fitness = (Double) i.next();
			if(fitnessSum == 0.0) {
				percentage.add(new Double(1.0 / ((double) parents.size())));
			}
			else {
				percentage.add(new Double(fitness.doubleValue() / fitnessSum));
			}
		}
		

		int individualCount = 0;
		//Randomly select children  (roulette wheel)
		//The chance to be selected is equal to the percentage of the own fitness compared to the sum of the fitness.
		Random rand = new Random();
		for(int i = 0; i < numberOfChildren; ++i) {
			double prob = rand.nextDouble();
			for(int j = 0; j < percentage.size(); ++j) {
				prob -= ((Double) percentage.elementAt(j)).doubleValue();
				if(prob <= 0.0) {
					EvolutionIndividual ind = new EvolutionIndividual(null);
					EvolutionIndividual parent = (EvolutionIndividual) parents.elementAt(j);
					parent.setNumberOfChildren(parent.getNumberOfChildren() + 1);
					ind.addParent(parent); 
					ind.addProperties(parent.getProperties());
					children.add(ind);
					individualCount++;
					break;
				}
			}
		}
		
		return children;
	}

	/**
	 * Returns a Vector with the new parents, sorted by importance.
	 */
	public Vector selectParents(Vector individuals) {
		Vector parents = new Vector();
		int fitnessFunctionIndex = getFitnessFunctionIndexValue().get();
		
		double fitnessSum = 0.0;
		for(Iterator i = individuals.iterator(); i.hasNext();) {
			EvolutionIndividual ind = (EvolutionIndividual) i.next();
			if(ind.getFitnessValues().size() > fitnessFunctionIndex) {
				fitnessSum += Math.max(0.0, ((Double) ind.getFitnessValues().elementAt(fitnessFunctionIndex)).doubleValue());
			}
		}
		
		double fitnessToConsider = fitnessSum / 100.0 * (mPercentageToConsider.get());
		
		if(fitnessToConsider == 0.0) {
			for(int i = 0; i < Math.min(individuals.size(), Math.max(1, (int) (individuals.size() / 100.0 * mPercentageToConsider.get()))); ++i) {
				parents.add(individuals.elementAt(i));
			}
		}
		else {
			LinkedList list = new LinkedList(individuals);
			while(fitnessToConsider > 0.0) {
				EvolutionIndividual best = null;
				double bestFitness = 0.0;
				for(Iterator i = list.iterator(); i.hasNext();) {
					EvolutionIndividual ind = (EvolutionIndividual) i.next();
					if(ind.getFitnessValues().size() > fitnessFunctionIndex) {
						double fitness = ((Double) ind.getFitnessValues().elementAt(fitnessFunctionIndex)).doubleValue();
						if(fitness > bestFitness) {
							bestFitness = fitness;
							best = ind;
						}
					}
				}
				if(best == null) {
					break;
				}
				parents.add(best);
				fitnessToConsider -= bestFitness;
				list.remove(best);
			}
		}
		
		//Sort the parents by their fitness
		parents = sortIndividualsByFitness(parents);
		return parents;
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mPercentageToConsider = (NormalizedDoubleValue) ValueManager.synchronizeValue(mPercentageToConsider, oldValue, newValue, this);
	}

}
