/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.fitness;

import de.fhg.iais.outdoor.simba.core.GenerationListener;
import de.fhg.iais.outdoor.simba.core.IndividualListener;
import de.fhg.iais.outdoor.simba.core.ParameterElement;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.core.SystemObject;
import de.fhg.iais.outdoor.simba.statistic.StatisticCalculator;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.HistoryValue;
import de.fhg.iais.outdoor.simba.value.ObjectValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public abstract class FitnessFunction extends ParameterElement
			implements SystemObject, ValueChangedListener, GenerationListener, IndividualListener
{
	protected DoubleValue mFitness;
	protected DoubleValue mMeanFitness;
	protected HistoryValue mGenerationMeanFitness;
	protected HistoryValue mGenerationBestFitness;
	protected HistoryValue mGenerationLowestFitness;
	protected int mNumberOfIndividuals;
	protected double mMeanFitnessTillLastTry;
	private int mCurrentTry;
	private BooleanValue mStaseActive;
	private boolean mStaseIsActive;
	private FitnessStatisticCalculator mGenerationStatisticsCalculator;
	
	/**
	 * Constructor.
	 * @param name the name to identify this FitnessFunction in the grafical user interface.
	 */
	public FitnessFunction(String name) {
		super(name, "/FitnessFunctions/" + name + "/Config/");
		
		Simulator.getInstance().addSystemObject(this);
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		ObjectValue value = new ObjectValue(this);
		int counter = 0;
		boolean ok = false;
		
		String fullName = name;
		
		while(!ok) {
			if(!vm.addValue("/FitnessFunctions/" + fullName, value)) {
				fullName =  name + ++counter;
			}
			else {
				ok = true;
			}
		}
		
		mFitness = new DoubleValue(0.0);
		mFitness.addValueReplacedListener(this);
		vm.addValue("/FitnessFunctions/" + fullName + "/Fitness", mFitness);
		
		mMeanFitness = new DoubleValue(0.0);
		mMeanFitness.addValueReplacedListener(this);
		vm.addValue("/FitnessFunctions/" + fullName + "/MeanFitness", mMeanFitness);
		
		mGenerationBestFitness = new HistoryValue();
		mGenerationBestFitness.addValueReplacedListener(this);
		vm.addValue("/FitnessFunctions/" + fullName + "/GenerationBestFitness", mGenerationBestFitness);
		
		
		mGenerationMeanFitness = new HistoryValue();
		mGenerationMeanFitness.addValueReplacedListener(this);
		vm.addValue("/FitnessFunctions/" + fullName + "/GenerationMeanFitness", mGenerationMeanFitness);
		
		mGenerationLowestFitness = new HistoryValue();
		mGenerationLowestFitness.addValueReplacedListener(this);
		vm.addValue("/FitnessFunctions/" + fullName + "/GenerationLowestFitness", mGenerationLowestFitness);
		
		mName = fullName;
	
		mGenerationStatisticsCalculator = new FitnessStatisticCalculator(this);
		Simulator.getInstance().getStatisticManager().addCalculator(mGenerationStatisticsCalculator);
		
		Simulator.getInstance().getScheduler().addIndividualListener(this);
		Simulator.getInstance().getScheduler().addGenerationListener(this);
	}
	

	/** 
	 * Initializes the fitness function during startup.
	 * This method can be overridden by subclasses. e.g. to acquire Value objects
	 * from the ValueManager.
	 */
	public boolean init() {
		mFitness.set(0.0);
		mMeanFitness.set(0.0);
		
		mStaseActive = Simulator.getInstance().getEvolutionManager().getStaseModeValue();
		mStaseActive.addValueChangedListener(this);
		
		return true;
	}

	/**
	 * Called during system clean up.
	 */
	public boolean cleanUp() {
		return true;
	}


	/**
	 * Caluclates the current fitness of the current try up to the current step;
	 * @return TODO
	 *
	 */
	public abstract double calculateFitness();
	
	
	/**
	 * Resets the fitness function between two tries.
	 * This method usually is overridden by subclasses to reset internal variables.
	 */
	public abstract void reset() ;
	
	
	/**
	 * Returns the fitness up to the current time step of the current try.
	 * @return the current fitness.
	 */
	public double getFitness() {
		return mFitness.get();
	}
	
	
	/**
	 * Returns the Value that holds the fitness.
	 * @return
	 */
	public DoubleValue getFitnessValue() {
		return mFitness;
	}
	
	/**
	 * Returns the mean fitness of all tries of the current behavior.
	 * @return
	 */
	public double getMeanFitness() {
		return mMeanFitness.get();
	}
	
	/**
	 * Returns the Value object that holds the mean fitness.
	 * @return
	 */
	public DoubleValue getMeanFitnessValue() {
		return mMeanFitness;
	}
	
	
	public DoubleValue getGenerationMeanFitnessValue() {
		return mGenerationMeanFitness;
	}
	
	
	public DoubleValue getGenerationBestFitnessValue() {
		return mGenerationBestFitness;
	}
	
	public DoubleValue getGenerationLowestFitnessValue() {
		return mGenerationLowestFitness;
	}
	
	
	/**
	 * Called by the FitnessManager to give the FitnessFunction a chance to
	 * reset values and to update the mean value between successive tries.
	 * 
	 * @param tryId the current try.
	 */
	public void prepareNextTry(int tryId) { //starting with 0
		if(tryId == 0) {
			//reset mean fitness
			mMeanFitness.set(0.0);
			mCurrentTry = 0;
			mMeanFitnessTillLastTry = 0.0;
		}
		else {
			double rawFitness = mMeanFitnessTillLastTry * ((double) mCurrentTry);
			mCurrentTry = tryId;
			mMeanFitnessTillLastTry = (rawFitness + mFitness.get()) / ((double) mCurrentTry);
		}
		mFitness.set(0.0);
		reset();
	}
	
	/**
	 * Called by the FitnessManager after each time step if this FitnessFunction
	 * is active. This method calls the customized fitness calculation method updateFitness
	 * and subsequently updates the internal fitness mean. The mean value is the arithmetic
	 * mean of the sum of the end fitness values of each try plus the current fitness up to 
	 * the current time step of the current try.
	 *
	 */
	public void update() {
		mFitness.set(calculateFitness());
		if(mCurrentTry == 0) {
			mMeanFitness.set(mFitness.get());
		}
		else {
			mMeanFitness.set(((mMeanFitnessTillLastTry * ((double) mCurrentTry)) + mFitness.get()) / ((double) (mCurrentTry + 1)));
		}
	}
	
	
	/**
	 * Sets the fitness for the current try up to the current time step.
	 * Usually this method is called from customized versions of updateFitness()
	 * to set the internal fitness value.
	 * 
	 * @param fitness the current fitness.
	 */
	public void setFitness(double fitness) {
		mFitness.set(fitness);
	}
	
	

	/**
	 * Called if one of the internal Value objects was replaced in the ValueManager.
	 * 
	 */
	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mFitness = (DoubleValue) ValueManager.synchronizeValue(mFitness, oldValue, newValue, this);
		mMeanFitness = (DoubleValue) ValueManager.synchronizeValue(mMeanFitness, oldValue, newValue, this);
		mStaseActive = (BooleanValue) ValueManager.synchronizeValue(mStaseActive, oldValue, newValue, this);
		mGenerationBestFitness  = (HistoryValue) ValueManager.synchronizeValue(mGenerationBestFitness, oldValue, newValue, this);
		mGenerationMeanFitness  = (HistoryValue) ValueManager.synchronizeValue(mGenerationMeanFitness, oldValue, newValue, this);
		mGenerationLowestFitness  = (HistoryValue) ValueManager.synchronizeValue(mGenerationLowestFitness, oldValue, newValue, this);
	}


	public String createParameterLog() {
		StringBuffer entry = new StringBuffer();
		entry.append("<fitnessFunction name=\"").append(getName());
		
		String parameterLog = super.createParameterLog();

		if(parameterLog.equals("")) {
			entry.append("\"/>\n");
		}
		else {
			entry.append("\">\n");
			entry.append(parameterLog);
			entry.append("</fitnessFunction>");
		}
		return entry.toString();
	}


	public void generationCompleted(int generationId) {
	}


	public void generationStarted(int generationId) {
		mNumberOfIndividuals = 0;
		mGenerationBestFitness.set(0.0);
		mGenerationMeanFitness.set(0.0);
		mGenerationLowestFitness.set(0.0);
	}
	


	public void valueChanged(Value value) {
		if(value == null) {
			return;
		}
		if(value == mStaseActive) {
			if(mStaseActive.get()) {
				mStaseIsActive = true;
			}
		}
	}
	


	public void individualCompleted(int individualId) {
		
		if(mStaseActive != null && !mStaseActive.get() && !mStaseIsActive) {
			if(individualId == 0) {
				mGenerationBestFitness.set(getMeanFitness());
				mGenerationLowestFitness.set(getMeanFitness());
				mGenerationMeanFitness.set(getMeanFitness());
				mNumberOfIndividuals++;
			}
			else {
				if(getMeanFitness() > mGenerationBestFitness.get()) {
					mGenerationBestFitness.set(getMeanFitness());
				}
				if(getMeanFitness() < mGenerationLowestFitness.get()) {
					mGenerationLowestFitness.set(getMeanFitness());
				}
				double meanFitness = mGenerationMeanFitness.get() * ((double) mNumberOfIndividuals);
				meanFitness += getMeanFitness();
				mNumberOfIndividuals++;
				mGenerationMeanFitness.set(meanFitness / ((double) mNumberOfIndividuals));
			}
			
		}
		if(mStaseActive != null && !mStaseActive.get()) {
			mStaseIsActive = false;
		}
	}


	public void individualStarted(int individualId) {
	}
	
	
	protected class FitnessStatisticCalculator extends StatisticCalculator {
		
		public FitnessStatisticCalculator(FitnessFunction owner) {
			super(owner.getName());
			addParameter(mGenerationBestFitness, "GenerationBestFitness");
			addParameter(mGenerationLowestFitness, "GenerationLowestFitness");
			addParameter(mGenerationMeanFitness, "GenerationMeanFitness");
		}

		public void calculate() {
		}
	}
	
	
	
}
