/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.fitnessfunctions;

import java.awt.geom.Point2D;
import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.fitness.FitnessFunction;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.NormalizedDoubleValue;
import de.fhg.iais.outdoor.simba.value.PointValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class ForwardMoverFitness extends FitnessFunction {

	private NormalizedDoubleValue mLeftVel;
	private NormalizedDoubleValue mRightVel;
	private PointValue mRobotPosition;
	private IntValue mMaxNumberOfStepsWithoutIncrease;
	private double mAccu;
	private int mIncreaseCounter;
	private Point2D.Double mLastPosition;
	
	public ForwardMoverFitness() {
		super("ForwardMoverFitness");
		mMaxNumberOfStepsWithoutIncrease = new IntValue(100);
		
		addParameter(mMaxNumberOfStepsWithoutIncrease, "MaxStepsWithoutIncrease");
	}

	
	public boolean init() {
		boolean ok = super.init();
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		DoubleValue lvel = vm.getDoubleValue("/SimObjects/Robot/VelocityLeft");
		DoubleValue rvel = vm.getDoubleValue("/SimObjects/Robot/VelocityRight");
		mRobotPosition = vm.getPointValue("/SimObjects/Robot/LocalPosition");
		
		if(lvel != null && rvel != null 
			&& lvel instanceof NormalizedDoubleValue
			&& rvel instanceof NormalizedDoubleValue
			&& mRobotPosition != null) 
		{
			mLeftVel = (NormalizedDoubleValue) lvel;
			mRightVel = (NormalizedDoubleValue) rvel;
			
			mLeftVel.addValueReplacedListener(this);
			mRightVel.addValueReplacedListener(this);
			mRobotPosition.addValueReplacedListener(this);
		}
		else {
			ok = false;
			Logger.getLogger("Simulator").warning("Could not find the motor neurons");
		}
		return ok;
	}




	public double calculateFitness() {
		
		//double diff = mLastPosition.distance(mRobotPosition.get()) * Math.signum(mLeftVel.get() + mRightVel.get());
		double diff = mLeftVel.get() + mRightVel.get() - Math.abs(mLeftVel.get() - mRightVel.get());
		mLastPosition.x = mRobotPosition.get().x;
		mLastPosition.y = mRobotPosition.get().y;
		mAccu += (diff / 100.0);
		if(diff == 0.0) {
			mIncreaseCounter++;
		}
		else {
			mIncreaseCounter = 0;
		}
		if(mIncreaseCounter > mMaxNumberOfStepsWithoutIncrease.get()) {
			Simulator.getInstance().getScheduler().terminateCurrentTry();
		}
		return mAccu;
	}

	public void reset() {
		mAccu = 0.0;
		mIncreaseCounter = 0;
		mLastPosition = new Point2D.Double(mRobotPosition.get().x, mRobotPosition.get().y);
	}


	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
		mRobotPosition = (PointValue) ValueManager.synchronizeValue(mRobotPosition, oldValue, newValue, this);
		mLeftVel = (NormalizedDoubleValue) ValueManager.synchronizeValue(mLeftVel, oldValue, newValue, this);
		mRightVel = (NormalizedDoubleValue) ValueManager.synchronizeValue(mRightVel, oldValue, newValue, this);
	}
	
	

}
