/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.fitnessfunctions;

import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.fitness.FitnessFunction;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.PointValue;
import de.fhg.iais.outdoor.simba.value.SimObjectValue;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class PathMemorizerFitness3 extends FitnessFunction {

	private SimObject mRightBound;
	private SimObject mLeftBound;
	private SimObject mRobot;
	private PointValue mStartAreaPosition;
	private BooleanValue mMovingLeft;
	private double mLastXPosition;
	private double mAccumulator;
	private DoubleValue mLeftThreshold;
	private DoubleValue mRightThreshold;
	private DoubleValue mGainToPower;
	private DoubleValue mGainPerDist;
	private BooleanValue mUseLeftGainForRightMovement;
	private BooleanValue mUseTurnFactor;
	private DoubleValue mLastTurningPoint;
	private DoubleValue mRatio;
	private DoubleValue mTurnPositionWeight;
	private DoubleValue mHitStartPositionWeight;
	
	
	public PathMemorizerFitness3() {
		super("PathMemorizerFitness3");
		
		mLeftThreshold = new DoubleValue(30.0);
		mRightThreshold = new DoubleValue(30.0);
		mGainToPower = new DoubleValue(1.0);
		mUseLeftGainForRightMovement = new BooleanValue(false);
		mUseTurnFactor = new BooleanValue(false);
		mTurnPositionWeight = new DoubleValue(0.0);
		mHitStartPositionWeight = new DoubleValue(1.0);
		
		addParameter(mLeftThreshold, "LeftThreshold");
		addParameter(mRightThreshold, "RightThreshold");
		addParameter(mGainToPower, "GainToPower");
		addParameter(mUseLeftGainForRightMovement, "UseLeftGainForRightMovement");
		addParameter(mUseTurnFactor, "UseTurnFactor");
		addParameter(mTurnPositionWeight, "TurnPositionWeight");
		addParameter(mHitStartPositionWeight, "HitStartPosWeight");
		
		mGainPerDist = new DoubleValue(1.0);
		Simulator.getInstance().getValueManager().addValue("/FitnessFunctions/" + getName() + "/Status/Gain", mGainPerDist);
		
		mMovingLeft = new BooleanValue(true);
		Simulator.getInstance().getValueManager().addValue("/FitnessFunctions/" + getName() + "/Status/MovingLeft", mMovingLeft);
		
		mLastTurningPoint = new DoubleValue(0.0);
		Simulator.getInstance().getValueManager().addValue("/FitnessFunctions/" + getName() + "/Status/LastTurningPoint", mLastTurningPoint);
		
		mRatio = new DoubleValue(1.0);
		Simulator.getInstance().getValueManager().addValue("/FitnessFunctions/" + getName() + "/Status/Ratio", mRatio);
	}
	
	

	public boolean init() {
		boolean ok =  super.init();
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		
		SimObjectValue v = vm.getSimObjectValue("/SimObjects/StartingArea");
		if(v != null) {
			mRightBound = v.get();
		}
		v = vm.getSimObjectValue("/SimObjects/TargetArea");
		if(v != null) {
			mLeftBound = v.get();
		}
		v = vm.getSimObjectValue("/SimObjects/Robot");
		if(v != null) {
			mRobot = v.get();
		}
		mStartAreaPosition = vm.getPointValue("/SimObjects/StartingArea/LocalPosition");
		
		if(mRightBound == null || mLeftBound == null || mRobot == null || mStartAreaPosition == null) {
			Logger.getLogger("Simulator").warning("Could not find the required SimObjects");
			return false;
		}
		
		return ok;
	}



	public double calculateFitness() {

		if(mLastXPosition == Double.MAX_VALUE) {
			mLastXPosition = mRobot.getGlobalPosition().x;
		}
		
		double diff = 0;
		if(mMovingLeft.get()) {
			diff = mLastXPosition - mRobot.getGlobalPosition().x;
			//check if the robot is still moving left
			if(diff < 0.0 && 
				mRobot.getGlobalPosition().x <= mLeftBound.getGlobalPosition().x + mLeftThreshold.get()) 
			{
				mMovingLeft.set(false);
				if(!mUseLeftGainForRightMovement.get()) {
					mGainPerDist.set(0.5);
				}
			}
		}
		else {
			diff = mRobot.getGlobalPosition().x - mLastXPosition;
			if(diff < 0.0
					&& mRobot.getGlobalPosition().x >= mRightBound.getGlobalPosition().x - mRightThreshold.get()
					&& mRobot.getGlobalPosition().x <= mRightBound.getGlobalPosition().x + mRightThreshold.get()) 
			{
				double distToRightArea = Math.abs(mRobot.getGlobalPosition().x - mRightBound.getGlobalPosition().x);
				double distToLastTurnPosition = Math.abs(mRobot.getGlobalPosition().x - mLastTurningPoint.get());
				double threshold = mRightThreshold.get();
				if(threshold == 0) {
					threshold = 1;
				}
				if(mLastTurningPoint.get() == Double.MAX_VALUE) {
					distToLastTurnPosition = 100.0;
				}
				else if(distToLastTurnPosition > 100.0) {
					distToLastTurnPosition = 100.0;
				}
				
				mGainPerDist.set(((1 - mHitStartPositionWeight.get())
							 	+ (mHitStartPositionWeight.get() 
							 			* ((threshold - distToRightArea) / threshold)))
							 * ((1 - mTurnPositionWeight.get()) 
								+  (mTurnPositionWeight.get() 
										* ((distToLastTurnPosition / 100.0)))));
				mMovingLeft.set(true);
				mLastTurningPoint.set(mRobot.getGlobalPosition().x);
			}
		}
		if(diff > 0.0 && mRobot.getGlobalPosition().x > mRightBound.getGlobalPosition().x) {
			diff *= -1;
		}
//		if(mRobot.getGlobalPosition().x > mStartAreaPosition.getX() + mRightThreshold.get()) {
//			diff = 0.0;
//		}
		if(diff != 0.0) {
			double d = Math.signum(diff) * Math.pow((Math.abs(diff) * mGainPerDist.get()), mGainToPower.get());
			mAccumulator += d;
		}
		
		mLastXPosition = mRobot.getGlobalPosition().x;
		
		return Math.max(0.0, mAccumulator);
	}

	public void reset() {
		mMovingLeft.set(true);
		mLastXPosition = Double.MAX_VALUE;
		mAccumulator = 0.0;
		mGainPerDist.set(0.5);
		mLastTurningPoint.set(Double.MAX_VALUE);
		mRatio.set(1.0);
	}

}
