/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.neuralnet;

import java.util.Iterator;
import java.util.Vector;



public class NeuralNetAgentInterface {

	private Vector mInputNeurons = new Vector();
	private Vector mOutputNeurons = new Vector();
	private Vector mInfoNeurons = new Vector();
	private String mName;

	/**
	 * Constructo.
	 * @param name the name of this NeuralNetworkController.
	 */
	public NeuralNetAgentInterface(String name) {
		mName = name;
	}
	
	/**
	 * Adds an input value that will normalized double values to neuron
	 * ranges from 0.0 to 1.0 or -1.0 to 1.0.
	 * 
	 * @param value the value to add.
	 * @return true if successful. If the value is already registered, then false is returend.
	 */
	public boolean addInputNeuron(NeuronWrapper value) {
		if(value == null || mInputNeurons.contains(value)) {
			return false;
		}
		mInputNeurons.add(value);
		return true;
	}
	
	/**
	 * Adds an output value that will normalized double values to neuron
	 * ranges from 0.0 to 1.0 or -1.0 to 1.0.
	 * 
	 * @param value the value to add.
	 * @return true if successful. If the value is already registered, then false is returend.
	 */
	public boolean addOutputNeuron(NeuronWrapper value) {
		if(value == null || mOutputNeurons.contains(value)) {
			return false;
		}
		mOutputNeurons.add(value);
		return true;
	}
	
	/**
	 * Adds an info value that will normalized double values to neuron
	 * ranges from 0.0 to 1.0 or -1.0 to 1.0.
	 * Info neurons are used in the fitness function and should be ignored by
	 * the real neural net.
	 * 
	 * @param value the value to add.
	 * @return true if successful. If the value is already registered, then false is returend.
	 */
	public boolean addInfoNeuron(NeuronWrapper value) {
		if(value == null || mInfoNeurons.contains(value)) {
			return false;
		}
		mInfoNeurons.add(value);
		return true;
	}
	
	/**
	 * Returns a Vector with all registered input neurons.
	 * @return all input neurons.
	 */
	public Vector getInputNeurons() {
		return mInputNeurons;
	}
	
	/**
	 * Returns a Vector with all regsitered output values.
	 * @return all output values.
	 */
	public Vector getOutputNeurons() {
		return mOutputNeurons;
	}

	/**
	 * Returns a Vector with all registered info neurons.
	 * @return
	 */
	public Vector getInfoNeurons() {
		return mInfoNeurons;
	}
	
	/**
	 * Returns the name of this NeuralNetController.
	 * @return the name.
	 */
	public String getName() {
		return mName;
	}
	
	/**
	 * Sets the name of this NeuralNetController.
	 * @param name
	 */
	public void setName(String name) {
		mName = name;
	}
	
	/**
	 * Returns the first neuron wrapper that matches the name, starting the search with the 
	 * input neurons, followed by the output neurons and finally the info neurons.
	 * @param name the name of the requested neuron wrapper.
	 * @return the requested neuron wrapper or null if not found.
	 */
	public NeuronWrapper getNeuron(String name) {
		for(Iterator i = mInputNeurons.iterator(); i.hasNext();) {
			NeuronWrapper neuron = ((NeuronWrapper) i.next());
			if(neuron.getName().equals(name)) {
				return neuron;
			}
		}
		for(Iterator i = mOutputNeurons.iterator(); i.hasNext();) {
			NeuronWrapper neuron = ((NeuronWrapper) i.next());
			if(neuron.getName().equals(name)) {
				return neuron;
			}
		}
		for(Iterator i = mInfoNeurons.iterator(); i.hasNext();) {
			NeuronWrapper neuron = ((NeuronWrapper) i.next());
			if(neuron.getName().equals(name)) {
				return neuron;
			}
		}
		return null;
	}
	
	
	public void setNeuronRange(double min, double max) {
		for(Iterator i = mInputNeurons.iterator(); i.hasNext();) {
			NeuronWrapper neuron = ((NeuronWrapper) i.next());
			neuron.setNormalizedValueRange(min, max);
		}
		for(Iterator i = mOutputNeurons.iterator(); i.hasNext();) {
			NeuronWrapper neuron = ((NeuronWrapper) i.next());
			neuron.setNormalizedValueRange(min, max);
		}
		for(Iterator i = mInfoNeurons.iterator(); i.hasNext();) {
			NeuronWrapper neuron = ((NeuronWrapper) i.next());
			neuron.setNormalizedValueRange(min, max);
		}
	}

}
