/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.neuralnet;

import java.util.Iterator;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.GenerationListener;
import de.fhg.iais.outdoor.simba.core.IndividualListener;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.core.StepListener;
import de.fhg.iais.outdoor.simba.core.SystemObject;
import de.fhg.iais.outdoor.simba.core.TryListener;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.value.ValueReplacedListener;

public class NeuralNetManager implements SystemObject, StepListener, TryListener, 
		IndividualListener, ValueReplacedListener, GenerationListener {

	private NeuralNet mCurrentNet;
	private Vector mNeuralNetListeners = new Vector();
	private Vector mAvailableTransferFunctions = new Vector();
	private NeuralNetAgentInterface mNeuralNetInterface;
	private Vector mListenerBufferVector = new Vector();
	private IntValue mIterationsPerStep;
	private int mIterationsPerStepBuffer;
	
	//private NeuralNet mDummyNet;
	
	public NeuralNetManager(ValueManager valueManager) {
		mAvailableTransferFunctions.add(new TransferFunctionTanh());
		mAvailableTransferFunctions.add(new TransferFunctionSigmoid());
		mAvailableTransferFunctions.add(new TransferFunctionRamp());
		mAvailableTransferFunctions.add(new TransferFunctionSigmoid2());
		mAvailableTransferFunctions.add(new TransferFunctionRamp2());
		mAvailableTransferFunctions.add(new TransferFunctionStep(3));
		mAvailableTransferFunctions.add(new TransferFunctionStep(5));
		mAvailableTransferFunctions.add(new TransferFunctionStep(9));
		
		
		mNeuralNetInterface = new NeuralNetAgentInterface("default");
		
		mIterationsPerStep = new IntValue(1);
		mIterationsPerStepBuffer = 1;
		mIterationsPerStep.addValueReplacedListener(this);
		
		valueManager.addValue("/NeuralNetManager/Control/IterationsPerStep", mIterationsPerStep);
	}
	
	public synchronized void setNeuralNetInterface(NeuralNetAgentInterface netInterface) {
		mNeuralNetInterface = netInterface;
	}
	
	public NeuralNetAgentInterface getNeuralNetInterface() {
		return mNeuralNetInterface;
	}
	
	public synchronized void setCurrentNet(NeuralNet net) {
		mCurrentNet = net;
		notifyCurrentNeuralNetworkChanged();
	}
	
	public NeuralNet getCurrentNet() {
		return mCurrentNet;
	}

	public boolean cleanUp() {
		return true;
	}

	public boolean init() {
		Simulator.getInstance().getScheduler().addTryListener(this);
		Simulator.getInstance().getScheduler().addStepListener(this);
		Simulator.getInstance().getScheduler().addIndividualListener(this);
		Simulator.getInstance().getScheduler().addGenerationListener(this);
		return true;
	}
	
	
	public void individualCompleted(int individualId) {
	}

	public void individualStarted(int individualId) {	
	}

	public void stepCompleted(int stepNumber) {
		notifyNetworkActivationsUpdated(mCurrentNet);
	}

	public void stepStarted(int stepNumber) {
		if(mCurrentNet != null) {
			//Execute the network the given number of times between two 
			//simulation steps.
			for(int i = 0; i < mIterationsPerStepBuffer; ++i) {
				mCurrentNet.updateNetwork();
			}
		}
	}

	public void tryCompleted(int tryId) {
	}

	public void tryStarted(int tryId) {
		if(mCurrentNet != null) {
			mCurrentNet.reset();
			notifyNetworkActivationsUpdated(mCurrentNet);
		}
	}
	
	
	
	public void generationCompleted(int generationId) {
		
	}

	public void generationStarted(int generationId) {
		mIterationsPerStepBuffer = mIterationsPerStep.get();
	}

	public synchronized boolean addNeuralNetListener(NeuralNetListener listener) {
		if(listener == null || mNeuralNetListeners.contains(listener)) {
			return false;
		}
		mNeuralNetListeners.add(listener);
		return true;
	}
	
	public synchronized boolean removeNeuralNetListener(NeuralNetListener listener) {
		if(listener == null || !mNeuralNetListeners.contains(listener)) {
			return false;
		}
		mNeuralNetListeners.remove(listener);
		return true;
	}
	
	public Vector getNeuralNetListeners() {
		return new Vector(mNeuralNetListeners);
	}
	
	public synchronized void notifyCurrentNeuralNetworkChanged() {
		mListenerBufferVector.clear();
		mListenerBufferVector.addAll(mNeuralNetListeners);
		for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
			((NeuralNetListener) i.next()).currentNeuralNetChanged(mCurrentNet);
		}
	}
	
	public synchronized void notifyNetworkActivationsUpdated(NeuralNet net) {
		mListenerBufferVector.clear();
		mListenerBufferVector.addAll(mNeuralNetListeners);
		for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
			((NeuralNetListener) i.next()).networkActivationsUpdated(net);
		}
	}
	
	public synchronized void notifyNetworkStructureChanged(NeuralNet net) {
		mListenerBufferVector.clear();
		mListenerBufferVector.addAll(mNeuralNetListeners);
		for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
			((NeuralNetListener) i.next()).networkStructureChanged(net);
		}
	}
	
	public void addTransferFunctionPrototype(TransferFunction transferFunction) {
		if(transferFunction == null || mAvailableTransferFunctions.contains(transferFunction)) {
			return;
		}
		mAvailableTransferFunctions.add(transferFunction);
	}
	
	public Vector getTransferFunctionPrototypes() {
		return new Vector(mAvailableTransferFunctions);
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		mIterationsPerStep = (IntValue) ValueManager.synchronizeValue(mIterationsPerStep, oldValue, newValue, this);
		
	}
	
	public IntValue getIterationsPerStepValue() {
		return mIterationsPerStep;
	}
	
	
}
