/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.simulators;

import java.awt.Color;
import java.util.Random;

import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.devices.Box;
import de.fhg.iais.outdoor.simba.devices.DistanceScanner;
import de.fhg.iais.outdoor.simba.devices.OmniDrive;
import de.fhg.iais.outdoor.simba.devices.SimpleBumpDetector;
import de.fhg.iais.outdoor.simba.fitnessfunctions.PathMemorizerFitness3;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetAgentInterface;
import de.fhg.iais.outdoor.simba.neuralnet.NeuronWrapper;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.visualization.SimbaDefaultFrame;

public class PathMemorizer extends SimbaDefaultFrame {

	public NeuralNetAgentInterface mMainController;
	
	public PathMemorizer(String args[]) {
		super("PathMemorizer", 700, 500);
	
		setDefaultOffset(-430.0, -100.0);
		setDefaultZoom(1.0);
		setVisualizationInterval(1);
		setSimulationDelay(15);
		
		mMainController = Simulator.getInstance().getNeuralNetManager().getNeuralNetInterface();
		
		//Add Fitness function.
		setFitnessFunction(new PathMemorizerFitness3(), 10000);
		
		setMasterEnvironment(new MyEnvironment());
		Simulator.getInstance().getNeuralNetManager().setNeuralNetInterface(mMainController);

		startSystem(mMainController, args);
	}
	

	private static int mEnvironmentId = 0;
	private static double[] mStartPositions = new double[10];
	
	public class MyEnvironment extends EnvironmentModel {
		
		private Box mLeftBox;
		private Box mRightBox;
		private Box mFloor;
		private OmniDrive mRobot;
		private Box mRobotBody;
		private DistanceScanner mDistanceLeft;
		private Box mStartingArea;
		private Box mTargetArea;
		private SimpleBumpDetector mBumpLeft;
		private SimpleBumpDetector mBumpRight;
		private BooleanValue mAscendingOrder;
		private BooleanValue mUseRandomizedStartingPositions;
		private DoubleValue mRightWallDistanceValue;
		private double mRightWallDistance;
		
		
		private double mStartPositionOffset;
	
		public MyEnvironment() {
			super("PathMemorizer");
			mStartPositions[0] = 270;
			mStartPositions[1] = 170;
			mStartPositions[2] = 70;
			mStartPositions[3] = -30;
			mStartPositions[4] = -100;
			mStartPositions[5] = -170;
			mStartPositions[6] = 220;
			mStartPositions[7] = 120;
			mStartPositions[8] = 20;
			mStartPositions[9] = -50;
			
			mAscendingOrder = new BooleanValue(false);
			mUseRandomizedStartingPositions = new BooleanValue(true);
			mRightWallDistanceValue = new DoubleValue(100.0);
			
			addParameter(mAscendingOrder, "AscendingOrder");
			addParameter(mUseRandomizedStartingPositions, "RandomizeStartingPositions");
			addParameter(mRightWallDistanceValue, "RightWallDistance");
			
		}
		
		public void resetEnvironmentObjects() {
			mEnvironmentId = 0;
			mRobot.setPosition(mStartPositionOffset, mRobot.getPosition().y);
			mStartingArea.setPosition(mRobot.getPosition());
			mRightBox.setPosition(mRobot.getPosition().x + mRightWallDistance, 15.0);
		}
		
		public void updateRandomizationParameters() {
			mRightWallDistance = mRightWallDistanceValue.get();
			Random rand = new Random();
			if(mUseRandomizedStartingPositions.get()) {
				mStartPositionOffset = -50.0 + (rand.nextDouble() * 200.0);
			}
			else {
				if(mAscendingOrder.get()) {
					mStartPositionOffset = 	mStartPositions[
					            (mStartPositions.length - 1) - (mEnvironmentId % mStartPositions.length)];
				}
				else {
					mStartPositionOffset = 	mStartPositions[(mEnvironmentId % mStartPositions.length)];       				
				}
			}
			mEnvironmentId++;
		}
		
		
	
		public String createRandomizationParameterLog() {
			String log =  super.createRandomizationParameterLog();
			log += getParamEntry("StartingPosition", Double.toString(mStartPositionOffset));
			return log;
		}

		public void createEnvironment() {
	
			ValueManager vm = Simulator.getInstance().getValueManager();
			
			//Set up scenario.
			mLeftBox = new Box("LeftBound", -400.0, 15.0, 0.0, 10.0, 30.0);
			mLeftBox.setShapeColor(Color.DARK_GRAY);
			addSimObject(mLeftBox);
			
			mRightBox = new Box("RightBound", 400.0, 15.0, 0.0, 10.0, 30.0);
			mRightBox.setShapeColor(Color.DARK_GRAY);
			addSimObject(mRightBox);
			
			mFloor = new Box("Floor", 0.0, -3.0, 0.0, 810.0, 6.0);
			mFloor.setShapeColor(Color.DARK_GRAY);
			addSimObject(mFloor);
			
			mStartingArea = new Box("StartingArea", 0.0, 15.0, 0.0, 30.0, 30.0);
			mStartingArea.setLayer(-5);
			mStartingArea.setShapeColor(Color.ORANGE);
			addSimObject(mStartingArea);
			
			mTargetArea = new Box("TargetArea", -370.0, 15.0, 0.0, 60.0, 30.0);
			mTargetArea.setLayer(-5);
			mTargetArea.setShapeColor(mStartingArea.getShapeColor());
			addSimObject(mTargetArea);
			
			mRobot = new OmniDrive("Robot", 0.0, 10.0, 0.0, -3.0, 3.0, 0.0, 0.0);
			mRobotBody = new Box("Robot/Body", 0.0, 0.0, 0.0, 20.0, 20.0);
			mRobotBody.setReferenceObject(mRobot);
			mRobotBody.setShapeColor(Color.GREEN);
			addSimObject(mRobot);
			addSimObject(mRobotBody);
			//output neuron for motor.
			{
				String name = "/SimObjects/" + mRobot.getName() + "/VelocityX";
				NeuronWrapper wrapper = new NeuronWrapper(name, mRobot.getVelocityX(), 
						mRobot.getMinVelocity(), mRobot.getMaxVelocity());
				vm.replaceValue(name, wrapper);
				mMainController.addOutputNeuron(wrapper);
			}
			
			mDistanceLeft = new DistanceScanner("DistanceScannerLeft", -10.0, 8.0, -180.0, 0.0, 100.0);
			mDistanceLeft.addObstacle(mLeftBox);
			mDistanceLeft.setReferenceObject(mRobotBody);
			
			addSimObject(mDistanceLeft);
			
			mMainController.addInputNeuron(NeuronWrapper.createWrapperForDistanceScanner(mDistanceLeft));
			
			mBumpLeft = new SimpleBumpDetector("BumpLeft", -10, 0.0, 180.0);
			mBumpLeft.setReferenceObject(mRobotBody);
			mBumpLeft.addObstacle(mLeftBox);
			addSimObject(mBumpLeft);
			Simulator.getInstance().getScheduler().addTerminateTryValue(mBumpLeft.getSensorValueObject());
			
			mBumpRight = new SimpleBumpDetector("BumpRight", 10, 0.0, 180.0);
			mBumpRight.setReferenceObject(mRobotBody);
			mBumpRight.addObstacle(mRightBox);
			addSimObject(mBumpRight);
			Simulator.getInstance().getScheduler().addTerminateTryValue(mBumpRight.getSensorValueObject());
			
			mMainController.setNeuronRange(-1.0, 1.0); //tanh
			
		}
		
	}
	
	
	
	public static void main(String[] args) {
		new PathMemorizer(args);
	}
}
