/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.statistic;

import java.util.Iterator;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.IndividualListener;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNetListener;
import de.fhg.iais.outdoor.simba.neuralnet.Neuron;
import de.fhg.iais.outdoor.simba.neuralnet.Synapse;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.HistoryValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;

public class NetworkStatistics extends StatisticCalculator implements  NeuralNetListener,
				IndividualListener, ValueChangedListener {
	
	
	private IntValue mNumberOfLoops;
	private IntValue mNumberOfNeurons;
	private IntValue mNumberOfSynapses;
	private IntValue mNumberOf2Neurons;
	private IntValue mNumberOf3Neurons;
	private HistoryValue mMaxNumberOfNeurons;
	private HistoryValue mMinNumberOfNeurons;
	private HistoryValue mMeanNumberOfNeurons;
	private HistoryValue mMaxNumberOfSynapses;
	private HistoryValue mMinNumberOfSynapses;
	private HistoryValue mMeanNumberOfSynapses;
	private HistoryValue mMinNumberOfLoops;
	private HistoryValue mMaxNumberOfLoops;
	private HistoryValue mMeanNumberOfLoops;
	private HistoryValue mMinNumberOf2Neurons;
	private HistoryValue mMaxNumberOf2Neurons;
	private HistoryValue mMeanNumberOf2Neurons;
	//TODO 2 neurons, 3 neurons, other ratings.
	
	private int mIndividualCount;
	private double mNeuronSum;
	private double mSynapseSum;
	private double mLoopSum;
	private double m2NeuronSum;
	
	private BooleanValue mStaseActive;
	private boolean mStaseWasActive;
	
	
	/**
	 * Creates a new NetworkStatistic object. This object calculates statistical 
	 * information about the current network, like the number of neurons and synapses,
	 * but also like the number of loops, 2-Neurons or 3-Neurons.
	 *
	 */
	public NetworkStatistics() {
		super("NetworkStatistics");
		
		mNumberOfLoops = new IntValue(0);
		mNumberOfNeurons = new IntValue(0);
		mNumberOfSynapses = new IntValue(0);
		mNumberOf2Neurons = new IntValue(0);
		mNumberOf3Neurons = new IntValue(0);
		mMaxNumberOfNeurons = new HistoryValue();
		mMinNumberOfNeurons = new HistoryValue();
		mMeanNumberOfNeurons = new HistoryValue();
		mMaxNumberOfSynapses = new HistoryValue();
		mMinNumberOfSynapses = new HistoryValue();
		mMeanNumberOfSynapses = new HistoryValue();
		mMinNumberOfLoops = new HistoryValue();
		mMaxNumberOfLoops = new HistoryValue();
		mMeanNumberOfLoops = new HistoryValue();
		mMinNumberOf2Neurons = new HistoryValue();
		mMaxNumberOf2Neurons = new HistoryValue();
		mMeanNumberOf2Neurons = new HistoryValue();
		
		addParameter(mNumberOfLoops, "NumberOfLoops");
		addParameter(mNumberOfNeurons, "NumberOfNeurons");
		addParameter(mNumberOfSynapses, "NumberOfSynapses");
		addParameter(mNumberOf2Neurons, "NumberOf2Neurons");
		addParameter(mNumberOf3Neurons, "NumberOf3Neurons");
		addParameter(mMaxNumberOfNeurons, "MaxNumberOfNeurons");
		addParameter(mMinNumberOfNeurons, "MinNumberOfNeurons");
		addParameter(mMeanNumberOfNeurons, "MeanNumberOfNeurons");
		addParameter(mMaxNumberOfSynapses, "MaxNumberOfSynapses");
		addParameter(mMinNumberOfSynapses, "MinNumberOfSynapses");
		addParameter(mMeanNumberOfSynapses, "MeanNumberOfSynapses");
		addParameter(mMinNumberOfLoops, "MinNumberOfLoops");
		addParameter(mMaxNumberOfLoops, "MaxNumberOfLoops");
		addParameter(mMeanNumberOfLoops, "MeanNumberOfLoops");
		addParameter(mMinNumberOf2Neurons, "MinNumberOf2Neurons");
		addParameter(mMaxNumberOf2Neurons, "MaxNumberOf2Neurons");
		addParameter(mMeanNumberOf2Neurons, "MeanNumberOf2Neurons");
		
		mIndividualCount = 1;
		Simulator.getInstance().getScheduler().addIndividualListener(this);
		Simulator.getInstance().getStatisticManager().addCalculator(this);
		Simulator.getInstance().getNeuralNetManager().addNeuralNetListener(this);
		
		mStaseActive = Simulator.getInstance().getEvolutionManager().getStaseModeValue();
		mStaseActive.addValueChangedListener(this);
		
	}
	
	
	
	
	public void calculate() {
		mNeuronSum = 0.0;
		mSynapseSum = 0.0;
		mLoopSum = 0.0;
		m2NeuronSum = 0.0;
	}




	/**
	 * Updates the Value objects with the current statistical data.
	 * @param currentNet
	 */
	public void updateStatistics(NeuralNet currentNet) {
		if(currentNet == null) {
			return;
		}
		
		int noLoops = 0;
		int noSynapses = 0;
		int noNeurons = 0;
		int no2Neurons = 0;
		int no3Neurons = 0;
		
		synchronized(currentNet) {
			noSynapses = currentNet.getSynapses().size();
			noNeurons = currentNet.getNeurons().size();
			noLoops = getNumberOfLoops(currentNet);
			no2Neurons = getNumberOf2Neurons(currentNet);
			no3Neurons = getNumberOf3Neurons(currentNet);
			
		}
		if(mNumberOfLoops.get() != noLoops) {
			mNumberOfLoops.set(noLoops);
		}
		if(mNumberOfNeurons.get() != noNeurons) {
			mNumberOfNeurons.set(noNeurons);
		}
		if(mNumberOfSynapses.get() != noSynapses) {
			mNumberOfSynapses.set(noSynapses);
		}
		if(mNumberOf2Neurons.get() != no2Neurons) {
			mNumberOf2Neurons.set(no2Neurons);
		}
		if(mNumberOf3Neurons.get() != no3Neurons) {
			mNumberOf3Neurons.set(no3Neurons);
		}
		
		
		
	}
	
	public void calculateGenerationStatistics() {
		
		int noLoops = mNumberOfLoops.get();
		int noSynapses = mNumberOfSynapses.get();
		int noNeurons = mNumberOfNeurons.get();
		int no2Neurons = mNumberOf2Neurons.get();
		//int no3Neurons = mNumberOf3Neurons.get();
		
		if(!mStaseWasActive) {
			mLoopSum += noLoops;
			mSynapseSum += noSynapses;
			mNeuronSum += noNeurons;
			m2NeuronSum += no2Neurons;
		}
		
//		Loop statistic
		if(noLoops > mMaxNumberOfLoops.get()) {
			mMaxNumberOfLoops.set((double) noLoops);
		}
		if(mIndividualCount == 1) {
			mMinNumberOfLoops.set(noLoops);
		}
		else {
			if(noLoops < mMinNumberOfLoops.get()) {
				mMinNumberOfLoops.set((double) noLoops);
			}
		}
		mMeanNumberOfLoops.set(((double) mLoopSum) / ((double) mIndividualCount));
		
		
		//Neuron statistic
		if(noNeurons > mMaxNumberOfNeurons.get()) {
			mMaxNumberOfNeurons.set((double) noNeurons);
		}
		if(mIndividualCount == 1) {
			mMinNumberOfNeurons.set(noNeurons);
		}
		else {
			if(noNeurons < mMinNumberOfNeurons.get()) {
				mMinNumberOfNeurons.set((double) noNeurons);
			}
		}
		mMeanNumberOfNeurons.set(((double) mNeuronSum) / ((double) mIndividualCount));
		
		
		
		//Synapse statistic
		if(noSynapses > mMaxNumberOfSynapses.get()) {
			mMaxNumberOfSynapses.set((double) noSynapses);
		}
		if(mIndividualCount == 1) {
			mMinNumberOfSynapses.set((double) noSynapses);
		}
		else {
			if(noSynapses < mMinNumberOfSynapses.get()) {
				mMinNumberOfSynapses.set((double) noSynapses);
			}
		}
		mMeanNumberOfSynapses.set(((double) mSynapseSum) / ((double) mIndividualCount));
		
		
		//2Neuron statistic
		if(no2Neurons > mMaxNumberOf2Neurons.get()) {
			mMaxNumberOf2Neurons.set((double) no2Neurons);
		}
		if(mIndividualCount == 1) {
			mMinNumberOf2Neurons.set((double) no2Neurons);
		}
		else {
			if(no2Neurons < mMinNumberOf2Neurons.get()) {
				mMinNumberOf2Neurons.set((double) no2Neurons);
			}
		}
		mMeanNumberOf2Neurons.set(((double) m2NeuronSum) / ((double) mIndividualCount));
		
	}
	
	/**
	 * Returns the number of direct loops (neuron with a synapse to itself).
	 * @param currentNet
	 * @return
	 */
	public int getNumberOfLoops(NeuralNet currentNet) {
		
		if(currentNet == null) {
			return 0;
		}
		
		int noLoops = 0;
		
		synchronized(currentNet) {
		
			Vector neurons = currentNet.getNeurons();
			
			for(Iterator i = neurons.iterator(); i.hasNext();) {
				Neuron n = (Neuron) i.next();
				Vector synapses = n.getSynapses();
				for(Iterator j = synapses.iterator(); j.hasNext();) {
					Synapse s = (Synapse) j.next();
					if(s.getSourceNeuron() == n) {
						noLoops++;
					}
				}
			}	
		}
		return noLoops;
	}
	
	/**
	 * Returns the number of 2-Neuron modules (n1 -> n2, n2 -> n1). 
	 * The strength of the synapses is not important.
	 * @param currentNet
	 * @return
	 */
	public int getNumberOf2Neurons(NeuralNet currentNet) {
		
		if(currentNet == null) {
			return 0;
		}
		int no2Neurons = 0;
		
		synchronized(currentNet) {
			
			Vector neurons = currentNet.getNeurons();
			
			for(Iterator i = neurons.iterator(); i.hasNext();) {
				Neuron n = (Neuron) i.next();
				
				Vector synapses = n.getSynapses();
				
				for(Iterator j = synapses.iterator(); j.hasNext();) {
					Synapse s = (Synapse) j.next();
					Neuron source = (Neuron) s.getSourceNeuron();
						if(source != n) {
						Vector sourceSynapses = source.getSynapses();
						for(Iterator k = sourceSynapses.iterator(); k.hasNext();) {
							if(((Synapse) k.next()).getSourceNeuron() == n) {
								no2Neurons++;
							}
						}
					}
					
				}
			}	
		}
		
		return no2Neurons / 2;
		
	}
	
	
	/**
	 * Returns the number of 3-Neuron modules (n1 -> n2, n2 -> n3, n3 -> n1).
	 * The strength of the synapses is not considered.
	 * @param currentNet
	 * @return
	 */
	public int getNumberOf3Neurons(NeuralNet currentNet) {
		
		if(currentNet == null) {
			return 0;
		}
		int no3Neurons = 0;
		
		synchronized(currentNet) {
			
			Vector neurons = currentNet.getNeurons();
			
			for(Iterator i = neurons.iterator(); i.hasNext();) {
				Neuron n = (Neuron) i.next();
				
				Vector synapses = n.getSynapses();
				
				for(Iterator j = synapses.iterator(); j.hasNext();) {
					Synapse s = (Synapse) j.next();
					Neuron source = (Neuron) s.getSourceNeuron();
						if(source != n) {
						Vector sourceSynapses = source.getSynapses();
						for(Iterator k = sourceSynapses.iterator(); k.hasNext();) {
							Synapse sourceSynapse = (Synapse) k.next();
							Neuron sourceOfSource = (Neuron) sourceSynapse.getSourceNeuron();
							if(sourceOfSource != n && sourceOfSource != source) {
								Vector sourceOfSourceSynapses = sourceOfSource.getSynapses();
								for(Iterator m = sourceOfSourceSynapses.iterator(); m.hasNext();) {
									if(((Synapse) m.next()).getSourceNeuron() == n) {
										no3Neurons++;
									}
								}
							}
						}
					}
					
				}
			}	
		}
		
		return no3Neurons / 3;
		
	}

	public void currentNeuralNetChanged(NeuralNet currentNet) {
		updateStatistics(currentNet);
	}

	public void networkActivationsUpdated(NeuralNet currentNet) {
	}

	public void networkStructureChanged(NeuralNet currentNet) {
		updateStatistics(currentNet);
	}
	
	

	public void individualCompleted(int individualId) {
	}


	public void individualStarted(int individualId) {
		mIndividualCount = individualId + 1;
		if(mStaseActive.get()) {
			mStaseWasActive = true;
		}
		else {
			mStaseWasActive = false;
		}
		calculateGenerationStatistics();
	}


	


	public void valueChanged(Value value) {
		if(value == null) {
			return;
		}
		if(value == mStaseActive) {
			if(mStaseActive.get()) {
				mStaseWasActive = true;
			}
		}
	}




	public void valueReplaced(Value oldValue, Value newValue) {
		mNumberOf2Neurons = (IntValue) ValueManager.synchronizeValue(mNumberOf2Neurons, oldValue, newValue, this);
		mNumberOf3Neurons = (IntValue) ValueManager.synchronizeValue(mNumberOf3Neurons, oldValue, newValue, this);
		mNumberOfLoops = (IntValue) ValueManager.synchronizeValue(mNumberOfLoops, oldValue, newValue, this);
		mNumberOfNeurons = (IntValue) ValueManager.synchronizeValue(mNumberOfNeurons, oldValue, newValue, this);
		mNumberOfSynapses = (IntValue) ValueManager.synchronizeValue(mNumberOfSynapses, oldValue, newValue, this);
		mMaxNumberOfNeurons = (HistoryValue) ValueManager.synchronizeValue(mMaxNumberOfNeurons, oldValue, newValue, this);
		mMinNumberOfNeurons = (HistoryValue) ValueManager.synchronizeValue(mMinNumberOfNeurons, oldValue, newValue, this);
		mMeanNumberOfNeurons = (HistoryValue) ValueManager.synchronizeValue(mMeanNumberOfNeurons, oldValue, newValue, this);
		mMaxNumberOfSynapses = (HistoryValue) ValueManager.synchronizeValue(mMaxNumberOfSynapses, oldValue, newValue, this);
		mMinNumberOfSynapses = (HistoryValue) ValueManager.synchronizeValue(mMinNumberOfSynapses, oldValue, newValue, this);
		mMeanNumberOfSynapses = (HistoryValue) ValueManager.synchronizeValue(mMeanNumberOfSynapses, oldValue, newValue, this);
		mMinNumberOfLoops = (HistoryValue) ValueManager.synchronizeValue(mMinNumberOfLoops, oldValue, newValue, this);
		mMaxNumberOfLoops = (HistoryValue) ValueManager.synchronizeValue(mMaxNumberOfLoops, oldValue, newValue, this);
		mMeanNumberOfLoops = (HistoryValue) ValueManager.synchronizeValue(mMeanNumberOfLoops, oldValue, newValue, this);
		mMinNumberOf2Neurons = (HistoryValue) ValueManager.synchronizeValue(mMinNumberOf2Neurons, oldValue, newValue, this);
		mMaxNumberOf2Neurons = (HistoryValue) ValueManager.synchronizeValue(mMaxNumberOf2Neurons, oldValue, newValue, this);
		mMeanNumberOf2Neurons = (HistoryValue) ValueManager.synchronizeValue(mMeanNumberOf2Neurons, oldValue, newValue, this);
	
	}
	
	

}
