/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.test;

import java.awt.Color;
import java.awt.geom.Point2D;

import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.SimObjectSet;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.devices.Box;
import de.fhg.iais.outdoor.simba.devices.DistanceScanner;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.NormalizedDoubleValue;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import junit.framework.TestCase;

public class TestDistanceScanner extends TestCase {

	protected void tearDown() throws Exception {
		super.tearDown();
		Simulator.getInstance().clearInstance();
	}
	
	/*
	 * Test method for 'de.fhg.iais.outdoor.simba.core.DistanceScanSimObject.DistanceScanSimObject(double, double, double, double, double, double, double)'
	 */
	public void testConstructor() {
		DistanceScanner obj = new DistanceScanner("", 
				1.0, 2.0, 0.1, 0.5, 6.5);
		
		assertEquals(obj.getPosition(), new Point2D.Double(1.0, 2.0));
		assertEquals(obj.getAngle(), 0.1, 0.001);
		assertEquals(obj.getMinScanDistance(), 0.5, 0.001);
		assertEquals(obj.getMaxScanDistance(), 6.5, 0.001);
		assertEquals(obj.getObstacles().size(), 0);
		assertEquals(obj.getDependencies().size(), 0);
		assertTrue(obj.isActive());
		assertTrue(obj.isVisible());
		
		assertSame(obj.getObstacleDetectedColor(), Color.red); //default
		obj.setObstacleDetectedColor(Color.BLUE);
		assertSame(obj.getObstacleDetectedColor(), Color.BLUE);

	}
	
	public void testGlobalStuffCalculation() {
		
		DistanceScanner obj = new DistanceScanner("", 
				1.0, 2.0, 0.1, 0.5, 6.5);
		
		assertEquals(obj.getGlobalAngle(), 0.1, 0.001);
		assertEquals(obj.getGlobalPosition().x, 1.0, 0.001);
		assertEquals(obj.getGlobalPosition().y, 2.0, 0.001);
		
		obj.setPosition(10.0, 12.0);
		obj.setAngle(43.21);
		
		assertEquals(obj.getGlobalAngle(), 0.1, 0.001);
		assertEquals(obj.getGlobalPosition().x, 1.0, 0.001);
		assertEquals(obj.getGlobalPosition().y, 2.0, 0.001);
		
		obj.update();
		assertEquals(obj.getGlobalAngle(), 43.21, 0.001);
		assertEquals(obj.getGlobalPosition().x, 10.0, 0.001);
		assertEquals(obj.getGlobalPosition().y, 12.0, 0.001);
	}
	
	/*
	 * Test method for 'de.fhg.iais.outdoor.simba.core.DistanceScanSimObject.update()'
	 */
	public void testUpdate() {
		DistanceScanner scanObj = new DistanceScanner("",
				1.0, 1.0, 0.0, 0.5, 6.5);
		
		scanObj.update();
		assertEquals(scanObj.getSensorReading(), 6.5, 0.001);
		
		Box obj = new Box("Box1", 5.5, 0.0, 0.0, 2.0, 3.0);
		scanObj.addObstacle(obj);
		
		scanObj.update();
		assertEquals(scanObj.getSensorReading(), 3.5, 0.0001);
		
		obj.setActive(false); //inactive objects are ignored
		scanObj.update();
		assertEquals(scanObj.getSensorReading(), 6.5, 0.0001);
		
		obj.setActive(true); 
		obj.setPosition(2.1, 0.0);
		obj.update();
		scanObj.update(); //this time the obstacle is closer than the min range.
		assertEquals(scanObj.getSensorReading(), 0.5, 0.0001);
		
		obj.setHeight(0.5); //obstacle is too small, so the scanner does not detect it.
		obj.update();
		scanObj.update();
		assertEquals(scanObj.getSensorReading(), 6.5, 0.0001);
		
		Box parent = new Box("Parent", -1.0, -1.0, 90.0, 1.0, 1.0);
		scanObj.setReferenceObject(parent);
		obj.setReferenceObject(parent);
		
		obj.setHeight(5.0);
		obj.update();
		scanObj.update(); //reference object moved coordinate system to -1/-1. Relation remains.
		assertEquals(scanObj.getSensorReading(), 0.5, 0.0001);
		
		obj.setPosition(5.5, 0.0);
		obj.update();
		scanObj.update(); //reference objects of obstacle and scanner differs. Distance differnt.
		assertEquals(scanObj.getSensorReading(), 3.5, 0.0001);
		
		parent.setAngle(0.0);
		parent.update();
		obj.update();
		Box parent2 = new Box("Parent2", -2.5, -1.0, 0.0, 1.0, 1.0);
		scanObj.setReferenceObject(parent2);
		scanObj.update(); //reference objects of obstacle and scanner differs. Distance differnt.
		assertEquals(scanObj.getSensorReading(), 5.0, 0.0001);
		
		
	}
	
	public void testObstacleSets() {
		
		DistanceScanner scanObj = new DistanceScanner("",
				1.0, 1.0, 0.0, 0.5, 6.5);
		
		SimObjectSet set = new SimObjectSet();
		
		scanObj.update();
		assertEquals(scanObj.getSensorReading(), 6.5, 0.001);
		
		Box obj = new Box("Box1", 5.5, 0.0, 0.0, 2.0, 3.0);
		assertTrue(set.add(obj));
		
		assertTrue(scanObj.addObstacleSet(set));
		scanObj.update();
		assertEquals(scanObj.getSensorReading(), 3.5, 0.0001);
		
		assertFalse(scanObj.addObstacleSet(set)); // already contained.
		scanObj.update();
		assertEquals(scanObj.getSensorReading(), 3.5, 0.0001);
		
		assertFalse(scanObj.removeObstacleSet(null));
		scanObj.update();
		assertEquals(scanObj.getSensorReading(), 3.5, 0.0001);
		
		assertTrue(scanObj.removeObstacleSet(set));
		scanObj.update();
		assertEquals(scanObj.getSensorReading(), 6.5, 0.0001);
		
		assertFalse(scanObj.removeObstacleSet(set)); //not contained.
		scanObj.update();
		assertEquals(scanObj.getSensorReading(), 6.5, 0.0001);
		
		
	}

	/*
	 * Test method for 'de.fhg.iais.outdoor.simba.core.DistanceScanSimObject.reset()'
	 */
	public void testReset() {

	}

	/*
	 * Test method for 'de.fhg.iais.outdoor.simba.core.DistanceScanSimObject.addObstacle(SimObject)'
	 */
	public void testAddObstacle() {
		DistanceScanner scanObj = new DistanceScanner("",
				1.0, 1.0, 45.0, 0.0, 4.5);
		
		assertEquals(scanObj.getObstacles().size(), 0);
		assertEquals(scanObj.getDependencies().size(), 0);
		SimObject o1 = new SimObject();
		SimObject o2 = new SimObject();
		assertTrue(scanObj.addObstacle(o1));
		assertEquals(scanObj.getObstacles().size(), 1);
		assertEquals(scanObj.getDependencies().size(), 1);
		
		assertFalse(scanObj.addObstacle(o1));
		assertEquals(scanObj.getObstacles().size(), 1);
		assertEquals(scanObj.getDependencies().size(), 1);
		
		assertTrue(scanObj.addDependency(o2));
		assertEquals(scanObj.getObstacles().size(), 1);
		assertEquals(scanObj.getDependencies().size(), 2);
		
		assertTrue(scanObj.addObstacle(o2));
		assertEquals(scanObj.getObstacles().size(), 2);
		assertEquals(scanObj.getDependencies().size(), 2); //was already registered as dependency
	}

	public void testValueReplaced() {
		DistanceScanner s = new DistanceScanner("ScannerX",
				1.0, 1.0, 45.0, 0.0, 4.5);
		
		ValueManager vm = Simulator.getInstance().getValueManager();
		DoubleValue distance = vm.getDoubleValue("/SimObjects/ScannerX/Distance");
		DoubleValue minDistance = vm.getDoubleValue("/SimObjects/ScannerX/MinScanDistance");
		DoubleValue maxDistance = vm.getDoubleValue("/SimObjects/ScannerX/MaxScanDistance");
		DoubleValue angle = vm.getDoubleValue("/SimObjects/ScannerX/LocalAngle");
		
		assertNotNull(distance);
		assertNotNull(minDistance);
		assertNotNull(maxDistance);
		assertNotNull(angle);
		
		minDistance.set(-1000.0);
		maxDistance.set(1000.0);
		distance.set(99.9);
		angle.set(55.5);
		
		assertEquals(s.getSensorReading(), 99.9, 0.001);
		assertEquals(s.getMinScanDistance(), -1000.0, 0.001);
		assertEquals(s.getMaxScanDistance(), 1000.0, 0.001);
		assertEquals(s.getAngle(), 55.5, 0.001);
		
		NormalizedDoubleValue dist2 = new NormalizedDoubleValue(111.11, 0.0, 1000.0, 0.0, 0.0);
		DoubleValue minDist2 = new DoubleValue(44.44);
		DoubleValue maxDist2 = new DoubleValue(88.8);
		DoubleValue angle2 = new DoubleValue(7.77);
		
		assertTrue(vm.replaceValue("/SimObjects/ScannerX/Distance", dist2));
		assertTrue(vm.replaceValue("/SimObjects/ScannerX/MinScanDistance", minDist2));
		assertTrue(vm.replaceValue("/SimObjects/ScannerX/MaxScanDistance", maxDist2));
		assertTrue(vm.replaceValue("/SimObjects/ScannerX/LocalAngle", angle2));
		
		assertEquals(s.getSensorReading(), 111.11, 0.001);
		assertEquals(s.getMinScanDistance(), 44.44, 0.001);
		assertEquals(s.getMaxScanDistance(), 88.8, 0.001);
		assertEquals(s.getAngle(), 7.77, 0.001);
		
	}

	

}
