/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.test;

import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.fitness.FitnessFunction;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import junit.framework.TestCase;

public class TestFitnessFunction extends TestCase {

	
	protected void tearDown() throws Exception {
		super.tearDown();
		Simulator.getInstance().clearInstance();
	}
	
	public void testInit() {
		AdapterFitnessFunction ff = new AdapterFitnessFunction("TestFitness");
		assertEquals(ff.getName(), "TestFitness");
		assertEquals(ff.getFitness(), 0.0, 0.0001);
		assertEquals(ff.getMeanFitness(), 0.0, 0.0001);
		assertTrue(ff.isActive());
		
		ff.setActive(false);
		assertFalse(ff.isActive());
		
		ff.setFitness(99.12);
		assertEquals(ff.getFitness(), 99.12, 0.0001);
		assertEquals(ff.getMeanFitness(), 0.0, 0.0001);
		
		ff.update();
		assertEquals(ff.getFitness(), 0.0, 0.0001); //because fitnessToReturn is 0.0
		
		ff.mFitnessToSet = 99.12;
		assertEquals(ff.getFitness(), 0.0, 0.0001);
		ff.update();
		
		assertEquals(ff.getFitness(), 99.12, 0.0001);
		assertEquals(ff.getMeanFitness(), 99.12, 0.0001);
		
		ff.setFitness(55.1);
		assertEquals(ff.getFitness(), 55.1, 0.0001);
		assertEquals(ff.getMeanFitness(), 99.12, 0.0001);
		
		//Test value registration.
		ValueManager vm = Simulator.getInstance().getValueManager();
		DoubleValue fitness = vm.getDoubleValue("/FitnessFunctions/TestFitness/Fitness");
		assertNotNull(fitness);
		assertEquals(fitness.get(), 55.1, 0.0001);
		
		DoubleValue meanFitness = vm.getDoubleValue("/FitnessFunctions/TestFitness/MeanFitness");
		assertNotNull(meanFitness);
		assertEquals(meanFitness.get(), 99.12, 0.0001);
		
		
		//Test value replaced.
		DoubleValue newFitness = new DoubleValue(123.456);
		assertTrue(vm.replaceValue("/FitnessFunctions/TestFitness/Fitness", newFitness));
		DoubleValue newFitnessPointer = vm.getDoubleValue("/FitnessFunctions/TestFitness/Fitness");
		assertNotSame(newFitnessPointer, fitness);
		assertSame(newFitnessPointer, newFitness);
		assertEquals(ff.getFitness(), 123.456, 0.0001);
		
		DoubleValue newMeanFitness = new DoubleValue(876.543);
		assertTrue(vm.replaceValue("/FitnessFunctions/TestFitness/MeanFitness", newMeanFitness));
		DoubleValue newMeanFitnessPointer = vm.getDoubleValue("/FitnessFunctions/TestFitness/MeanFitness");
		assertNotSame(newMeanFitnessPointer, meanFitness);
		assertSame(newMeanFitnessPointer, newMeanFitness);
		assertEquals(ff.getMeanFitness(), 876.543, 0.0001);
		
	}
	

	public void testMeanUpdate() {
		
		AdapterFitnessFunction ff = new AdapterFitnessFunction("TestFitness");
		
		assertEquals(ff.getFitness(), 0.0, 0.0001);
		assertEquals(ff.getMeanFitness(), 0.0, 0.0001);
		
		ff.mFitnessToSet = 123.234;
		
		//executes only the fitness update, not the mean fitness update.
		assertEquals(ff.calculateFitness(), 123.234, 0.001);
		assertEquals(ff.getFitness(), 0.0, 0.0001);
		assertEquals(ff.getMeanFitness(), 0.0, 0.0001);
		
		//includes both, the fitness update and the mean fitness update.
		ff.update();
		assertEquals(ff.getFitness(), 123.234, 0.0001);
		assertEquals(ff.getMeanFitness(), 123.234, 0.0001);
		
		//next call to update
		ff.mFitnessToSet = 50.0;
		ff.update();
		assertEquals(ff.getFitness(), 50.0, 0.0001);
		assertEquals(ff.getMeanFitness(), 50.0, 0.0001);
		
		//signal the next try
		ff.mFitnessToSet = 0.0;
		ff.prepareNextTry(1);
		assertEquals(ff.getFitness(), 0.0, 0.0001);
		assertEquals(ff.getMeanFitness(), 50.0, 0.0001);
		
		ff.update();
		assertEquals(ff.getFitness(), 0.0, 0.0001);
		assertEquals(ff.getMeanFitness(), 25.0, 0.0001); //50 %
		
		ff.mFitnessToSet = 100.0;
		ff.update();
		assertEquals(ff.getFitness(), 100.0, 0.0001);
		assertEquals(ff.getMeanFitness(), 75.0, 0.0001); //50 %
		
		
		ff.prepareNextTry(5);
		ff.mFitnessToSet = 0.0;
		ff.update();
		assertEquals(ff.getFitness(), 0.0, 0.0001);
		assertEquals(ff.getMeanFitness(), 25.0, 0.0001); //150 / 6;
		
		ff.mFitnessToSet = 66.6;
		ff.update();
		assertEquals(ff.getFitness(), 66.6, 0.0001);
		assertEquals(ff.getMeanFitness(), 36.1, 0.0001); //25.0 + 11.1
		
		
	}

	
	public class AdapterFitnessFunction extends FitnessFunction {
		
		public int mUpdateCounter;
		public double mFitnessToSet;
		
		public AdapterFitnessFunction(String name) {
			super(name);	
		}
		
		

		public void reset() {
		}

		public double calculateFitness() {
			mUpdateCounter++;
			return mFitnessToSet;
		}
		
		
	}

}
