/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.value;

public class NormalizedDoubleValue extends DoubleValue {

	private double mMinValue;
	private double mMaxValue;
	private double mMinNormalizedValue;
	private double mMaxNormalizedValue;
	private boolean mInvertNormalized;
	
	/**
	 * Constructor.
	 * @param value the initial value. Must be in the range specified by minvalue and maxValue.
	 * @param minValue the minimal real value this Value can contain. Smaller values are truncated to this minimal value.
	 * @param maxValue the maximal real value this Value can contain. Greater values are truncated to this maximal value.
	 * @param minNormalizedValue the minimal normalized value. Smaller normalized values are truncated to this smaller value.
	 * @param maxNormalizedValue the maximal normalized value. Greater normalized values are truncated to this maximal calue.
	 */
	public NormalizedDoubleValue(double value, double minValue, double maxValue,
			double minNormalizedValue, double maxNormalizedValue) 
	{
		super();
		if(minValue < maxValue) {
			mMinValue = minValue;
			mMaxValue = maxValue;
		}
		else {
			mMinValue = -1000.0;
			mMaxValue = 1000.0;
		}
		if(minNormalizedValue < maxNormalizedValue) {
			mMinNormalizedValue = minNormalizedValue;
			mMaxNormalizedValue = maxNormalizedValue;
		}
		else {
			mMinNormalizedValue = 0.0;
			mMaxNormalizedValue = 1.0;
		}
		if(value < mMinValue) {
			value = mMinValue;
		}
		else if(value > mMaxValue) {
			value = mMaxValue;
		}
		mValue = value;
		mInvertNormalized = false;
	}
	
	
	/**
	 * Sets the value of this DoubleValue in non-normalized values.
	 * Tha value is truncated to the range if it is out of the specified 
	 * bounds.
	 * 
	 * @param value the value to set.
	 */
	public void set(double value) {
		if(value < mMinValue) {
			value = mMinValue;
		}
		else if(value > mMaxValue) {
			value = mMaxValue;
		}
		super.set(value);
	}
	
	/**
	 * Returns the value in a normalized range.
	 * The range between min and max of the value is mapped to a value
	 * between min and max of the normalized range. 
	 * 
	 * @return the normalized value.
	 */
	public double getNormalized() {
		double value = (mMinNormalizedValue + (((mMaxNormalizedValue - mMinNormalizedValue) 
				/ (mMaxValue - mMinValue)) * (mValue - mMinValue)));
		if(mInvertNormalized) {
			return mMinNormalizedValue + (mMaxNormalizedValue - value);
		}
		else {
			return value;
		}
	}
	
	/**
	 * Sets the value in a normalized form. Values that are smaller or greater than
	 * the minimal and maximal range of the normalized value are truncated to fit 
	 * into this range.
	 * 
	 * @param value the value to set in the normalized value range. 
	 */
	public void setNormalized(double value) {
		if(mInvertNormalized) {
			value = mMinNormalizedValue + (mMaxNormalizedValue - value);
		}
		if(value < mMinNormalizedValue) {
			value = mMinNormalizedValue;
		}
		else if(value > mMaxNormalizedValue) {
			value = mMaxNormalizedValue;
		}
		
		set((mMinValue + (((mMaxValue - mMinValue) 
				/ (mMaxNormalizedValue - mMinNormalizedValue)) * (value - mMinNormalizedValue))));
	}
	
	/**
	 * Sets the range for the non-normalized values. Min must be smaller than max. If not, then 
	 * nothing is changed.
	 * 
	 * @param min the new minimum for non-normalized values.
	 * @param max the new maximum for non-normalized values.
	 */
	public void setValueRange(double min, double max) {
		if(min < max) {
			mMinValue = min;
			mMaxValue = max;
		}
		if(mValue < mMinValue) {
			set(mMinValue);
		}
		else if(mValue > mMaxValue) {
			set(mMaxValue);
		}
	}
	
	
	/**
	 * Sets the range for the normalized values. Min must be smaller than max. If not, then
	 * nothing is changed.
	 * 
	 * @param min the new minimum for normalized values.
	 * @param max the new maximum for normalized values.
	 */
	public void setNormalizedValueRange(double min, double max) {
		if(min < max) {
			mMinNormalizedValue = min;
			mMaxNormalizedValue = max;
		}
	}
	
	/**
	 * Returns the minimum for non-normalized values.
	 * @return the minimum for non-normalized values.
	 */
	public double getMinValue() {
		return mMinValue;
	}
	
	
	/**
	 * Returns the maximum for non-normalized values.
	 * @return the maximum for non-normalized values.
	 */
	public double getMaxValue() {
		return mMaxValue;
	}
	
	/**
	 * Returns the minimum for normalized values.
	 * @return the minimum for ormalized values.
	 */
	public double getMinNormalizedValue() {
		return mMinNormalizedValue;
	}
	
	/**
	 * Returns the maximum for normalized values.
	 * @return the maximum for normalized values.
	 */
	public double getMaxNormalizedValue() {
		return mMaxNormalizedValue;
	}
	
	//TODO test
	public boolean setNormalizedValueFromString(String value) {
		try {
			double v = Double.parseDouble(value);
			if(v != getNormalized()) {
				setNormalized(v);
			}
			return true;
		} catch (NumberFormatException e) {
			return false;
		}
	}
	
	public String getNormalizedValueAsString() {
		return Double.toString(getNormalized());
	}
	
	/**
	 * Invers the output of the normalized values.
	 * Inversion means that the output is max - current instead of current.
	 * @param inverted
	 */
	public void setInverted(boolean inverted) {
		if(inverted != mInvertNormalized) {
			double normalized = getNormalized();
			mInvertNormalized = inverted;
			setNormalized(mMinNormalizedValue + (mMaxNormalizedValue - normalized));
		}
		else {
			mInvertNormalized = inverted;
		}
	}
	
	public boolean isInverted() {
		return mInvertNormalized;
	}
	
}
