/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.value;

import java.util.Iterator;
import java.util.Vector;


public class Value {
	
	private Vector mValueChangedListeners = new Vector();
	private Vector mValueReplacedListeners = new Vector();
	private Vector mListenerBufferVector = new Vector();
	private int mNotifyCount;
	
	public Value() {
		mNotifyCount = 0;
	}
	
	/**
	 * Returns the content of the Value as String.
	 * This method should be implemented in a suitable way in each subclass.
	 * @return the string representation of the Values content.
	 */
	public String getValueAsString() {
		return "";
	}
	
	
	/**
	 * Sets the content of the Value from a String.
	 * This method should be implemented in a suitable way in each subclass.
	 * @return true if the parameter could successfully be parsed to set the value.
	 */
	public boolean setValueFromString(String value) {
		return false;
	}
	
	
	/**
	 * Adds a ValueChangedListener.
	 * @param listener the listener to add.
	 * @return true if successful, false if already contained.
	 */
	public synchronized boolean addValueChangedListener(ValueChangedListener listener) {
		if(listener == null || mValueChangedListeners.contains(listener)) {
			return false;
		}
		mValueChangedListeners.add(listener);
		return true;
	}
	
	
	/**
	 * Removes a ValueChangedListener from a Value object.
	 * 
	 * @param listener the listener to remove.
	 * @return true if successful. False if the listener was not contained.
	 */
	public synchronized boolean removeValueChangedListener(ValueChangedListener listener) {
		if(listener == null || !mValueChangedListeners.contains(listener)) {
			return false;
		}
		mValueChangedListeners.remove(listener);
		return true;
	}
	
	/**
	 * Adds a ValueReplacedListener.
	 * @param listener the listener to add.
	 * @return true if successful, false if already contained.
	 */
	public synchronized boolean addValueReplacedListener(ValueReplacedListener listener) {
		if(listener == null || mValueReplacedListeners.contains(listener)) {
			return false;
		}
		mValueReplacedListeners.add(listener);
		return true;
	}
	
	
	/**
	 * Removes a ValueReplacedListener from a Value object.
	 * 
	 * @param listener the listener to remove.
	 * @return true if successful. False if the listener was not contained.
	 */
	public synchronized boolean removeValueReplacedListener(ValueReplacedListener listener) {
		if(listener == null || !mValueReplacedListeners.contains(listener)) {
			return false;
		}
		mValueReplacedListeners.remove(listener);
		return true;
	}
	
	/**
	 * Returns a Vector with all registered ValueChangedListeners.
	 * @return all ValueChangedListeners.
	 */
	public Vector getValueChangedListeners() {
		return new Vector(mValueChangedListeners);
	}
	
	/**
	 * Returns a Vector with all registered ValueChangedListeners.
	 * @return all ValueChangedListeners
	 */
	public Vector getValueReplacedListeners() {
		return new Vector(mValueReplacedListeners);
	}
	
	/**
	 * Notifies all listeners that the value of this Value changed.
	 * @return
	 */
	public  synchronized void notifyValueChanged() {
		
		if(mNotifyCount <= 0) {
			mListenerBufferVector.clear();
			mListenerBufferVector.addAll(mValueChangedListeners);
			mNotifyCount = 0;
		}
		mNotifyCount++;
		for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
			((ValueChangedListener) i.next()).valueChanged(this);
		}
		mNotifyCount--;
	}

	/**
	 * Notifies all listeners that the value of this Value was replaced by a new one.
	 * @return
	 */
	public synchronized void notifyValueReplaced(Value newValue) {
		mListenerBufferVector.clear();
		mListenerBufferVector.addAll(mValueChangedListeners);
		for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
			((ValueChangedListener) i.next()).valueReplaced(this, newValue);
		}
		mListenerBufferVector.clear();
		mListenerBufferVector.addAll(mValueReplacedListeners);
		for(Iterator i = mListenerBufferVector.iterator(); i.hasNext();) {
			((ValueReplacedListener) i.next()).valueReplaced(this, newValue);
		}
	}
	
}
