/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.value;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.regex.PatternSyntaxException;


public class ValueManager {

	private Hashtable mValues = new Hashtable();
	
	public ValueManager() {
		
	}
	
	
	/**
	 * Returns true if the given name already is associated with
	 * a Value object.
	 * 
	 * @param name the name to check.
	 * @return true if this name is already in use.
	 */
	public boolean nameExists(String name) {
		return mValues.containsKey(name);
	}
	
	/**
	 * Adds a Value object to the ValueManager associated with a name.
	 * This name can be used to access this Value object later. The same
	 * Value object can be associated with several names by calling this
	 * method multiple times with different names as parameters. 
	 * 
	 * @param name the name this Value will be associated with.
	 * @param value the value to add to the manager.
	 * @return true if successful.
	 */
	public boolean addValue(String name, Value value) {
		if(name == null || value == null || mValues.containsKey(name)) {
			return false;
		}
		mValues.put(name, value);
		return true;
	}
	
	/**
	 * Adds a Value object to the ValueManager associated with a name.
	 * If there is already a Value that is associated with that name, 
	 * then the old Value is removed and replaced by the new one.
	 * 
	 * @param name the name this Value will be associated with.
	 * @param value the value to add to the manager.
	 * @return true if successful.
	 */
	public boolean replaceValue(String name, Value value) {
		if(name == null || value == null || !mValues.containsKey(name)) {
			return false;
		}
		Value v = (Value) mValues.put(name, value);
		if(v != null) {
			v.notifyValueReplaced(value);
		}
		return true;
	}
	
	/**
	 * Returns the Value object associated with the name given in the parameter list.
	 * @param name the associated name of the desired Value object.
	 * @return the Value object associated with the name, if no Value is found, then null is returned.
	 */
	public Value getValue(String name) {
		if(name == null || !mValues.containsKey(name)) {
			return null;
		}
		return (Value) mValues.get(name);
	}
	
	/**
	 * Returns a OBjectValue object associated with the given name.
	 * If no such object exists or if the associated Value object
	 * is not of type ObjectValue, then null is returned.
	 * @param name the name of the desired ObjectValue object.
	 * @return the desired ObjectValue if found, otherwise null.
	 */
	public ObjectValue getObjectValue(String name) {
		if(name == null || !mValues.containsKey(name)) {
			return null;
		}
		Object obj = mValues.get(name);
		if(obj != null && obj instanceof ObjectValue) {
			return (ObjectValue) obj;
		}
		return null;
	}
	
	/**
	 * Returns a DoubleValue object associated with the given name.
	 * If no such object exists or if the associated Value object
	 * is not of type DoubleValue, then null is returned.
	 * @param name the name of the desired DoubleValue object.
	 * @return the desired DoubleValue if found, otherwise null.
	 */
	public DoubleValue getDoubleValue(String name) {
		if(name == null || !mValues.containsKey(name)) {
			return null;
		}
		Object obj = mValues.get(name);
		if(obj != null && obj instanceof DoubleValue) {
			return (DoubleValue) obj;
		}
		return null;
	}
	
	/**
	 * Returns a IntValue object associated with the given name.
	 * If no such object exists or if the associated Value object
	 * is not of type IntValue, then null is returned.
	 * @param name the name of the desired IntValue object.
	 * @return the desired IntValue if found, otherwise null.
	 */
	public IntValue getIntValue(String name) {
		if(name == null || !mValues.containsKey(name)) {
			return null;
		}
		Object obj = mValues.get(name);
		if(obj != null && obj instanceof IntValue) {
			return (IntValue) obj;
		}
		return null;
	}
	
	/**
	 * Returns a BooleanValue object associated with the given name.
	 * If no such object exists or if the associated Value object
	 * is not of type IntValue, then null is returned.
	 * @param name the name of the desired BooleanValue object.
	 * @return the desired BooleanValue if found, otherwise null.
	 */
	public BooleanValue getBooleanValue(String name) {
		if(name == null || !mValues.containsKey(name)) {
			return null;
		}
		Object obj = mValues.get(name);
		if(obj != null && obj instanceof BooleanValue) {
			return (BooleanValue) obj;
		}
		return null;
	}
	
	/**
	 * Returns a SimObjectValue object associated with the given name.
	 * If no such object exists or if the associated Value object
	 * is not of type SimObjectValue, then null is returned.
	 * @param name the name of the desired SimObjectValue object.
	 * @return the desired SimObjectValue if found, otherwise null.
	 */
	public SimObjectValue getSimObjectValue(String name) {
		if(name == null || !mValues.containsKey(name)) {
			return null;
		}
		Object obj = mValues.get(name);
		if(obj != null && obj instanceof SimObjectValue) {
			return (SimObjectValue) obj;
		}
		return null;
	}
	
	/**
	 * Returns a StringValue object associated with the given name.
	 * If no such object exists or if the associated Value object
	 * is not of type StringValue, then null is returned.
	 * @param name the name of the desired StringValue object.
	 * @return the desired StringValue if found, otherwise null.
	 */
	public StringValue getStringValue(String name) {
		if(name == null || !mValues.containsKey(name)) {
			return null;
		}
		Object obj = mValues.get(name);
		if(obj != null && obj instanceof StringValue) {
			return (StringValue) obj;
		}
		return null;
	}
	
	/**
	 * Returns a PointValue object associated with the given name.
	 * If no such object exists or if the associated Value object
	 * is not of type PointValue, then null is returned.
	 * @param name the name of the desired PointValue object.
	 * @return the desired PointValue if found, otherwise null.
	 */
	public PointValue getPointValue(String name) {
		if(name == null || !mValues.containsKey(name)) {
			return null;
		}
		Object obj = mValues.get(name);
		if(obj != null && obj instanceof PointValue) {
			return (PointValue) obj;
		}
		return null;
	}
	
	/**
	 * Returns a HistoryValue object associated with the given name.
	 * If no such object exists or if the associated Value object
	 * is not of type HistoryValue, then null is returned.
	 * @param name the name of the desired HistoryValue object.
	 * @return the desired HistoryValue if found, otherwise null.
	 */
	public HistoryValue getHistoryValue(String name) {
		if(name == null || !mValues.containsKey(name)) {
			return null;
		}
		Object obj = mValues.get(name);
		if(obj != null && obj instanceof HistoryValue) {
			return (HistoryValue) obj;
		}
		return null;
	}
	
	/**
	 * Returns a Vector with all value objects. The Vector might contain duplicates.
	 * @return a vector containing all Value objects.
	 */
	public Vector getValues() {
		return new Vector(mValues.values());
	}
	
	/**
	 * Returns a Vector with all Value objects that are associated with names
	 * starting with the given namePrefix.
	 * @param namePrefix the namePrefix of all objects that will be returned.
	 * @return a Vector with all objects that are associated with names staring with namePrefix.
	 */
	public Vector getValuesStartingWith(String namePrefix) {
		Vector objects = new Vector();
		if(namePrefix == null) {
			return objects;
		}
		Enumeration names = mValues.keys();
		for(; names.hasMoreElements();) {
			String name = (String) names.nextElement();
			if(name.startsWith(namePrefix)) {
				Object obj = mValues.get(name);
				if(!objects.contains(obj)) {
					objects.add(obj);
				}
			}
		}
		return objects;
	}
	
	public Vector getValuesMatchingPattern(String regularExpression) {
		Vector objects = new Vector();
		if(regularExpression == null) {
			return objects;
		}
		Enumeration names = mValues.keys();
		for(; names.hasMoreElements();) {
			String name = (String) names.nextElement();
			try {
				if(name.matches(regularExpression)) {
					Object obj = mValues.get(name);
					if(!objects.contains(obj)) {
						objects.add(obj);
					}
				}
			} catch (PatternSyntaxException e) {
				break;
			}
		}
		return objects;
	}
	
	public Vector getValueNamesMatchingPattern(String regularExpression) {
		Vector objects = new Vector();
		if(regularExpression == null) {
			return objects;
		}
		Enumeration names = mValues.keys();
		for(; names.hasMoreElements();) {
			String name = (String) names.nextElement();
			try {
				if(name.matches(regularExpression)) {
					objects.add(name);
				}
			} catch (PatternSyntaxException e) {
				break;
			}
		}
		return objects;
	}
	
	//TODO getNamesOfObject()
	
	
	/**
	 * Returns an Enumeration with all names that are associated with Values.
	 * @return an Enumeration with all names.
	 */
	public Enumeration getValueNames() {
		return mValues.keys();
	}
	
	public static Value synchronizeValue(Value value, Value oldValue, Value newValue, 
			ValueReplacedListener owner) 
	{
		if(oldValue != null 
			&& newValue != null 
			&& oldValue == value
			&& value.getClass().isInstance(newValue))
		{
			oldValue.removeValueReplacedListener(owner);
			if(owner instanceof ValueChangedListener) {
				if(oldValue.getValueChangedListeners().contains(owner)) {
					oldValue.removeValueChangedListener((ValueChangedListener) owner);
					newValue.addValueChangedListener((ValueChangedListener) owner);
				}
			}
				
			newValue.addValueReplacedListener(owner);
			return newValue;
		}
		return value;
	}
	
	//Test purposes only.
	public void removeAllValues() {
		mValues.clear();
	}
	
	

}
 
