/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.visualization;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.util.logging.Logger;

import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;

public class SimulationDelayController implements Visualization, ValueChangedListener, 
								ChangeListener, OptimizePerformanceListener {

	private JPanel mDisplay;
	private JSlider mSlider;
	private IntValue mSimulationDelay;
	private final double mIncrementResolution = 1.5;
	private boolean mChangingDelay;
	private int mCurrentDelay;
	
	public SimulationDelayController() {
		this(SwingConstants.VERTICAL);	
	}
	
	public SimulationDelayController(int orientation) {
		mDisplay = new JPanel(new BorderLayout());
		mSlider = new JSlider(orientation, 0, 50, 0);
		mSlider.addChangeListener(this);
		mDisplay.add(BorderLayout.CENTER, mSlider);
		mSlider.setPreferredSize(new Dimension(20, 40));
		mChangingDelay = false;
		
		Simulator.getInstance().addOptimizePerformanceListener(this);
	}
	
	public boolean init() {
		boolean ok = true;
		
		acquireValueObjects();
		
		return ok;
	}
	
	private void acquireValueObjects() {
		mSimulationDelay = Simulator.getInstance().getValueManager().getIntValue("/Simulation/Delay");
		if(mSimulationDelay == null) {
			Logger.getLogger("Simulator")
				.warning("SimulationDelayController: Could not find IntValue:\n" +
						"[/Simulation/Delay]");
		}
		else {
			mSlider.setValue((int) Math.pow((double) mSimulationDelay.get(), 1.0 / mIncrementResolution));
			mCurrentDelay = mSlider.getValue();
			mSimulationDelay.addValueChangedListener(this);
		}
	}
	
	public boolean cleanUp() {
		boolean ok = true;
		
		return ok;
	}
	
	public JPanel getPanel() {
		return mDisplay;
	}
	
	public void valueChanged(Value value) {
		if(mSimulationDelay != null && value == mSimulationDelay && !mChangingDelay) {
			mSlider.setValue((int) Math.pow((double) mSimulationDelay.get(), 1.0 / mIncrementResolution));
			mCurrentDelay = mSlider.getValue();
		}
	}
	
	/**
	 * Called if the one of the registered Value objects were replaced in the ValueManager.
	 */
	public void valueReplaced(Value oldValue, Value newValue) {
		if(oldValue != newValue && oldValue != null) {
			oldValue.removeValueChangedListener(this);
			if(mSimulationDelay == oldValue) {
				acquireValueObjects();
			}
		}
	}
	
	public void stateChanged(ChangeEvent e) {
		if(mSimulationDelay != null) {
			int delay = (int) Math.pow(mSlider.getValue(), mIncrementResolution);
			
			mChangingDelay = true;
			mSimulationDelay.set(delay);
			mChangingDelay = false;
			mCurrentDelay = mSlider.getValue();
			
			Simulator.getInstance().reportStatus("Simulation delay set to " + delay + " milliseconds.");
		}
	}
	
	/**
	 * Sets the maximal delay this slider can set.
	 * @param maxDelay the maximal delay.
	 */
	public void setMaximalDelay(int maxDelay) {
		mSlider.setMaximum(maxDelay);
		if(mSimulationDelay != null && mSimulationDelay.get() > maxDelay) {
			mChangingDelay = true;
			mSimulationDelay.set(maxDelay);
			mChangingDelay = false;
		}
	}
	
	/**
	 * Returns the maximum this slider control can set.
	 * @return the maximum.
	 */
	public int getMaximalDelay() {
		return mSlider.getMaximum();
	}

	public void optimizePerformance(boolean optimize) {
		int currentDelay = mCurrentDelay;
		if(optimize) {
			mSlider.setValue(mSlider.getMinimum());
		}
		else {
			mSlider.setValue(currentDelay);
		}
		mCurrentDelay = currentDelay;
	}
	
	
	
	
}
