/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.visualization;

import java.util.Iterator;
import java.util.Vector;

import de.fhg.iais.outdoor.simba.core.Simulator;


public abstract class VisualizationArea implements Visualization, VisibilityListener {
	
	protected Vector mVisualizations = new Vector();
	protected Vector mVisibilityChangedListeners = new Vector();
	
	protected abstract boolean installVisualization(Visualization visu, String positionIdentifier);
	
	/**
	 * Adds a new Visualization to the VisualizationArea. This method calls method
	 * installVisualization which should be overridden to implement the installation
	 * of the new Visualization at a suitable position. The position is encoded 
	 * in the positionIdentifier and can be used freely.
	 * 
	 * @param visu the visualization to add.
	 * @param positionIdentifier the identifier that can be used to determin the location of the visualization.
	 * @return true if the Visualization could be added successfully.
	 */
	public boolean addVisualization(Visualization visu, String positionIdentifier) {
		if(visu == null) {
			return false;
		}
		if(!mVisualizations.contains(visu)) {
			mVisualizations.add(visu);
			Simulator.getInstance().addSystemObject(visu);
			Simulator.getInstance().getVisualizationManager().addVisualization(visu);
			if(visu instanceof VisibilityListener) {
				addVisibilityListener((VisibilityListener) visu);
			}
		}
		return installVisualization(visu, positionIdentifier);
	}
	
	/**
	 * Returns a Vector containing all Visualizations that are registered here.
	 * @return
	 */
	public Vector getVisualizations() {
		return new Vector(mVisualizations);
	}
	

	public boolean init() {
		return true;
	}
	
	public boolean cleanUp() {
		return true;
	}
	
	/**
	 * Returns true if the JDialog is visible.
	 * @return true if the dialog is visible.
	 */
	public boolean isVisible() {
		return getPanel().isVisible();
	}
	
	/**
	 * Adds a VisibilityListener. All registered listeners will be notified if the 
	 * visibility of the JDialog changed.
	 * 
	 * @param listener the listener to add.
	 * @return true if successful.
	 */
	public synchronized boolean addVisibilityListener(VisibilityListener listener) {
		if(listener == null || mVisibilityChangedListeners.contains(listener)) {
			return false;
		}
		mVisibilityChangedListeners.add(listener);
		return true;
	}
	
	/**
	 * Removes a VisibilityListener.
	 * 
	 * @param listener the listener to remove.
	 * @return true if successful.
	 */
	public synchronized boolean removeVisibilityListener(VisibilityListener listener) {
		if(listener == null || !mVisibilityChangedListeners.contains(listener)) {
			return false;
		}
		mVisibilityChangedListeners.remove(listener);
		return true;
	}
	
	/**
	 * Notifies all registered VisibilityListeners that the visibiility of the JDialog changed.
	 */
	public synchronized void notifyVisibilityChanged(boolean visible) {
		for(Iterator i = mVisibilityChangedListeners.iterator(); i.hasNext();) {
			((VisibilityListener) i.next()).visibilityChanged(visible, this);
		}
	}

	public void visibilityChanged(boolean visible, Object source) {
		notifyVisibilityChanged(visible);
	}
	
	

}
