/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.visualization.neuron;

import java.util.Iterator;
import java.util.Vector;
import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.core.SystemObject;
import de.fhg.iais.outdoor.simba.neuralnet.NeuralNet;
import de.fhg.iais.outdoor.simba.neuralnet.Neuron;
import de.fhg.iais.outdoor.simba.neuralnet.Synapse;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;


public class NeuronEditFunctions implements SystemObject, ValueChangedListener {
	
	private NeuronVisualizationEnvironment mNeuronVisualizationEnvironment;
	private NetworkViewer mNetworkViewer;
	private IntValue mControlKeys;
	private Vector mClip = new Vector();
	
	public NeuronEditFunctions(NeuronVisualizationEnvironment env, NetworkViewer viewer) {
		mNeuronVisualizationEnvironment = env;
		mNetworkViewer = viewer;
		Simulator.getInstance().addSystemObject(this);
	}

	public boolean init() {
		ValueManager vm = Simulator.getInstance().getValueManager();
		mControlKeys = vm.getIntValue("/Visualization/Neurons/Control/LastControlKey");
		
		if(mControlKeys == null) {
			Logger.getLogger("Simulator").warning("Could not find the Value for LastControlKey.");
		}
		else {
			mControlKeys.addValueChangedListener(this);
		}
		return true;
	}
	
	public boolean cleanUp() {
		return true;
	}
	
	public void copyNeurons() {
		Vector selectedObjects = mNetworkViewer.getSelectedObjects();

		mClip.clear();
		mClip.addAll(selectedObjects);
	}
	
	public void pasteNeurons() {
		Vector selectedObjects = mNetworkViewer.getSelectedObjects();
		NeuralNet net = Simulator.getInstance().getNeuralNetManager().getCurrentNet();
		
		if(mClip.isEmpty()) {
			return;
		}
		
		if(net == null) {
			return;
		}
		
		Vector clonedNeurons = new Vector();
		Vector newVisuObjects = new Vector();
		
		boolean networkChanged = false;
		synchronized(net) {
			for(Iterator i = mClip.iterator(); i.hasNext();) {
				SimObject obj = (SimObject) i.next();
				if(!obj.isActive()) {
					continue;
				}
				if(obj instanceof NeuronObject) {
					NeuronObject neuronObject = (NeuronObject) obj;
					Neuron neuron = neuronObject.getNeuron();
					if(neuron != null && neuron.getType().equals(Neuron.HIDDEN)) {
						clonedNeurons.add(neuron);
						NeuronObject no = new NeuronObject(neuronObject.getPosition().x,
								neuronObject.getPosition().y, neuronObject.getAngle());
						newVisuObjects.add(no);
					}
				}
			}
			if(clonedNeurons.isEmpty()) {
				return;
			}
			clonedNeurons = NeuralNet.cloneNeuronGroup(clonedNeurons);
			if(clonedNeurons.size() != newVisuObjects.size()) {
				Logger.getLogger("Simulator").warning("Problem during copy task!");
			}
			for(int i = 0; i < clonedNeurons.size(); ++i) {
				NeuronObject no = (NeuronObject) newVisuObjects.elementAt(i);
				Neuron n = (Neuron) clonedNeurons.elementAt(i);
				no.setNeuron(n);
				no.setActive(true);
				no.setMarked(true);
				net.addNeuron(n);
				mNeuronVisualizationEnvironment.addSimObject(no);
				networkChanged = true;
			}
		}
		if(networkChanged) {
			net.updateNetwork();
			Simulator.getInstance().getNeuralNetManager().notifyNetworkStructureChanged(net);
			for(Iterator i = selectedObjects.iterator(); i.hasNext();) {
				((SimObject) i.next()).setMarked(false);
			}
			mNetworkViewer.setSelectedObjects(newVisuObjects);
		}
	}
	
	public void deleteObjects() {
		
		Vector selectedObjects = mNetworkViewer.getSelectedObjects();
		if(selectedObjects.isEmpty()) {
			return;
		}
		
		NeuralNet net = Simulator.getInstance().getNeuralNetManager().getCurrentNet();
		if(net == null) {
			return;
		}
		
		boolean networkChanged = false;
		synchronized(net) {
			
			for(Iterator i = selectedObjects.iterator(); i.hasNext();) {
				SimObject obj = (SimObject) i.next();
				if(!obj.isActive()) {
					continue;
				}
				if(obj instanceof NeuronObject) {
					Neuron neuron = ((NeuronObject) obj).getNeuron();
					if(neuron != null && neuron.getType().equals(Neuron.HIDDEN)) {
						net.removeNeuron(neuron);
						networkChanged = true;
					}
				}
				else if(obj instanceof SynapseObject) {
					Synapse synapse = ((SynapseObject) obj).getSynapse();
					if(synapse != null && synapse.getTargetNeuron() != null) {
						Neuron target = synapse.getTargetNeuron();
						target.removeSynapse(synapse);
						networkChanged = true;
					}
				}
			}
		}
		if(networkChanged) {
			net.updateNetwork();
			Simulator.getInstance().getNeuralNetManager().notifyNetworkStructureChanged(net);
			for(Iterator i = selectedObjects.iterator(); i.hasNext();) {
				((SimObject) i.next()).setMarked(false);
			}
			mNetworkViewer.setSelectedObjects(new Vector());
		}
	}
	
	public void arrangeHorizontally() {
		Vector selectedObjects = mNetworkViewer.getSelectedObjects();
		
		double furthestXPosition = -1 * Double.MAX_VALUE;
		boolean hasRelevantObjects = false;
		for(Iterator i = selectedObjects.iterator(); i.hasNext();) {
			SimObject obj = (SimObject) i.next();
			if(!obj.isActive()) {
				continue;
			}
			if((obj instanceof NeuronObject) || (obj instanceof SynapseObject)) {
				hasRelevantObjects = true;
				if(furthestXPosition < obj.getPosition().x) {
					furthestXPosition = obj.getPosition().x;
				}
			}
		}
		if(hasRelevantObjects && furthestXPosition > (-1 * Double.MAX_VALUE)) {
			for(Iterator i = selectedObjects.iterator(); i.hasNext();) {
				SimObject obj = (SimObject) i.next();
				if(!obj.isActive()) {
					continue;
				}
				if((obj instanceof NeuronObject) || (obj instanceof SynapseObject)) {
					obj.setPosition(furthestXPosition, obj.getPosition().y);
				}
			}
		}
	}
	
	public void arrangeVertically() {
		Vector selectedObjects = mNetworkViewer.getSelectedObjects();
		
		double furthestYPosition = (-1 * Double.MAX_VALUE);
		boolean hasRelevantObjects = false;
		for(Iterator i = selectedObjects.iterator(); i.hasNext();) {
			SimObject obj = (SimObject) i.next();
			if(!obj.isActive()) {
				continue;
			}
			if((obj instanceof NeuronObject) || (obj instanceof SynapseObject)) {
				hasRelevantObjects = true;
				if(furthestYPosition < obj.getPosition().y) {
					furthestYPosition = obj.getPosition().y;
				}
			}
		}
		if(hasRelevantObjects && furthestYPosition > (-1 * Double.MAX_VALUE)) {
			for(Iterator i = selectedObjects.iterator(); i.hasNext();) {
				SimObject obj = (SimObject) i.next();
				if(!obj.isActive()) {
					continue;
				}
				if((obj instanceof NeuronObject) || (obj instanceof SynapseObject)) {
					obj.setPosition(obj.getPosition().x, furthestYPosition);
				}
			}
		}
	}
	
	public void moveObjects(double xIncr, double yIncr) {
Vector selectedObjects = mNetworkViewer.getSelectedObjects();

		for(Iterator i = selectedObjects.iterator(); i.hasNext();) {
			SimObject obj = (SimObject) i.next();
			if(!obj.isActive()) {
				continue;
			}
			if((obj instanceof NeuronObject) || (obj instanceof SynapseObject)) {
				obj.setPosition(obj.getPosition().x + xIncr, obj.getPosition().y + yIncr);
			}
		}
	}

	public void valueChanged(Value value) {
		if(value == null) {
			return;
		}
		System.out.println("key: " + mControlKeys.get());
		if(value == mControlKeys) {
			switch(mControlKeys.get()) {
				case 3:
					copyNeurons();
					break;
				case 22:
					pasteNeurons();
					break;
				case 4:
					deleteObjects();
					break;
				case 12:
					arrangeHorizontally();
					break;
				case 8:
					arrangeVertically();
					break;
				case 26:
					moveObjects(-1.0, 0.0);
					break;
				case 21:
					moveObjects(0.0, 1.0);
					break;
				case 9:
					moveObjects(0.0, -1.0);
					break;
				case 15:
					moveObjects(1.0, 0.0);
					break;
				case 23:
					moveObjects(-10.0, 0.0);
					break;
				case 5:
					moveObjects(0.0, 10.0);
					break;
				case 18:
					moveObjects(0.0, -10.0);
					break;
				case 20:
					moveObjects(10.0, 0.0);
					break;
				default:
			}
		}
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		mControlKeys = (IntValue) ValueManager.synchronizeValue(mControlKeys, oldValue, newValue, this);
	}
	
	
	

}
