/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.visualization.neuron;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.text.NumberFormat;
import java.util.Locale;
import java.util.logging.Logger;

import de.fhg.iais.outdoor.simba.core.EnvironmentModel;
import de.fhg.iais.outdoor.simba.core.SimObject;
import de.fhg.iais.outdoor.simba.devices.Box;
import de.fhg.iais.outdoor.simba.devices.TextLabel;
import de.fhg.iais.outdoor.simba.neuralnet.Synapse;
import de.fhg.iais.outdoor.simba.value.BooleanValue;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.PointValue;
import de.fhg.iais.outdoor.simba.value.Value;

public class SynapseObject extends SimObject {
	
	private NeuronObject mSourceNeuron;
	private NeuronObject mTargetNeuron;
	private Synapse mSynapse;
	private double mThickness;
	private Polygon mArrowShape;
	private TextLabel mTextLabel;
	private DoubleValue mSynapseStrengthValue;
	private PointValue mSourcePosition;
	private PointValue mTargetPosition;
	private Box mSelectedBox;
	private BooleanValue mEnabledValue;
	
	public SynapseObject(double x, double y, double angle) {
		super("Synapse", x, y, angle, new Polygon(), false);
		
		mTextLabel = new TextLabel("SynapseStrengthText", 0, 0, 0, "", 6, false);
		mTextLabel.setLayer(6);
		mTextLabel.setReferenceObject(this);
		
		setShape(mTextLabel.getShape());
		setShapeColor(Color.GREEN);
		setFillShape(true);
		setLayer(-1);
		
		mArrowShape = new Polygon();
		mArrowShape.addPoint(-15, -7);
		mArrowShape.addPoint(0, 0);
		mArrowShape.addPoint(-15, 7);
		
		mSelectedBox = new Box("MarkedBox", 0.0, 0.0, 0.0, 40.0, 20.0, false);
		mSelectedBox.setReferenceObject(this);
		mSelectedBox.setLayer(50);
		mSelectedBox.setActive(false);
		mSelectedBox.setShapeTransparency(0.3);
		mSelectedBox.setShapeColor(Color.RED);
		mMarked.addValueChangedListener(this);

		setActive(false);
	}

	public void addChildren(EnvironmentModel model) {
		super.addChildren(model);
		model.addSimObject(mTextLabel);
		model.addSimObject(mSelectedBox);
	}

	public boolean init() {
		boolean ok = super.init();
		
		return ok;
	}

	public void reset() {
		super.reset();
	}

	public synchronized boolean update() {
		boolean ok = super.update();

		if(mSynapse != null && mTargetNeuron != null && mSourceNeuron != null) {
			double strength = mSynapse.getStrength();
			mThickness = Math.max(1.0, Math.min(4.0, Math.abs(strength)));
			String value = NumberFormat.getNumberInstance(Locale.GERMAN).format(strength);
			mTextLabel.setText(value);
			if(strength >= 0.0) {
				setShapeColor(Color.GREEN);
			}
			else {
				setShapeColor(Color.CYAN);
			}
			if(mEnabledValue != null) {
				mTextLabel.setCrossedOut(!mEnabledValue.get());
			}
		}
		
		return ok;
	}
	
	public synchronized void centerPosition() {
		if(mTargetNeuron != null && mSourceNeuron != null && isActive()) {
			
			mTargetNeuron.invalidateCalculationBuffers();
			mSourceNeuron.invalidateCalculationBuffers();
			invalidateCalculationBuffers();
			
			Point2D.Double p1 = mTargetNeuron.getGlobalPosition();
			Point2D.Double p2 = mSourceNeuron.getGlobalPosition();
			
			double dist = p1.distance(p2);
			double angle = Math.atan2(p2.y - p1.y, p2.x - p1.x);
			
			
			double x = (Math.cos(angle) * 0.5 * dist) + p1.x;
			double y = (Math.sin(angle) * 0.5 * dist) + p1.y;
			
			if(mTargetNeuron == mSourceNeuron) {
				y += 70.0;
				x -= 10.0;
			}
			
			Point2D.Double p = new Point2D.Double(x, y);
			
			setPosition(p);
			
		}
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		super.valueReplaced(oldValue, newValue);
	}
	
	

	public void valueChanged(Value value) {
		super.valueChanged(value);
		if(value == null) {
			return;
		}
		if(value == mSynapseStrengthValue) {
			update();
		}
		else if(value == mTargetPosition || value == mSourcePosition) {
			//TODO
		}
		else if(value == mMarked) {
			mSelectedBox.setActive(mMarked.get());
		}
		else if(value == mEnabledValue) {
			mTextLabel.setCrossedOut(!mEnabledValue.get());
		}
	}

	public synchronized void setSynapse(Synapse synapse, NeuronObject sourceNeuron, NeuronObject targetNeuron) {
		if(mSynapseStrengthValue != null) {
			mSynapseStrengthValue.removeValueChangedListener(this);
			mSynapseStrengthValue = null;
		}
		if(mEnabledValue != null) {
			mEnabledValue.removeValueChangedListener(this);
			mEnabledValue = null;
		}
		mSynapse = synapse;
		if(mSourcePosition != null) {
			mSourcePosition.removeValueChangedListener(this);
			mSourcePosition = null;
		}
		mSourceNeuron = sourceNeuron;
		if(mTargetPosition != null) {
			mTargetPosition.removeValueChangedListener(this);
			mTargetPosition = null;
		}
		mTargetNeuron = targetNeuron;
		setActive(mSourceNeuron != null && mTargetNeuron != null && mSynapse != null);
		if(mSynapse != null && mSourceNeuron != null && mTargetNeuron != null) {
			mName = "Synapse-" + mSourceNeuron.getNeuron().getName()
					+ "-" + mTargetNeuron.getNeuron().getName();
			mSourcePosition = sourceNeuron.getPositionValue();
			mSourcePosition.addValueChangedListener(this);
			mTargetPosition = targetNeuron.getPositionValue();
			mTargetPosition.addValueChangedListener(this);
			mSynapseStrengthValue = mSynapse.getStrengthValue();
			mSynapseStrengthValue.addValueChangedListener(this);
			mEnabledValue = mSynapse.getEnabledValue();
			mEnabledValue.addValueChangedListener(this);
			update();
		}
	}
	
	public NeuronObject getSourceNeuron() {
		return mSourceNeuron;
	}
	
	public NeuronObject getTargetNeuron() {
		return mTargetNeuron;
	}
	
	public Synapse getSynapse() {
		return mSynapse;
	}
	
	public void setActive(boolean active) {
		super.setActive(active);
		mTextLabel.setActive(active);
		if(!active) {
			mSelectedBox.setActive(false);
		}
	}

	public synchronized void paintSelf(Graphics2D g) {
		if(mSynapse != null && mTargetNeuron != null && mSourceNeuron != null) {
			Color c = g.getColor();
			Composite composite = g.getComposite();
			g.setColor(mShapeColor);
			g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float) mTransparency.get()));
			
			AffineTransform transformation = getAffineTransform();
			try {
				transformation = transformation.createInverse();
			}
			catch(NoninvertibleTransformException ex) {
				Logger.getLogger("Simulator")
					.warning("DistanceToPointScanner.paintSelf: Could not invert AffineTransform.");
				g.setColor(c);
				return;
			}
			Stroke oldStroke = g.getStroke();
			g.setStroke(new BasicStroke((int) mThickness));
			
			
			
			Point2D.Double object = new Point2D.Double();
			Point2D.Double ownPosition = new Point2D.Double(0.0, 0.0);
			if(mTargetNeuron == mSourceNeuron) {
				ownPosition.x = 30.0;
				ownPosition.y = 0.0;
			}
			transformation.transform(mSourceNeuron.getGlobalPosition(), object);
			g.drawLine((int) ownPosition.x, (int) ownPosition.y, (int) Math.round(object.x), (int) Math.round(object.y));
			transformation.transform(mTargetNeuron.getGlobalPosition(), object);
			g.drawLine(0, 0, (int) Math.round(object.x), (int) Math.round(object.y));
			if(mTargetNeuron == mSourceNeuron) {
				g.drawLine(0, 0, 30, 0);
			}
			
			
			g.setStroke(oldStroke);
			
			double angle = Math.atan2(object.y, object.x);
			double distance = object.distance(0.0, 0.0);
			
			AffineTransform transform = g.getTransform();
			transform.rotate(angle);
			distance = Math.max(distance - 35.0, 15.0);
			transform.translate(distance, 0);
			g.setTransform(transform);
			g.fillPolygon(mArrowShape);
			transform.translate(-distance, 0);
			transform.rotate(-angle);
			g.setTransform(transform);
			
			
//			g.setColor(Color.green);
//			g.fillPolygon(mShape);
			g.setColor(Color.BLACK);

			transform.scale(1.0, -1.0);
			g.setTransform(transform);
			
//			NumberFormat.getInstance(Locale.GERMAN).setMinimumFractionDigits(1);
//			NumberFormat.getInstance(Locale.GERMAN).setMaximumFractionDigits(2);
//			String strength = NumberFormat.getInstance(Locale.GERMAN).format(mSynapse.getStrength());
//			g.drawChars(strength.toCharArray(), 0, strength.length(), -20, 5);
			
			
			
			g.setComposite(composite);
			g.setColor(c);
		}
	}
	
	

}
