/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.visualization.statistics;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.JPanel;

import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.HistoryValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueReplacedListener;
import de.fhg.iais.outdoor.simba.visualization.VisibilityListener;


/**
 * The PlotPanel is able to plot a number of HistoryValues in 
 * different colors. The names of the plotted HistoryValues is displayed in the left upper 
 * corner in the color the plot of this HistoryValue is painted. 
 * 
 * 
 * By clicking at any position in the StatisticsPlotPanel the index of the corresponding
 * generation and the value that corresponds to the y position of the clicked location, 
 * will be written to the internal Value objects mSelectedYAxisValue and mSelectedIndexValue.
 * These values can be observed to react on changes of these Values.
 * 
 * The range to plot can be specified manually. If the plot mode is set to autoRange, then
 * the upper bound of the plotted range is automatically set to the highest possible index.
 * The lower bound is unaffected of the autoRange mode.
 * 
 * The scaling can also be specified manually. If the plot mode is set to autoScale, then the 
 * scale is automatically choosen so that all values of all histories will fit exactly to the 
 * plot panel.
 * 
 * 
 * @author Christian Rempis
 *
 */
public class PlotDisplay extends JPanel implements ValueReplacedListener, VisibilityListener,
					MouseListener {

	public static final long serialVersionUID = 123L; 
	private double mUpperBound;
	private double mLowerBound;
	private IntValue mFirstIndexToPlot;
	private IntValue mLastIndexToPlot;
	private boolean mAutoRange;
	private boolean mAutoScale;
	private double mScale;
	private Hashtable mHistoryColors = new Hashtable();
	private Hashtable mHistoryNames = new Hashtable();
	private Vector mHistories = new Vector();
	private Vector mColors = new Vector();
	private boolean mVisible;
	
	private double mLastIndexDistance;
	private double mLastUpperBound;
	private double mLastLowerBound;
	
	private DoubleValue mSelectedYAxisValue;
	private IntValue mSelectedIndexValue;
	
	public PlotDisplay() {
		setBorder(BorderFactory.createEtchedBorder());
		
		mLowerBound = 1.0;
		mUpperBound = 0.0;
		mScale = 1.0;
		
		mColors.add(Color.GREEN);
		mColors.add(Color.RED);
		mColors.add(Color.BLUE);
		mColors.add(Color.BLACK);
		mColors.add(Color.CYAN);
		mColors.add(Color.MAGENTA);
		mColors.add(Color.YELLOW);
		mColors.add(Color.PINK);
		
		mAutoScale = true;
		mAutoRange = true;
		
		mFirstIndexToPlot = new IntValue(0);
		mLastIndexToPlot = new IntValue(0);
		mUpperBound = 0.1;
		mLowerBound = 0.0;
		mVisible = true;
		
		mSelectedIndexValue = new IntValue();
		mSelectedYAxisValue = new DoubleValue();
		
		addMouseListener(this);
	}
	
	public synchronized boolean addHistory(HistoryValue history, String name, Color color) {
		if(history == null || name == null || mHistoryNames.containsKey(history)) {
			return false;
		}
		mHistoryNames.put(history, name);
		mHistories.add(history);
		if(color == null) {
			color = (Color) mColors.elementAt(mHistories.size() % mColors.size());
		}
		mHistoryColors.put(history, color);
		history.addValueReplacedListener(this);
		drawContent();
		return true;
	}
	
	public synchronized boolean removeHistory(HistoryValue history) {
		if(history == null || !mHistoryNames.containsKey(history)) {
			return false;
		}
		mHistoryNames.remove(history);
		mHistories.remove(history);
		mHistoryColors.remove(history);
		history.removeValueReplacedListener(this);
		drawContent();
		return true;
	}
	
	public synchronized void clearPlotter() {
		mHistories.clear();
		mHistoryNames.clear();
		mHistoryColors.clear();
		drawContent();
	}
	
	public void setScale(double scale) {
		mScale = scale;
	}
	
	public void setAutoScale(boolean autoScale) {
		mAutoScale = autoScale;
		drawContent();
	}
	
	public void setAutoRange(boolean autoRange) {
		mAutoRange = autoRange;
		drawContent();
	}
	
	public void setRangeToPlot(int min, int max) {
		if(min >= 0 && min <= max) {
			if(mFirstIndexToPlot.get() != min) {
				mFirstIndexToPlot.set(min);
			}
			if(mLastIndexToPlot.get() != max) {
				mLastIndexToPlot.set(max);
			}
			mAutoRange = false;
		}
		else {
			mFirstIndexToPlot.notifyValueChanged();
			mLastIndexToPlot.notifyValueChanged();
		}
		drawContent();
	}
	
	public IntValue getFirstIndexValue() {
		return mFirstIndexToPlot;
	}
	
	public IntValue getLastIndexValue() {
		return mLastIndexToPlot;
	}
	
	public IntValue getSelectedIndexValue() {
		return mSelectedIndexValue;
	}
	
	public DoubleValue getSelectedYAxisValue() {
		return mSelectedYAxisValue;
	}

	
	/**
	 * Triggers a repaint, if this PlotDisplay is visible.
	 *
	 */
	public void drawContent() {
		if(mVisible) {
			repaint();
		}
	}
	
	
	protected synchronized void paintComponent(Graphics g) {
		
		if(!mVisible) {
			return;
		}
		
		Graphics2D g2d = (Graphics2D) g;
		Color c = g2d.getColor();

		//calcualte current upper and lower bound
		if(mAutoScale) {
			mUpperBound = Double.MIN_VALUE;
			mLowerBound = Double.MAX_VALUE;
			for(Iterator i = mHistories.iterator(); i.hasNext();) {
				HistoryValue value = (HistoryValue) i.next();
				synchronized (value) {
					Vector values = value.getHistory();
					if(mAutoRange || mLastIndexToPlot.get() > values.size() - 2) { //TODO check if dangerous (if one of the values is too small or too big.
						int lastIndex = Math.max(0, values.size() - 2);
						if(lastIndex != mLastIndexToPlot.get()) {
							mLastIndexToPlot.set(lastIndex);
						}
					}
					for(int j = 0; j < values.size() - 1; ++j) {
						if(j < mFirstIndexToPlot.get() || j > mLastIndexToPlot.get()) {
							continue;
						}
						Double d = (Double) values.elementAt(j);
						if(d == null) {
							continue;
						}
						double dv = d.doubleValue();
						if(dv > mUpperBound) {
							mUpperBound = dv;
						}
						if(dv < mLowerBound) {
							mLowerBound = dv;
						}
					}
				}
			}
			//make sure that the bounds are set.
			if(mUpperBound == Double.MIN_VALUE) {
				mUpperBound = 1.0;
			}
			if(mLowerBound == Double.MAX_VALUE) {
				mLowerBound = 0.0;
			}
		}
		//avoid to have the same values for the upper and lower bound (avoid to divide by zero) 
		if(mUpperBound == mLowerBound) {
			mUpperBound += 0.1;
		}
		mLastLowerBound = mLowerBound;
		mLastUpperBound = mUpperBound;

		paintHistory(g2d);
		
		g2d.setColor(c);

	}
	
	private void paintHistory(Graphics2D g2d) {

		//Index range
		int firstIndex = mFirstIndexToPlot.get();
		int lastIndex = mLastIndexToPlot.get();
		
		//Clear area.
		g2d.setColor(Color.WHITE);
		g2d.fillRect(0, 0, getWidth(), getHeight());

		int index = 0;

		//Draw the raster of the diagram.
		double verticalScale = ((double) getHeight() - 20) / 2.0;
		g2d.setColor(Color.LIGHT_GRAY);
		g2d.drawLine(0, 9, getWidth(), 9);
		g2d.drawLine(0, (int) (verticalScale) + 10, getWidth(), (int) (1 * verticalScale) + 10);
		g2d.drawLine(0, (int) (2 * verticalScale) + 11, getWidth(), (int) (2 * verticalScale) + 11);
		g2d.drawLine(0, (int) (0.5 * verticalScale) + 10, getWidth(), (int) (0.5 * verticalScale) + 10);
		g2d.drawLine(0, (int) (1.5 * verticalScale) + 10, getWidth(), (int) (1.5 * verticalScale) + 10);
		
		
		//Plot the history of each neuron in the color of that neuron.
		for(Iterator h = mHistories.iterator(); h.hasNext();) {
			HistoryValue val = (HistoryValue) h.next();
			
			synchronized(val) {
				Collection p = val.getHistory();
			
				g2d.setColor((Color) mHistoryColors.get(val));				
				double pointDist = ((double) getWidth() - 6) / ((double) (Math.min(p.size() - 2, lastIndex) - firstIndex));
				mLastIndexDistance = pointDist;
				
				int pointId = 0;
				Point point = new Point(0, 0);
				int j = 0;
				int drawnPoints = 0;
				for(Iterator i = p.iterator(); i.hasNext() && j < p.size() - 1;) {
					Double d = (Double) i.next();
					Point nextPoint = null;
					if(d == null) {
						//ignore missing values (HistoryValue adds a null for missing values).
						nextPoint = point;
						if(j >= firstIndex && j <= lastIndex) {
							//incmrenet drawn points to fill the gap if a value is missing.
							drawnPoints++;
						}
					}
					else {
						nextPoint = new Point(2 + (int) (drawnPoints * pointDist), (int) (((((-1 + (((2.0 / (mUpperBound - mLowerBound)) * (d.doubleValue() - mLowerBound)))) * mScale) - 1.0) * -1) * verticalScale) + 10);
						if(j >= firstIndex && j <= lastIndex) {
							//Draw a dot for each value.
							g2d.fillOval(nextPoint.x - 1, nextPoint.y - 1, 3, 3);
							drawnPoints++;
						}
					}
					if(pointId != 0 && j >= firstIndex && j <= lastIndex) {
						//Draw a line between each dots.
						g2d.drawLine(point.x, point.y, nextPoint.x, nextPoint.y);
					}
					j++;
					pointId++;
					point = nextPoint;
					
				}
			}
			index++;
			
			
			//Plot the names of the HistoryValues in thir colors.
			index = 0;
			for(Iterator hist = mHistories.iterator();hist.hasNext();) {
				HistoryValue history = (HistoryValue) hist.next();
			
				g2d.setColor((Color) mHistoryColors.get(history));

				String name = (String) mHistoryNames.get(history);
				int offset = Math.max(0, name.length() - 25);
				g2d.drawChars(name.toCharArray(), offset, name.length() - offset, 10, 25 + (15 * index));
				
				index++;
			}
			
			g2d.setColor(Color.BLACK);
			//Plot the vertical range
			g2d.setColor(Color.BLACK);
			double upperBound = (double) (mUpperBound / mScale);
			double lowerBound = (double) (mLowerBound / mScale);
			double midLine = ((double) (mLowerBound + ((double) (mUpperBound - mLowerBound) / 2.0))) / mScale;
			String upper = Double.toString(upperBound);
			String lower = Double.toString(lowerBound);
			String mid = Double.toString(midLine);
			int upperMaxSize = upperBound >= 0.0 ? 4 : 5;
			int midMaxSize = midLine >= 0.0 ? 4 : 5;
			int lowerMaxSize = lowerBound >= 0.0 ? 4 : 5;
			g2d.drawChars(upper.toCharArray(), 0, Math.min(upperMaxSize, upper.length()), 5, 13);
			g2d.drawChars(mid.toCharArray(), 0, Math.min(midMaxSize, mid.length()), 5, (int) (verticalScale) + 14);
			g2d.drawChars(lower.toCharArray(), 0, Math.min(lowerMaxSize, lower.length()), 5, (int) (2 * verticalScale) + 14);
		}

		
		
	}

//	/**
//	 * Whenever one of the HistoryValues changes, then the auto range flag is 
//	 * set, so the upper bound of the horizional range is automatically readjusted
//	 * to the highest index.
//	 */
//	public void valueChanged(Value value) {
//		mAutoRange = true;
//		repaint();
//	}

	public void valueReplaced(Value oldValue, Value newValue) {
		//TODO
	}

	/**
	 * Called when this plotter becomes visible or invisible.
	 * If invisible, then all computationally expensive calculations are skipped.
	 */
	public void visibilityChanged(boolean visible, Object source) {
		mVisible = visible;
		if(visible) {
			repaint();
		}
	}

	public void mouseClicked(MouseEvent e) {
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void mousePressed(MouseEvent e) {
		Point pos = e.getPoint();
		if(mLastIndexDistance != 0) {
			mSelectedIndexValue.set(((int) ((((double) pos.x) - 2.0 + (mLastIndexDistance / 2.0)) / mLastIndexDistance)) 
					+ mFirstIndexToPlot.get());
			double verticalScale = ((double) getHeight() - 20) / 2.0;
			if(mLastLowerBound != mLastUpperBound && mScale != 0.0 && verticalScale != 0.0) {
				mSelectedYAxisValue.set((((((((((double) pos.y) - 10.0) /  verticalScale) * -1) + 1.0) / mScale) + 1) 
						/ (2.0 / (mLastUpperBound - mLastLowerBound))) + mLastLowerBound);
			}
		}
	}

	public void mouseReleased(MouseEvent e) {
	}
	
	
	
	
	
	
}
