/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.visualization.statistics;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.HistoryValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.visualization.VisibilityListener;
import de.fhg.iais.outdoor.simba.visualization.Visualization;


/**
 * This Visualization can be used to plot all available statistic Values in the system.
 * The StatisticsHistoryPlotPanel automatically aquires all available statistic Values
 * and provides them in a JList object where the user can select one or more of these 
 * Values to plot them. 
 * 
 * The embedded StatisticsPlotPanel provides two text fields where the user can specify
 * the plotted range by setting a lower and an upper bound. The upper bound will be reset
 * automatically to the current highest index after each generation. If a user wants to 
 * investigate a lower range in detail, then the system should be set to pause mode to 
 * prevent the system from readjusting the upper bound when completing the next generation.
 * 
 * By clicking at any position in the StatisticsPlotPanel the index of the corresponding
 * generation and the value that corresponds to the y position of the clicked location, 
 * will be displayed at the left bottom of the panel.
 * 
 * The remove all button removes all observed Values from the plotter.
 * 
 * @author Christian Rempis
 *
 */
public class StatisticsHistoryPlotPanel implements Visualization, VisibilityListener {

	protected static int mHistoryPlotCounter = 0;
	
	private JPanel mDisplay;
	private StatisticsPlotPanel mPlotPanel;
	private JList mHistorySelector;
	private JScrollPane mScrollPane;
	private Vector mNameList = new Vector();
	private Vector mValueList = new Vector();
	private JButton mRemoveAllButton;
	private int mId;
	
	
	public StatisticsHistoryPlotPanel() {
	
		mId = mHistoryPlotCounter++;
		
		mDisplay = new JPanel(new BorderLayout());
		mDisplay.setBorder(BorderFactory.createRaisedBevelBorder());
		
		mPlotPanel = new StatisticsPlotPanel();
		mHistorySelector = new JList();
		
		mScrollPane = new JScrollPane(mHistorySelector);
		mScrollPane.setPreferredSize(new Dimension(150, 50));
		
		JPanel controlPanel = new JPanel(new BorderLayout());
		controlPanel.add(mScrollPane, BorderLayout.CENTER);
		
		mDisplay.add(mPlotPanel, BorderLayout.CENTER);
		mDisplay.add(controlPanel, BorderLayout.EAST);

		mHistorySelector.addListSelectionListener(new ListSelectionListener() {

			public void valueChanged(ListSelectionEvent e) {
				Object[] names = mHistorySelector.getSelectedValues();
				Vector selectedNames = new Vector();
				for(int i = 0; i < names.length; ++i) {
					selectedNames.add(names[i]);
				}
				for(int i = 0; i < mNameList.size(); ++i) {
					String name = (String) mNameList.elementAt(i);
					if(!selectedNames.contains(name)) {
						mPlotPanel.getPlotDisplay().removeHistory((HistoryValue) mValueList.elementAt(i));
					}
				}
				for(int i = 0; i < mNameList.size(); ++i) {
					String name = (String) mNameList.elementAt(i);
					String displayedName = name.substring(0, name.indexOf("  -"));
					if(selectedNames.contains(name)) {
						mPlotPanel.getPlotDisplay().addHistory((HistoryValue) mValueList.elementAt(i), displayedName, null); 
					}
				}
			}
		});
		
		mRemoveAllButton = new JButton("Remove All");
		controlPanel.add(mRemoveAllButton, BorderLayout.SOUTH);
		mRemoveAllButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				mPlotPanel.getPlotDisplay().clearPlotter();
				mHistorySelector.setSelectedIndices(new int[0]);
			}
		});
		
		
	}
	
	
	public JPanel getPanel() {
		return mDisplay;
	}

	public boolean cleanUp() {
		
		Object[] selectedNames = mHistorySelector.getSelectedValues();
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(Simulator.getInstance().getConfigDirectoryName() + "historyplot" + mId + ".conf"));
			for(int i = 0; i < selectedNames.length; ++i) {
				writer.write((String) selectedNames[i] + "\n");
			}
			writer.close();
		} catch (IOException e) {
		}
		
		mValueList.clear();
		mNameList.clear();
		return true;
	}

	public boolean init() {
		
		//Collect all statistics Values and their Names and initialize the 
		//JList with these value names.
		ValueManager vm = Simulator.getInstance().getValueManager();
		Vector valueNames = vm.getValueNamesMatchingPattern("/Statistics/.*");
		for(Iterator i = valueNames.iterator(); i.hasNext();) {
			String name = (String) i.next();
			Value value = vm.getHistoryValue(name);
			if(value != null) {
				int valueNameIndex = name.lastIndexOf("/");
				if(valueNameIndex < name.length() + 1) {
					valueNameIndex++;
				}
				String valueName = name.substring(valueNameIndex);
				String statisticsName = name.substring(12);
				if(statisticsName.indexOf("/") != -1) {
					statisticsName = statisticsName.substring(0, statisticsName.indexOf("/"));
				}
				//sort
				boolean addedValue = false;
				for(int j = 0; j < mNameList.size(); ++j) {
					String localName = (String) mNameList.elementAt(j);
					if(valueName.compareToIgnoreCase(localName) < 0) {
						mNameList.add(j, valueName + "   -   [" + statisticsName + "]");
						mValueList.add(j, value);
						addedValue = true;
						break;
					}
				}
				if(!addedValue) {
					mNameList.add(valueName + "   -   [" + statisticsName + "]");
					mValueList.add(value);
				}
				
			}
		}
		mHistorySelector.setListData(mNameList);
		
		
		//Read last selected history values.
		Vector selectedNames = new Vector();
		try {
			BufferedReader reader = new BufferedReader(
					new FileReader(Simulator.getInstance().getConfigDirectoryName() + "historyplot" + mId + ".conf"));
			String name = reader.readLine();
			while(name != null) {
				selectedNames.add(name);
				name = reader.readLine();
			}
			reader.close();
		} catch (FileNotFoundException e) {
		} catch (IOException e) {
		}
		Vector selectedIndices = new Vector();
		for(int i = 0; i < mNameList.size(); ++i) {
			String name = (String) mNameList.elementAt(i);
			for(Iterator j = selectedNames.iterator(); j.hasNext();) {
				if(((String) j.next()).equals(name)) {
					selectedIndices.add(new Integer(i));
					break;
				}
			}
		}
		int[] indices = new int[selectedIndices.size()];
		for(int i = 0; i < selectedIndices.size(); ++i) {
			indices[i] = ((Integer) selectedIndices.elementAt(i)).intValue();
		}
		mHistorySelector.setSelectedIndices(indices);

		
		mScrollPane.revalidate();
		return true;
	}


	public void visibilityChanged(boolean visible, Object source) {
		mPlotPanel.visibilityChanged(visible, source);
	}
	
	

	
}
