/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.visualization.statistics;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import de.fhg.iais.outdoor.simba.core.GenerationListener;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.DoubleValue;
import de.fhg.iais.outdoor.simba.value.IntValue;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.visualization.VisibilityListener;

/**
 * The StatisticsPlotPanel uses a PlotDisplay object to plot a number of HistoryValues in 
 * different colors. The names of the plotted HistoryValues is displayed in the left upper corner in the color
 * the plot of this HistoryValue is painted. 
 * 
 * The Panel provides two text fields where the user can specify
 * the plotted range by setting a lower and an upper bound. The upper bound will be reset
 * automatically to the current highest index after each generation. If a user wants to 
 * investigate a lower range in detail, then the system should be set to pause mode to 
 * prevent the system from readjusting the upper bound when completing the next generation.
 * 
 * By clicking at any position in the StatisticsPlotPanel the index of the corresponding
 * generation and the value that corresponds to the y position of the clicked location, 
 * will be displayed at the left bottom of the panel.
 * 
 * @author Christian Rempis
 *
 */
public class StatisticsPlotPanel extends JPanel implements ValueChangedListener, 
					VisibilityListener, GenerationListener {
	
	
	public static final long serialVersionUID = 1235L;
	
	private PlotDisplay mDisplay;
	private JLabel mRangeLabel;
	private JTextField mMinIndex;
	private JTextField mMaxIndex;
	private JLabel mSelectedPositionLabel;
	private IntValue mFirstIndexValue;
	private IntValue mLastIndexValue;
	private DoubleValue mSelectedYAxisValue;
	private IntValue mSelectedIndexValue;

	public StatisticsPlotPanel() {
		mDisplay = new PlotDisplay();
		
		//Connect with the range Values of the plotter.
		mFirstIndexValue = mDisplay.getFirstIndexValue();
		mLastIndexValue = mDisplay.getLastIndexValue();
		mSelectedYAxisValue = mDisplay.getSelectedYAxisValue();
		mSelectedIndexValue = mDisplay.getSelectedIndexValue();
		mFirstIndexValue.addValueChangedListener(this);
		mLastIndexValue.addValueChangedListener(this);
		mSelectedIndexValue.addValueChangedListener(this);
		mSelectedYAxisValue.addValueChangedListener(this);
		
		
		//Text fields for the range specification.
		mMinIndex = new JTextField(5);
		mMaxIndex = new JTextField(5);
		mMinIndex.setText(Integer.toString(mFirstIndexValue.get()));
		mMaxIndex.setText(Integer.toString(mLastIndexValue.get()));
		mRangeLabel = new JLabel("Range: ");
		
		mSelectedPositionLabel = new JLabel("Last selected: none");
		
		mMinIndex.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				updateRange();
			}
		});
		
		mMaxIndex.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				updateRange();
			}
		});
		
		JPanel rangePanel = new JPanel();
		rangePanel.add(mRangeLabel);
		rangePanel.add(mMinIndex);
		rangePanel.add(mMaxIndex);
		
		JPanel lowerPanel = new JPanel(new BorderLayout());
		lowerPanel.add(rangePanel, BorderLayout.EAST);
		lowerPanel.add(mSelectedPositionLabel, BorderLayout.WEST);
		
		setLayout(new BorderLayout());
		add(lowerPanel, BorderLayout.SOUTH);
		add(mDisplay, BorderLayout.CENTER);
		
		Simulator.getInstance().getScheduler().addGenerationListener(this);
	
	}
	
	public PlotDisplay getPlotDisplay() {
		return mDisplay;
	}
	
	
	/**
	 * Updates the range of the PlotDisplay with the values specified in the
	 * two text fields. If the range is not valid, then the range text fields
	 * are reset to the current range settings.
	 */
	public void updateRange() {
		try {
			int minIndex = Integer.parseInt(mMinIndex.getText());
			int maxIndex = Integer.parseInt(mMaxIndex.getText());
			mDisplay.setRangeToPlot(minIndex, maxIndex);
		} catch (NumberFormatException e) {
			mMinIndex.setText(Integer.toString(mFirstIndexValue.get()));
		}
	}

	public void valueChanged(Value value) {
		if(value == null) {
			return;
		}
		if(value == mFirstIndexValue) {
			mMinIndex.setText(Integer.toString(mFirstIndexValue.get()));
		}
		else if(value == mLastIndexValue) {
			mMaxIndex.setText(Integer.toString(mLastIndexValue.get()));
		}
		else if(value == mSelectedIndexValue || value == mSelectedYAxisValue) {
			mSelectedPositionLabel.setText("Last selected: (" + mSelectedIndexValue.get()
					+ "/" + mSelectedYAxisValue.get());
		}
	}

	public void valueReplaced(Value oldValue, Value newValue) {
		mFirstIndexValue = (IntValue) ValueManager.synchronizeValue(mFirstIndexValue, oldValue, newValue, this);
		mLastIndexValue = (IntValue) ValueManager.synchronizeValue(mLastIndexValue, oldValue, newValue, this);
	}

	public void visibilityChanged(boolean visible, Object source) {
		mDisplay.visibilityChanged(visible, source);
	}

	public void generationCompleted(int generationId) {
		mDisplay.setAutoRange(true); //this automatically paints the content and readjusts the range.
	}

	public void generationStarted(int generationId) {
		mDisplay.setAutoRange(true); 
		
	}
	
	
	

}
