/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.visualization.valuepanel;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;

import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueChangedListener;
import de.fhg.iais.outdoor.simba.value.ValueManager;


/**
 * A ValuePanel is a JPanel that provides the control and visualization elements to 
 * manipulate and display a single Value object. The content of the connected Value object
 * is displayed in a JTextField. As long as the checkbox on the left of this text field is set, 
 * the JTextField will be updated automatically if the connected Value object changes. In this 
 * case all changes made to the text field will be overwritten. 
 * 
 * The content of the text field can be changed manually. The color of the text field indicates
 * whether the displayed value and the content of the connected Value are the same. 
 * If the text field background is white, then both contents match, otherwise the background color
 * is red. 
 * A modified content can be set to the connected Value object by pressing the return key 
 * in the text field or by calling the setValueFromContentField() method.
 * 
 * The constructor allows some customization of the ValuePanel. The panel may display the 
 * type of the connected Value object. Also the size of the JTextField can be specified.
 * The displayed name can be choosen freely, while the valueName should point to a real
 * Value in the ValueManager. This is important to support saving and loading from 
 * ValuePanels.
 * 
 * @author Christian Rempis
 *
 */
public class ValuePanel extends JPanel implements ValueChangedListener {
	
	public static final long serialVersionUID = 1001L;
	
	protected JLabel mNameLabel;
	protected JLabel mTypeLabel;
	protected JCheckBox mActivateCheckbox;
	protected JTextField mValueContent;
	protected Value mValue;
	protected boolean mContentChanged;
	protected boolean mShowType;
	protected boolean mObservationEnabled;
	protected String mValueName;
	
	public ValuePanel(Value value, String displayedName, int numberOfColumns, boolean showType, boolean activated) {
		this(value, displayedName, displayedName, numberOfColumns, showType, activated);
	}
	
	/**
	 * Constructor.
	 * 
	 * @param value the value to connect to. This value may be null (the ValuePanel then is disabled).
	 * @param valueName the path name of this Value in the ValueManger. This name should be valid.
	 * @param displayedName the name to display in the ValuePanel to identify the connected Value.
	 * @param numberOfColumns the size of the JTextField.
	 * @param showType indicates whether the type of the connected Value should be shown.
	 * @param activated indicates whether the checkbox is set after construction.
	 */
	public ValuePanel(Value value, String valueName, String displayedName, int numberOfColumns, boolean showType, boolean activated)  {
		
		mObservationEnabled = true;
		mShowType = showType;
		
		mValueName = valueName;
		mNameLabel = new JLabel(displayedName);
		mTypeLabel = new JLabel();
		mActivateCheckbox = new JCheckBox();
		mValueContent = new JTextField();
		mValueContent.setColumns(numberOfColumns);
		
		if(activated) {
			mActivateCheckbox.setSelected(true);
		}
		else {
			mActivateCheckbox.setSelected(false);
		}
		
		//checks if the content is different to the value's content and if the return button was pressed.
		mValueContent.addKeyListener(new KeyAdapter() {
			public void keyReleased(KeyEvent e) {
				if(mValue != null) {
					if(e.getKeyCode() == KeyEvent.VK_ENTER) {
						setValueFromContentField();
					}
					else {
						checkContentFieldChanged();
					}
				}
			}
		});
		
		//changes between listen and passive mode of the content text field.
		mActivateCheckbox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				handleActivateCheckboxChanged();
			}
		});

		setLayout(new BorderLayout());
		JPanel panel = new JPanel();
		panel.add(mActivateCheckbox);
		panel.add(mValueContent);
		if(mShowType) {
			panel.add(mTypeLabel);
		}
		
		add(mNameLabel, BorderLayout.NORTH);
		add(panel, BorderLayout.WEST);
		
		setValue(value);
		disableIfNoValueSet();
		
	}
	
	
	/**
	 * Returns the path name where the connected Value can be found in the ValueManager.
	 * @return
	 */
	public String getValueName() {
		return mValueName;
	}
	
	/**
	 * Returns the name that is displayed to identify the connected Value object.
	 * @return
	 */
	public String getDisplayedName() {
		return mNameLabel.getText();
	}
	
	/**
	 * Returns the embedded Value that is controlled by this ValuePanel.
	 * @return
	 */
	public Value getValue() {
		return mValue;
	}
	
	/**
	 * Sets the activation checkbox.
	 * @param selected
	 */
	public void setActivationCheckboxSelected(boolean selected) {
		mActivateCheckbox.setSelected(selected);
		handleActivateCheckboxChanged();
	}
	
	/**
	 * Handles the ActionEvent that is triggered by the ActivationCheckbox.
	 * If the checkbox is checked, then the content field is in listen mode, 
	 * so each change of the embedded Value is printed out to the content field,
	 * overwriting any other content in this field.
	 * If the checkbox is not checked, then the embedded Value will not be 
	 * observed as ValueChangedListener.
	 * 
	 * @param e the ActionEvent triggered on the ActivationCheckBox.
	 */
	protected void handleActivateCheckboxChanged() {
		disableIfNoValueSet();
		if(mValue != null) {
			if(mActivateCheckbox.isSelected() && mObservationEnabled) {
				mValue.addValueChangedListener(this);
				setContentFieldFromValue();
			}
			else {
				mValue.removeValueChangedListener(this);
			}
		}
	}
	
	
	/**
	 * Sets the Value object for this ValuePanel. If the Value is null, then 
	 * the ValuePanel and all of its components will be deactivated.
	 * @param value the new Value to write to and read from.
	 * @param valueName the path name where the Value can be found in the valueManager.
	 * @param nametoDisplay the name that is displayed in the ValuePanel.
	 */
	public void setValue(Value value, String valueName, String nameToDisplay) {
		if(mValue != null) {
			mValue.removeValueReplacedListener(this);
			mValue.removeValueChangedListener(this);
		}
		mValue = value;
		if(mValue != null) {
			mValue.addValueReplacedListener(this);
			mValueContent.setText(mValue.getValueAsString());
			if(mShowType) {
				String name = value.getClass().getName();
				int index = name.lastIndexOf(".");
				if(index != -1) {
					name = name.substring(index + 1);
				}
				mTypeLabel.setText(name);
			}
			mNameLabel.setText(nameToDisplay);
			mValueName = valueName;
			//mActivateCheckbox.setSelected(true); //was removed to allow disabled startup of a panel.
			handleActivateCheckboxChanged();
		}
		else {
			disableIfNoValueSet();
		}
	}
	
	/**
	 * Sets the Value object for this ValuePanel. If the Value is null, then 
	 * the ValuePanel and all of its components will be deactivated.
	 * @param value the new Value to write to and read from.
	 */
	public void setValue(Value value) {
		setValue(value, mValueName, mNameLabel.getText());
	}

	/**
	 * Called if the observed embedded Value was changed.
	 * Refreshes the content text field with the current Value.
	 */
	public void valueChanged(Value value) {
		if(value == mValue) {
			setContentFieldFromValue();
		}
	}
	
	/**
	 * Tries to set the embedded Value object with the string found
	 * in the content text field. If the string could be parsed to a 
	 * valid value in the Value's setValueFromString method, then the
	 * new value is set. Otherwise the current value of the Value object
	 * is printed in the content text field, overwriting any previous value.
	 */
	public void setValueFromContentField() {
		if(mValue != null) {
			mValue.setValueFromString(mValueContent.getText());
			mValueContent.setText(mValue.getValueAsString());
			checkContentFieldChanged();
		}
	}
	
	/**
	 * Sets the content of the JTextField with the content of the connected Value.
	 */
	public void setContentFieldFromValue() {
		if(mValue != null) {
			mValueContent.setText(mValue.getValueAsString());
			checkContentFieldChanged();
		}
	}
	
	/**
	 * Enables of disables the observation mode. The observation mode
	 * is indicated by the check box. If selected, then the connected Value
	 * is observed, otherwise the content of the text field is not changed
	 * when the content of the Value changes.
	 * 
	 * @param observe
	 */
	public void enableObservation(boolean observe) {
		if(observe && !mObservationEnabled) {
			mObservationEnabled = true;
			handleActivateCheckboxChanged();
		}
		else if(!observe && mObservationEnabled) {
			mObservationEnabled = false;
			handleActivateCheckboxChanged();
		}
	}

	/**
	 * Checks whether the string in the content text field differs from the
	 * current content of the embedded Value object. If the value differs, then
	 * the content text field's background color is changed to red. If the values are
	 * the same, then the background color is set to white.
	 * 
	 * @return true if the content of the content text field differs from the current Value content.
	 */
	public boolean checkContentFieldChanged() {
		if(!mValue.getValueAsString().equals(mValueContent.getText())) {
			if(!mContentChanged) {
				mContentChanged = true;
			}
			mValueContent.setBackground(Color.RED);
		}
		else {
			if(mContentChanged) {
				mContentChanged = false;
			}
			mValueContent.setBackground(Color.WHITE);
		}
		return mContentChanged;
	}

	/**
	 * Called when the embedded Value was replaced by another Value in the ValueManager.
	 */
	public void valueReplaced(Value oldValue, Value newValue) {
		mValue = (Value) ValueManager.synchronizeValue(mValue, oldValue, newValue, this);
		setValue(mValue);
	}
	
	/**
	 * Enables or diables this JPanel.
	 * If the embedded Value is null, then the JPanel and all of its child components are disabled.
	 * If the Value is not null, then the JPanel and all of its child components are enabled 
	 * or disabled according to the parameter of this method.
	 */
	public void disableIfNoValueSet() {
		if(mValue == null || !isEnabled()) {
			mValueContent.setEnabled(false);
			mValueContent.setBackground(Color.LIGHT_GRAY);
			mActivateCheckbox.setEnabled(false);
			//mActivateCheckbox.setSelected(false);
			//super.setEnabled(false);
		}
		else {
			mActivateCheckbox.setEnabled(true);
			checkContentFieldChanged();
			mValueContent.setEnabled(true);
			//super.setEnabled(active);
		}
		
	}
	
	
	/**
	 * Enables of disables the entire ValuePanel.
	 */
	public void setEnabled(boolean enabled) {
		super.setEnabled(enabled);
		disableIfNoValueSet();
	}

	/**
	 * Returns the content of the text field.
	 * @return
	 */
	public String getValueContent() {
		return mValueContent.getText();
	}
	
	/**
	 * Sets the string that is displayed in the text field.
	 * When this method is used to set the text field content, then the 
	 * background is ALWAYS set to red, even if the content matches the content 
	 * of the connected Value.
	 * @param text
	 */
	public void setValueContent(String text) {
		mValueContent.setText(text);
		mContentChanged = true;
		mValueContent.setBackground(Color.RED);
	}
	
	

}
