/***************************************************************************
 *   SIMBA Neuro-Workbench                                                 *
 *                                                                         *
 *   Fraunhofer Institute IAIS                                             *
 *   (Intelligent Analysis and Information Systems)                        *
 *   Sankt Augustin, Germany                                               *
 *   INDY (Intelligent Dynamics) Group                                     *
 *                                                                         *
 *   Project homepage: simba.x-bot.org                                     *
 *                                                                         *
 *   Copyright (C) 2007 by Christian Rempis                                *
 *   Contact: Christian Rempis                                             *
 *   christian.rempis@uni-osnabrueck.de                                    *
 *                                                                         *
 *                                                                         *
 *   Acknowledgments:                                                      *
 *   The SIMBA Workbench is part of the DFG Project OUTDOOR.               *
 *   This work was partially funded (2007) by DFG SPP 1125.                *
 *                                                                         *
 *                                                                         *
 *   License Agreement:                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/


package de.fhg.iais.outdoor.simba.visualization.valuepanel;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;

import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;

import de.fhg.iais.outdoor.simba.communication.NetFileFilter;
import de.fhg.iais.outdoor.simba.core.Simulator;
import de.fhg.iais.outdoor.simba.value.Value;
import de.fhg.iais.outdoor.simba.value.ValueManager;
import de.fhg.iais.outdoor.simba.visualization.OptimizePerformanceListener;
import de.fhg.iais.outdoor.simba.visualization.VisibilityListener;
import de.fhg.iais.outdoor.simba.visualization.Visualization;


/**
 * The ValuePanelDisplay is a flexible component to work with ValuePanels. 
 * This class can be used to create user interfaces in which the user can
 * select, change, observe, load or save all Value objects that are registered
 * at the ValueManager.
 * 
 * When the save button is pressed, then all contents of the text fields of the
 * ValuePanels are saved, NOT the content of the connected Value objects. 
 * 
 * When loading values, then the text fields of all ValuePanels, that were contained 
 * in the loaded file, are set to red, even if the new content of the Value still
 * matches the content of the connected Value object. A warning message will be displayed
 * if not all value objects in the loaded file could be found in the value manager.
 * 
 * After loading a file, the ValuePanelDisplay displays only thouse ValuePanels that
 * represent the loaded values. All other ValuePanels are removed.
 * 
 * To select a Value to display, the user can enter a name of name pattern in the
 * upper text field. Any regular java expression is allowed. The combo box automatically
 * shows all names of value objects that still fit the regular expression. If enter is
 * pressed then all matching value objects are added to the ValuePanelDisplay.
 * 
 * Examples of regular expressions are: 
 * "/Statistic" - selects all Values that start with "/Statistic".
 * "/Stat.*Neuro" - selects all Values that start with "/Stat" and afterwards contain the string "Neuro".
 * 
 * 
 * 
 * @author Christian Rempis
 *
 */
public class ValuePanelDisplay implements Visualization, VisibilityListener, OptimizePerformanceListener {
	
	private JPanel mDisplay;
	private JComboBox  mAddTextField;
	private JButton mAddButton;
	private JButton mApplyButton;
	private JButton mUpdateButton;
	private JButton mSelectAllButton;
	private JButton mSaveButton;
	private JButton mLoadButton;
	private boolean mAllSelected;
	private JButton mRemoveAllButton;
	private Box mValuePanelArea;
	private Vector mValues = new Vector();
	private Vector mPanelWrappers = new Vector();
	private int mNumberOfTextFieldColumns;
	private JScrollPane mScrollPane;
	private boolean mVisible;
	private boolean mOptimizePerformance;
	private JLabel mNumberOfValuesLabel;
	private String mInitialCollection;
	private JFileChooser mFileChooser;
	private int mDisplayId;
	
	protected static int mValueDisplayCounter = 0;
	
	
	public ValuePanelDisplay() {
		this("");
	}

	public ValuePanelDisplay(String initialCollection) {
		
		mDisplayId = mValueDisplayCounter++;
		
		mInitialCollection = initialCollection;
		mNumberOfTextFieldColumns = 15;
		mVisible = false;
		mOptimizePerformance = false;
		
		mFileChooser = new JFileChooser();
		mFileChooser.setFileFilter(new NetFileFilter("Value Files", "val"));
		
		mDisplay = new JPanel(new BorderLayout());
		
		JPanel addPanel = new JPanel(new BorderLayout());
		mAddTextField = new JComboBox();
		mAddTextField.setEditable(true);
		mNumberOfValuesLabel = new JLabel("");
		JPanel addTextPanel = new JPanel(new BorderLayout());
		addTextPanel.add(mAddTextField, BorderLayout.CENTER);
		addTextPanel.add(mNumberOfValuesLabel, BorderLayout.EAST);
		mAddButton = new JButton("Add");
		addPanel.add(addTextPanel, BorderLayout.CENTER);
		addPanel.add(mAddButton, BorderLayout.EAST);
		
		mAddButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				handleAddButton();
			}
		});
		
		mAddTextField.getEditor().getEditorComponent().addKeyListener(new KeyAdapter() {
			public void keyReleased(KeyEvent e) {
				if(e.getKeyCode() == KeyEvent.VK_ENTER) {
					handleAddButton();
					mAddTextField.hidePopup();
					return;
				}
				int code = e.getKeyCode();
				if(code == KeyEvent.VK_KP_DOWN || code == KeyEvent.VK_KP_UP 
						|| code == KeyEvent.VK_UP || code == KeyEvent.VK_DOWN
						|| code == KeyEvent.VK_LEFT || code == KeyEvent.VK_RIGHT
						|| code == KeyEvent.VK_SHIFT) {
					return;
				}
				JTextField editor = ((JTextField) mAddTextField.getEditor().getEditorComponent());
				int position = editor.getCaretPosition();
				String name = (String) ((JTextField) mAddTextField.getEditor().getEditorComponent()).getText();
				Vector names = Simulator.getInstance().getValueManager().getValueNamesMatchingPattern(name + ".*");
				mAddTextField.removeAllItems();
				
				for(Iterator i = names.iterator(); i.hasNext();) {
					mAddTextField.addItem(i.next());
				}
				((JTextField) mAddTextField.getEditor().getEditorComponent()).setText(name);
				editor.setCaretPosition(position);
				
				
				if(names.isEmpty()) {
					mAddButton.setEnabled(false);
					mNumberOfValuesLabel.setText("");
					mAddTextField.hidePopup();
				}
				else {
					mAddButton.setEnabled(true);
					mNumberOfValuesLabel.setText("" + names.size());
					mAddTextField.showPopup();
				}
			}
		});
		
		mAddButton.setEnabled(false);
		
		
		JPanel controlPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));
		mRemoveAllButton = new JButton("Remove All");
		mSelectAllButton = new JButton("Select All");
		mUpdateButton = new JButton("Update");
		mApplyButton = new JButton("Apply");
		mSaveButton = new JButton("Save");
		mLoadButton = new JButton("Load");
		controlPanel.add(mRemoveAllButton);
		controlPanel.add(mSelectAllButton);
		controlPanel.add(mSaveButton);
		controlPanel.add(mLoadButton);
		controlPanel.add(mUpdateButton);
		controlPanel.add(mApplyButton);
		
		
		mApplyButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				for(Iterator i = mPanelWrappers.iterator(); i.hasNext();) {
					ValuePanel vp = (ValuePanel) ((PanelWrapper) i.next()).getWrappedPanel();
					if(vp.checkContentFieldChanged()) {
						vp.setValueFromContentField();
					}
				}
			}
		});
		
		mUpdateButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				for(Iterator i = mPanelWrappers.iterator(); i.hasNext();) {
					ValuePanel vp = (ValuePanel) ((PanelWrapper) i.next()).getWrappedPanel();
					if(vp.checkContentFieldChanged()) {
						vp.setContentFieldFromValue();
					}
				}
			}
		});
		
		mSelectAllButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if(mAllSelected) {
					mSelectAllButton.setText("Select All");
					mAllSelected = false;
				}
				else {
					mSelectAllButton.setText("Deselect All");
					mAllSelected = true;
					//TODO select all.
				}
				for(Iterator i = mPanelWrappers.iterator(); i.hasNext();) {
					ValuePanel vp = (ValuePanel) ((PanelWrapper) i.next()).getWrappedPanel();
					vp.setActivationCheckboxSelected(mAllSelected);
				}
			}
		});
		
		mRemoveAllButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				removeAllValues();
			}
		});
		
		mSaveButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				saveValues();
			}
		});
		
		mLoadButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				loadValues();
			}
		});
		
		checkApplyButtonEnabled();
		checkUpdateButtonEnabled();
		
		
		JPanel mainPanel = new JPanel(new BorderLayout());
		mValuePanelArea = Box.createVerticalBox();
		mainPanel.add(mValuePanelArea, BorderLayout.NORTH);
		mScrollPane = new JScrollPane(mainPanel);
		
		
		mDisplay.add(addPanel, BorderLayout.NORTH);
		mDisplay.add(mScrollPane, BorderLayout.CENTER);
		mDisplay.add(controlPanel, BorderLayout.SOUTH);
		
		Simulator.getInstance().addOptimizePerformanceListener(this);
	}
	
	/**
	 * Adds new PanelWrappers with ValuePanels to the Visualization.
	 *
	 */
	private void handleAddButton() {
		addValuesByPattern(((String) mAddTextField.getSelectedItem()) + ".*");
	}
	
	public void addValuesByPattern(String pattern) {
		synchronized(this) {
			ValueManager vm = Simulator.getInstance().getValueManager();
			
			Vector values = vm.getValueNamesMatchingPattern(pattern);
			addValues(values);
		}
	}
	
	public void addValues(Vector valueNames) {
		if(valueNames == null) {
			return;
		}
		ValueManager vm = Simulator.getInstance().getValueManager();
		for(Iterator i = valueNames.iterator(); i.hasNext();) {
			String name = (String) i.next();
			Value v = vm.getValue(name);
			addValue(v, name, true);
		}
		
		mScrollPane.revalidate();
	}
	
	public PanelWrapper addValue(Value value, String valueName, boolean activated) {
		if(value != null && !mValues.contains(value)) {
			PanelWrapper pw = new PanelWrapper(
					new ValuePanel(value, valueName, mNumberOfTextFieldColumns, true, activated), this);
			mValues.add(value);
			mPanelWrappers.add(pw);
			mValuePanelArea.add(pw);
			mValuePanelArea.validate();
			return pw;
		}
		return null;
	}
	
	
	public void removeAllValues() {
		synchronized(this) {
			for(Iterator i = mPanelWrappers.iterator(); i.hasNext();) {
				ValuePanel vp = (ValuePanel) ((PanelWrapper) i.next()).getWrappedPanel();
				vp.setValue(null);
			}
			mValues.clear();
			mPanelWrappers.clear();
			
			mValuePanelArea.removeAll();
			mValuePanelArea.invalidate();
			mScrollPane.revalidate();
		}
	}
	
	public void checkUpdateButtonEnabled() {
		//TODO
	}
	
	public void checkApplyButtonEnabled() {
		//TODO
	}
	
	
	//Saves the currently displayed values to a file. The content of the text files is saved,
	//not the current value of the represented Value object.
	public void saveValues() {
		synchronized(this) {
			//choose directory.
			Properties sysProps = Simulator.getInstance().getSystemProperties();
			String directory = sysProps.getProperty("/FileChooser/CurrentDirectory");
			if(directory != null) {
				mFileChooser.setCurrentDirectory(new File(directory));
			}
			mFileChooser.setApproveButtonText("Save");
			
			if(mFileChooser.showSaveDialog(null) == JFileChooser.APPROVE_OPTION) {
				File file = mFileChooser.getSelectedFile();
				if(!file.getName().endsWith(".val")) {
					file = new File(file.getAbsolutePath() + ".val");
				}
				//Memorize directory
				directory = mFileChooser.getCurrentDirectory().getPath();
				sysProps.put("/FileChooser/CurrentDirectory", directory);
				
				if(file.exists()) {
					int res = JOptionPane.showConfirmDialog(null, "A file with name " + file.getName() + " already exists.\nDo you want to overwrite it?", 
									"File exists warning", JOptionPane.OK_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE);
					if(res == JOptionPane.CANCEL_OPTION) {
						return;
					}
				}
				
				Properties props = new Properties();
				for(Iterator i = mPanelWrappers.iterator(); i.hasNext();) {
					PanelWrapper wrapper = (PanelWrapper) i.next();
					ValuePanel panel = (ValuePanel) wrapper.getWrappedPanel();
					if(panel == null) {
						continue;
					}
					//System.out.println("Name: " + panel.getValueName() + "  content: " + panel.getValueContent());
					props.setProperty(panel.getValueName(), panel.getValueContent());
				}
				
				try {
					FileOutputStream output = new FileOutputStream(file);
					props.store(output, "Saved Values for the Simba Simulator.");
				} catch (FileNotFoundException e) {
					JOptionPane.showMessageDialog(null, "Could not save the Values.", 
							"Save error", JOptionPane.WARNING_MESSAGE);
			
				} catch (IOException e) {
					JOptionPane.showMessageDialog(null, "Could not save the Values.", 
							"Save error", JOptionPane.WARNING_MESSAGE);
				}
				
			}
			else {
				//Memorize directory
				directory = mFileChooser.getCurrentDirectory().getPath();
				sysProps.put("/FileChooser/CurrentDirectory", directory);
			}
		}
	}
	
	//Loads a set of values from a file.
	public void loadValues() {
		synchronized(this) {
			//choose directory.
			Properties sysProps = Simulator.getInstance().getSystemProperties();
			String directory = sysProps.getProperty("/FileChooser/CurrentDirectory");
			if(directory != null) {
				mFileChooser.setCurrentDirectory(new File(directory));
			}
			mFileChooser.setApproveButtonText("Load");
			
			if(mFileChooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
				File file = mFileChooser.getSelectedFile();
				if(!file.getName().endsWith(".val")) {
					file = new File(file.getAbsolutePath() + ".val");
				}
				//Memorize directory
				directory = mFileChooser.getCurrentDirectory().getPath();
				sysProps.put("/FileChooser/CurrentDirectory", directory);
				
				Properties props = new Properties();
				try {
					FileInputStream inputStream = new FileInputStream(file);
					props.load(inputStream);
				} catch (FileNotFoundException e1) {
					JOptionPane.showMessageDialog(null, "Could not find the specified file" + file.getName(), 
							"File does not exist", JOptionPane.WARNING_MESSAGE);
					return;
				} catch (IOException e1) {
					JOptionPane.showMessageDialog(null, "Could not load the Values.", 
							"Load error", JOptionPane.WARNING_MESSAGE);
					return;
				}
				
				removeAllValues();
				
				String errorMessage = "Could not find the following Values:\n\n";
				boolean errorOccured = false;
				ValueManager vm = Simulator.getInstance().getValueManager();
				Enumeration keys = props.keys();
				for(; keys.hasMoreElements();) {
					String valueName = (String) keys.nextElement();
					
					Value v = vm.getValue(valueName);
					if(v == null) {
						errorMessage += "- " + valueName + "\n";
						errorOccured = true;
						continue;
					}
					PanelWrapper pw = addValue(v, valueName, false);
					if(pw != null) {
						ValuePanel panel = (ValuePanel) pw.getWrappedPanel();
						if(panel != null) {
							panel.setValueContent(props.getProperty(valueName));
						}
					}
				}
				if(errorOccured) {
					JOptionPane.showMessageDialog(null, errorMessage, 
							"Warning", JOptionPane.WARNING_MESSAGE);
				}
			}
			else {
				//Memorize directory
				directory = mFileChooser.getCurrentDirectory().getPath();
				sysProps.put("/FileChooser/CurrentDirectory", directory);
			}
		}
	}

	/**
	 * Returns the JPanel that contains the components for this Visualization.
	 */
	public JPanel getPanel() {
		return mDisplay;
	}

	public boolean cleanUp() {
		
		if(mInitialCollection.equals("")) {
			try {
				BufferedWriter writer = new BufferedWriter(new FileWriter(
						Simulator.getInstance().getConfigDirectoryName() + "valuepanel" + mDisplayId + ".conf"));
				for(Iterator i = mPanelWrappers.iterator(); i.hasNext();) {
					PanelWrapper pw = (PanelWrapper) i.next();
					if(pw != null && pw.getWrappedPanel() != null) {
						ValuePanel panel = (ValuePanel) pw.getWrappedPanel();
						writer.write(panel.getValueName() + "\n");
					}
				}
				writer.close();
			} catch (IOException e) {
			}
		}
		
		return true;
	}

	public boolean init() {
		synchronized(this) {
			if(!mInitialCollection.equals("")) {
				addValuesByPattern(mInitialCollection);
			}
			else {
				try {
					BufferedReader reader = new BufferedReader(
							new FileReader(Simulator.getInstance().getConfigDirectoryName() + "valuepanel" + mDisplayId + ".conf"));
					Vector valueNames = new Vector();
					String name = reader.readLine();
					while(name != null) {
						valueNames.add(name);
						name = reader.readLine();
					}
					addValues(valueNames);
				} catch (IOException e) {
				}
			}
		}
		return true;
	}
	
	
	/**
	 * Called whenever the source object gets visible or invisible. Usually this is a
	 * Frame like the FrameVisualizationArea that indicates whether or not the Frame
	 * with all its content is visible. 
	 * In the case that the Frame is not visible, all observations of all ValuePanels
	 * are temporarily disabled to increase the performance (if the frame is not seen
	 * then there is no need to request values).
	 */
	public void visibilityChanged(boolean visible, Object source) {
		mVisible = visible;
		for(Iterator i = mPanelWrappers.iterator(); i.hasNext();) {
			ValuePanel vp = (ValuePanel) ((PanelWrapper) i.next()).getWrappedPanel();
			vp.enableObservation(mVisible && !mOptimizePerformance);
		}
	}
	
	
	/**
	 * Called if the system was set to optimized performance mode.
	 * In this mode this Visualization disables the observation of all ValuePanels in the
	 * PanelWrappers to avoid computations for displaying values.
	 */
	public void optimizePerformance(boolean optimize) {
		mOptimizePerformance = optimize;
		
		//Prevent concurrent modification exception through changes of the Valuelisteners of the performance Value.
		Value optimizePerformanceValue = Simulator.getInstance().getPerformanceOptimizationValue();
		
		
		for(Iterator i = mPanelWrappers.iterator(); i.hasNext();) {
			ValuePanel vp = (ValuePanel) ((PanelWrapper) i.next()).getWrappedPanel();
			if(vp.getValue() != optimizePerformanceValue) { //exclude the ValuePanel of the Performance Value.
				vp.enableObservation(mVisible && !mOptimizePerformance);
			}
		}
	}

	
	/**
	 * Moves the specified PanelWrapper up or down in the list of Wrappers.
	 * @param up if true, then the PanelWrapper is moved up, otherwise it is moved down.
	 * @param wrapper the PanelWrapper to move.
	 */
	public void moveWrapper(boolean up, PanelWrapper wrapper) {
		
		synchronized (this) {
			for(int i = 0; i < mValuePanelArea.getComponentCount(); ++i) {
				if(wrapper == mValuePanelArea.getComponent(i)) {
					if(up && i > 0) {
						mValuePanelArea.remove(wrapper);
						mValuePanelArea.add(wrapper, i - 1);
						break;
					}
					else if(!up && i < mValuePanelArea.getComponentCount() - 1) {
						mValuePanelArea.remove(wrapper);
						mValuePanelArea.add(wrapper, i + 1);
						break;
					}
				}
			}
		}
		mValuePanelArea.validate();
		mScrollPane.revalidate();
	}
	
	
	/**
	 * Removes the specified PanelWrapper cleanly from the visualization component.
	 * @param wrapper
	 */
	public void removeWrapper(PanelWrapper wrapper) {
		if(wrapper == null) {
			return;
		}
		synchronized (this) {
			mValues.remove(((ValuePanel) wrapper.getWrappedPanel()).getValue());
			((ValuePanel) wrapper.getWrappedPanel()).setValue(null);
			mPanelWrappers.remove(wrapper);
			mValuePanelArea.remove(wrapper);
			mValuePanelArea.validate();
			mScrollPane.revalidate();
		}
	}

}
